/*
FILE:     _maxit.C
*/
/*
VERSION:  8.200
*/
/*
DATE:     3/18/2015
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2015 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <map>
#include <set>
#include <vector>

#include "_maxit.h"
#include "erf_category.h"
#include "ndb_misclib.h"
#include "erf_util.h"
#include "util.h"
#include "_fileobj_ioutil.h"

#define NUM_TOKENS  91

typedef struct {
        CifString asym_ids;
        CifString symmetries;
} _ASYM_SYMM_PAIR;

typedef struct {
        int biol_id;
        ReVarCifArray<_ASYM_SYMM_PAIR> asym_symm_array;
} _BIOL_ASYM_SYMMETRY;

typedef struct {
        CifString asym_id;
        ReVarCifArray<CifString> symmetry;
} _ASYM_SYMMETRY;

typedef struct {
        int biol_id;
        ReVarCifArray<_ASYM_SYMMETRY> asym_array;
} _STRUC_BIOL;

typedef struct {
        CifString entity_id;
        CifString n_terminal;
        CifString c_terminal;
        int       is_protein;
} _ENTITY_TERMINAL_SEQ;

static char merge_tokens[NUM_TOKENS][7] = {
       "BVALUE", "COFOFC", "CRMET1", "CRMETH", "DENSTY", "DIFPTL", "DTMEA1", "DTMEAS", "DTMETH", "DTTEMP",
       "EXPDTA", "FREERF", "NCSMOD", "NMRECD", "NMREXP", "NMRPST", "NMRSDT", "NMRSFT", "NMRSMB", "NMRSMP",
       "NMRSPM", "PERCOM", "PERWAT",  "PHVAL", "R4SCUT", "RADIAT", "REFERR", "REFESU", "REFINE", "REFLEC",
       "REFMET", "REFREM", "RESTOT", "RFACTR", "RMSCON", "RMSDIN", "RMSDIS", "RMSGEN", "RMSISN", "RMSISO",
       "RMSNBO", "RMSPLA", "RMSRF1", "RMSRF2", "RMSRF3", "RMSRF4", "RMSRF5", "RMSRF6", "RMSRF7", "RMSRF8",
       "RMSTN1", "RMSTN2", "RMSVOL", "RNOCUT", "SHELC",  "SHELL",  "SHELXL", "SOLMOD", "SFTWAR", "WAVLEN",
       "XFILE1", "XFILE2", "XFILE3", "XFILES", "SOLEXP", "SOLMDL", "CMRFMT", "CEMFIT", "CMRNST", "EMSAMP",
       "EMDATA", "ASSEMB", "VIRPAR", "TLSRNG", "TLSGRO", "TTENSR", "LTENSR", "STENSR", "RFTPOS", "RFMPOS",
       "RFLPOS", "RFTTHR", "RFMTHR", "RFLTHR", "CCPNCS", "TMLPTL", "PEROBS", "NCSGRO", "TWIN",   "NCSTLS",
       "PHENCS"
};

static void get_line_from_file(FILE *fp, CifString &string);

static void add_terminal_residues(ISTable *Table, int &row, CifString &id, CifString &entity,
               CifString &pdb_num, CifString &pdb_id, const int id_index, const int entity_index,
               const int mon_index, const int num_index, const int pdb_index, const int type,
               ReVarCifArray<_ENTITY_TERMINAL_SEQ> &_array);

// static void get_pdbformat_symmetry(char *symm, const char *old_symm);

_MAXIT::_MAXIT(const int mxtokens, const int mxjrnltokens): _VALIDATION(mxtokens, mxjrnltokens)
{
       _alloc_sheet = 0; _n_sheet = 0; _sheet = NULL;
       // _alloc_ssbond = 0; _n_ssbond = 0; _ssbond = NULL;
       _alloc_hydbnd = 0; _n_hydbnd = 0; _hydbnd = NULL;
       _alloc_sltbrg = 0; _n_sltbrg = 0; _sltbrg = NULL;
       _alloc_site = 0; _n_site = 0; _site = NULL;
       _alloc_het = 0; _n_het = 0; _het = NULL;
       _alloc_modifb = 0; _n_modifb = 0; _modifb = NULL;
       _alloc_modifp = 0; _n_modifp = 0; _modifp = NULL;
       _alloc_modifs = 0; _n_modifs = 0; _modifs = NULL;
       _alloc_modres = 0; _n_modres = 0; _modres = NULL;
       _n_entity = 0;
       _entity = NULL;
       _exist_helix = 0;
       _exist_sheet = 0;
       _exist_ssbond = 0;
       _create_final_file = 0;
       _n_biol_unit_info = 0; _biol_unit_info = NULL;
       _remark_decimal_precision = FORCE_DECIMAL_PRECISION;
       _joint_methods.Clear();
       _current_method.Clear();
       _automatic_modres_flag = 1;
}

_MAXIT::~_MAXIT()
{
       int i, j;
       Clear_Structural_Feature();
       if (_entity) {
            for (i = 0; i < _n_entity; i++) {
                 if (_entity[i].Type) delete [] _entity[i].Type;
                 if (_entity[i].Source) delete [] _entity[i].Source;
                 if (_entity[i].Description) delete [] _entity[i].Description;
            }
            delete [] _entity;
       }
       if (_biol_unit_info) {
            for (i = 0; i < _n_biol_unit_info; i++) {
                 if (_biol_unit_info[i]) {
                      if (_biol_unit_info[i]->stbiol) {
                           for (j = 0; j < _biol_unit_info[i]->n_stbiol; j++) {
                                if (_biol_unit_info[i]->stbiol[j]) {
                                     delete _biol_unit_info[i]->stbiol[j];
                                }
                           }
                           delete [] _biol_unit_info[i]->stbiol;
                      }
                      delete _biol_unit_info[i];
                 }
            }
            delete [] _biol_unit_info;
       }
}

void _MAXIT::Clear_Structural_Feature()
{
       int i;
       if (_sheet) {
            for (i = 0; i < _n_sheet; i++)
                 if (_sheet[i]) delete _sheet[i];
            delete [] _sheet;
       }
       if (_ssbond) {
            for (i = 0; i < _n_ssbond; i++)
                 if (_ssbond[i]) delete _ssbond[i];
            delete [] _ssbond;
       }
       if (_hydbnd) {
            for (i = 0; i < _n_hydbnd; i++)
                 if (_hydbnd[i]) delete _hydbnd[i];
            delete [] _hydbnd;
       }
       if (_sltbrg) {
            for (i = 0; i < _n_sltbrg; i++) {
                 if (_sltbrg[i]) {
                      if (_sltbrg[i]->details) delete _sltbrg[i]->details;
                      delete _sltbrg[i];
                 }
            }
            delete [] _sltbrg;
       }
       Clear_Site();
       if (_het) {
            for (i = 0; i < _n_het; i++)
                 if (_het[i]) delete _het[i];
            delete [] _het;
       }
       if (_modifb) {
            for (i = 0; i < _n_modifb; i++)
                 if (_modifb[i]) delete _modifb[i];
            delete [] _modifb;
       }
       if (_modifp) {
            for (i = 0; i < _n_modifp; i++)
                 if (_modifp[i]) delete _modifp[i];
            delete [] _modifp;
       }
       if (_modifs) {
            for (i = 0; i < _n_modifs; i++)
                 if (_modifs[i]) delete _modifs[i];
            delete [] _modifs;
       }
       if (_modres) {
            for (i = 0; i < _n_modres; i++) {
                 if (_modres[i]) {
                      if (_modres[i]->details) delete [] _modres[i]->details;
                      delete _modres[i];
                 }
            }
            delete [] _modres;
       }
}

void _MAXIT::Clear_Site()
{
       if (_site) {
            for (int i = 0; i < _n_site; i++) {
                 if (_site[i]) {
                      if (_site[i]->initIndex)
                           delete [] _site[i]->initIndex;
                      if (_site[i]->initRes)
                           delete [] _site[i]->initRes;
                      if (_site[i]->SymOP) {
                           for (int j = 0; j < _site[i]->numRes; j++) {
                                if (_site[i]->SymOP[j]) delete [] _site[i]->SymOP[j];
			   }
                           delete [] _site[i]->SymOP;
                      }
                      if (_site[i]->residue) delete [] _site[i]->residue;
                      delete _site[i];
                 }
            }
            delete [] _site;
       }
       _site = NULL; _n_site = 0; _alloc_site = 0;
}

void _MAXIT::Insert_a_Sheet(const char *id, const int numStrands, const int initIndex,
               const int initRes, const int endIndex, const int endRes, const int sense,
               const char *curAtom, const int curIndex, const int curRes,
               const char *prevAtom, const int prevIndex, const int prevRes)
{
       int i;
       _SHEET** tmp = NULL;
       if (_n_sheet >= _alloc_sheet) {
            _alloc_sheet += 20;
            tmp = _sheet;
            _sheet = new _SHEET*[_alloc_sheet];
            for (i = 0; i < _n_sheet; i++) _sheet[i] = tmp[i];
            for (i = _n_sheet; i < _alloc_sheet; i++) _sheet[i] = NULL;
            delete [] tmp;
       }
       _sheet[_n_sheet] = new _SHEET;
       strcpy(_sheet[_n_sheet]->sheetID, id);
       _sheet[_n_sheet]->numStrands = numStrands;
       _sheet[_n_sheet]->initIndex = initIndex;
       _sheet[_n_sheet]->initRes = initRes;
       _sheet[_n_sheet]->endIndex = endIndex;
       _sheet[_n_sheet]->endRes = endRes;
       _sheet[_n_sheet]->sense = sense;
       strcpy(_sheet[_n_sheet]->curAtom, curAtom);
       _sheet[_n_sheet]->curIndex = curIndex;
       _sheet[_n_sheet]->curRes = curRes;
       strcpy(_sheet[_n_sheet]->prevAtom, prevAtom);
       _sheet[_n_sheet]->prevIndex = prevIndex;
       _sheet[_n_sheet]->prevRes = prevRes;
       _n_sheet++;
}

void _MAXIT::Insert_a_Hydbnd(const char *FstAtom, const char *altLoc1, const char *FstName,
           const int FstIndex, const int FstRes, const char *H_Atom, const char *altLocH,
           const char *H_Name, const int H_Index, const int H_Res, const char *SndAtom,
           const char *altLoc2, const char *SndName, const int SndIndex, const int SndRes,
           const char *SymOP_1, const char *SymOP_2)
{
       int i;
       _HYDBND** tmp = NULL;
       for (i = 0; i < _n_hydbnd; i++) {
            if (!strcmp(_hydbnd[i]->FstAtom, FstAtom) &&
                !strcmp(_hydbnd[i]->altLoc1, altLoc1) &&
                !strcmp(_hydbnd[i]->FstName, FstName) &&
                _hydbnd[i]->FstIndex == FstIndex && _hydbnd[i]->FstRes == FstRes &&
                !strcmp(_hydbnd[i]->H_Atom, H_Atom) &&
                !strcmp(_hydbnd[i]->altLocH, altLocH) &&
                !strcmp(_hydbnd[i]->H_Name, H_Name) &&
                _hydbnd[i]->H_Index == H_Index && _hydbnd[i]->H_Res == H_Res &&
                !strcmp(_hydbnd[i]->SndAtom, SndAtom) &&
                !strcmp(_hydbnd[i]->altLoc2, altLoc2) &&
                !strcmp(_hydbnd[i]->SndName, SndName) &&
                _hydbnd[i]->SndIndex == SndIndex && _hydbnd[i]->SndRes == SndRes &&
                !strcmp(_hydbnd[i]->SymOP_1, SymOP_1) &&
                !strcmp(_hydbnd[i]->SymOP_2, SymOP_2)) break;
       }
       if (i < _n_hydbnd) return;

       if (_n_hydbnd >= _alloc_hydbnd) {
            _alloc_hydbnd += 5;
            tmp = _hydbnd;
            _hydbnd = new _HYDBND*[_alloc_hydbnd];
            for (i = 0; i < _n_hydbnd; i++) _hydbnd[i] = tmp[i];
            for (i = _n_hydbnd; i < _alloc_hydbnd; i++) _hydbnd[i] = NULL;
            delete [] tmp;
       }
       _hydbnd[_n_hydbnd] = new _HYDBND;
       strcpy(_hydbnd[_n_hydbnd]->FstAtom, FstAtom);
       strcpy(_hydbnd[_n_hydbnd]->altLoc1, altLoc1);
       strcpy(_hydbnd[_n_hydbnd]->FstName, FstName);
       _hydbnd[_n_hydbnd]->FstIndex = FstIndex;
       _hydbnd[_n_hydbnd]->FstRes = FstRes;
       strcpy(_hydbnd[_n_hydbnd]->H_Atom, H_Atom);
       strcpy(_hydbnd[_n_hydbnd]->altLocH, altLocH);
       strcpy(_hydbnd[_n_hydbnd]->H_Name, H_Name);
       _hydbnd[_n_hydbnd]->H_Index = H_Index;
       _hydbnd[_n_hydbnd]->H_Res = H_Res;
       strcpy(_hydbnd[_n_hydbnd]->SndAtom, SndAtom);
       strcpy(_hydbnd[_n_hydbnd]->altLoc2, altLoc2);
       strcpy(_hydbnd[_n_hydbnd]->SndName, SndName);
       _hydbnd[_n_hydbnd]->SndIndex = SndIndex;
       _hydbnd[_n_hydbnd]->SndRes = SndRes;
       strcpy(_hydbnd[_n_hydbnd]->SymOP_1, SymOP_1);
       strcpy(_hydbnd[_n_hydbnd]->SymOP_2, SymOP_2);
       _n_hydbnd++;
}

void _MAXIT::Insert_a_Sltbrg(const char *FstAtom, const char *altLoc1, const char *FstName,
           const int FstIndex, const int FstRes, const char *SndAtom, const char *altLoc2,
           const char *SndName, const int SndIndex, const int SndRes, const char *SymOP_1,
           const char *SymOP_2)
{
       int i;
       _SLTBRG** tmp = NULL;
       for (i = 0; i < _n_sltbrg; i++) {
            if (!strcmp(_sltbrg[i]->FstAtom, FstAtom) &&
                !strcmp(_sltbrg[i]->altLoc1, altLoc1) &&
                !strcmp(_sltbrg[i]->FstName, FstName) &&
                _sltbrg[i]->FstIndex == FstIndex && _sltbrg[i]->FstRes == FstRes &&
                !strcmp(_sltbrg[i]->SndAtom, SndAtom) &&
                !strcmp(_sltbrg[i]->altLoc2, altLoc2) &&
                !strcmp(_sltbrg[i]->SndName, SndName) &&
                _sltbrg[i]->SndIndex == SndIndex && _sltbrg[i]->SndRes == SndRes &&
                !strcmp(_sltbrg[i]->SymOP_1, SymOP_1) &&
                !strcmp(_sltbrg[i]->SymOP_2, SymOP_2)) break;
       }
       if (i < _n_sltbrg) return;

       if (_n_sltbrg >= _alloc_sltbrg) {
            _alloc_sltbrg += 5;
            tmp = _sltbrg;
            _sltbrg = new _SLTBRG*[_alloc_sltbrg];
            for (i = 0; i < _n_sltbrg; i++) _sltbrg[i] = tmp[i];
            for (i = _n_sltbrg; i < _alloc_sltbrg; i++) _sltbrg[i] = NULL;
            delete [] tmp;
       }
       _sltbrg[_n_sltbrg] = new _SLTBRG;
       strcpy(_sltbrg[_n_sltbrg]->FstAtom, FstAtom);
       strcpy(_sltbrg[_n_sltbrg]->altLoc1, altLoc1);
       strcpy(_sltbrg[_n_sltbrg]->FstName, FstName);
       _sltbrg[_n_sltbrg]->FstIndex = FstIndex;
       _sltbrg[_n_sltbrg]->FstRes = FstRes;
       strcpy(_sltbrg[_n_sltbrg]->SndAtom, SndAtom);
       strcpy(_sltbrg[_n_sltbrg]->altLoc2, altLoc2);
       strcpy(_sltbrg[_n_sltbrg]->SndName, SndName);
       _sltbrg[_n_sltbrg]->SndIndex = SndIndex;
       _sltbrg[_n_sltbrg]->SndRes = SndRes;
       strcpy(_sltbrg[_n_sltbrg]->SymOP_1, SymOP_1);
       strcpy(_sltbrg[_n_sltbrg]->SymOP_2, SymOP_2);
       _sltbrg[_n_sltbrg]->details = NULL;
       _n_sltbrg++;
}

void _MAXIT::Insert_a_Site(const char *SiteID, const int initIndex, const int initRes, const char *SymOP_1)
{
       int i, j, *tmp_int = NULL;
       char **tmp_char = NULL, SymOP[7];
       _SITE **tmp_site = NULL;

       if (SymOP_1 == NULL || !strcmp(SymOP_1, "") || !strcmp(SymOP_1, "?") || !strcmp(SymOP_1, "."))
            strcpy(SymOP, "1_555");
       else strcpy(SymOP, SymOP_1);

       for (i = 0; i < _n_site; i++) {
            if (!strcmp(SiteID, _site[i]->SiteID)) break;
       }
       if (i == _n_site) {
            if (_n_site >= _alloc_site) {
                 _alloc_site += 10;
                 tmp_site = _site;
                 _site = new _SITE*[_alloc_site];
                 for (j = 0; j < _n_site; j++) _site[j] = tmp_site[j];
                 for (j = _n_site; j < _alloc_site; j++) _site[j] = NULL;
                 delete [] tmp_site;
            }
            _site[_n_site] = new _SITE;
            _n_site++;
            strcpy(_site[i]->SiteID, SiteID);
            _site[i]->numRes = 0;
            _site[i]->alloc_numRes = 5;
            _site[i]->initIndex = new int[5];
            _site[i]->initRes = new int[5];
            _site[i]->SymOP = new char*[5];
            _site[i]->residue = NULL;
            for (j = 0; j < _site[i]->alloc_numRes; j++) {
                 _site[i]->SymOP[j] = NULL;
            }
       }
       for (j = 0; j < _site[i]->numRes; j++) {
            if (_site[i]->initIndex[j] == initIndex &&
                _site[i]->initRes[j] == initRes &&
                !strcmp(_site[i]->SymOP[j], SymOP)) break;
       }
       if (j == _site[i]->numRes) {
            if (_site[i]->numRes >= _site[i]->alloc_numRes) {
                 _site[i]->alloc_numRes += 5;
                 tmp_int = _site[i]->initIndex;
                 _site[i]->initIndex = new int[_site[i]->alloc_numRes];
                 for (j = 0; j < _site[i]->numRes; j++)
                      _site[i]->initIndex[j] = tmp_int[j];
                 delete [] tmp_int;
                 tmp_int = _site[i]->initRes;
                 _site[i]->initRes = new int[_site[i]->alloc_numRes];
                 for (j = 0; j < _site[i]->numRes; j++)
                      _site[i]->initRes[j] = tmp_int[j];
                 delete [] tmp_int;
                 tmp_char = _site[i]->SymOP;
                 _site[i]->SymOP = new char*[_site[i]->alloc_numRes];
                 for (j = 0; j < _site[i]->numRes; j++)
                      _site[i]->SymOP[j] = tmp_char[j];
                 delete [] tmp_char;
                 for (j = _site[i]->numRes; j < _site[i]->alloc_numRes; j++)
                      _site[i]->SymOP[j] = NULL;
            }
            _site[i]->initIndex[_site[i]->numRes] = initIndex;
            _site[i]->initRes[_site[i]->numRes] = initRes;
            _site[i]->SymOP[_site[i]->numRes] = new char[strlen(SymOP)+1];
            strcpy(_site[i]->SymOP[_site[i]->numRes], SymOP);
            _site[i]->numRes++;
       }
}

void _MAXIT::Insert_a_Het(const int initIndex, const int initRes, const int numAtoms)
{
       int i;
       _HET** tmp = NULL;
       for (i = 0; i < _n_het; i++) {
            if (_het[i]->initIndex == initIndex && _het[i]->initRes == initRes) break;
       }
       if (i < _n_het) return;

       if (_n_het >= _alloc_het) {
            _alloc_het += 5;
            tmp = _het;
            _het = new _HET*[_alloc_het];
            for (i = 0; i < _n_het; i++) _het[i] = tmp[i];
            for (i = _n_het; i < _alloc_het; i++) _het[i] = NULL;
            delete [] tmp;
       }
       _het[_n_het] = new _HET;
       _het[_n_het]->initIndex = initIndex;
       _het[_n_het]->initRes = initRes;
       _het[_n_het]->numAtoms = numAtoms;
       _n_het++;
}

void _MAXIT::Insert_a_Modifb(const char *Name, const char *FstAtom, const char *FstName,
           const int FstIndex, const int FstRes, const char *SndAtom, const char *SndName,
           const int SndIndex, const int SndRes, const char *Replaced_Atom)
{
       int i;
       _MODIFB** tmp = NULL;
       for (i = 0; i < _n_modifb; i++) {
            if (!strcmp(_modifb[i]->FstAtom, FstAtom) &&
                !strcmp(_modifb[i]->FstName, FstName) &&
                _modifb[i]->FstIndex == FstIndex && _modifb[i]->FstRes == FstRes &&
                !strcmp(_modifb[i]->SndAtom, SndAtom) &&
                !strcmp(_modifb[i]->SndName, SndName) &&
                _modifb[i]->SndIndex == SndIndex && _modifb[i]->SndRes == SndRes &&
                !strcmp(_modifb[i]->Replaced_Atom, Replaced_Atom)) break;
       }
       if (i < _n_modifb) return;

       if (_n_modifb >= _alloc_modifb) {
            _alloc_modifb += 5;
            tmp = _modifb;
            _modifb = new _MODIFB*[_alloc_modifb];
            for (i = 0; i < _n_modifb; i++) _modifb[i] = tmp[i];
            for (i = _n_modifb; i < _alloc_modifb; i++) _modifb[i] = NULL;
            delete [] tmp;
       }
       _modifb[_n_modifb] = new _MODIFB;
       strcpy(_modifb[_n_modifb]->Name, Name);
       strcpy(_modifb[_n_modifb]->FstAtom, FstAtom);
       strcpy(_modifb[_n_modifb]->FstName, FstName);
       _modifb[_n_modifb]->FstIndex = FstIndex;
       _modifb[_n_modifb]->FstRes = FstRes;
       strcpy(_modifb[_n_modifb]->SndAtom, SndAtom);
       strcpy(_modifb[_n_modifb]->SndName, SndName);
       _modifb[_n_modifb]->SndIndex = SndIndex;
       _modifb[_n_modifb]->SndRes = SndRes;
       strcpy(_modifb[_n_modifb]->Replaced_Atom, Replaced_Atom);
       _n_modifb++;
}

void _MAXIT::Insert_a_Modifp(const char *Name, const char *FstAtom, const char *FstName,
           const int FstIndex, const int FstRes, const char *SndAtom, const char *SndName,
           const int SndIndex, const int SndRes, const char *Replaced_Atom)
{
       int i;
       _MODIFP** tmp = NULL;
       for (i = 0; i < _n_modifp; i++) {
            if (!strcmp(_modifp[i]->FstAtom, FstAtom) &&
                !strcmp(_modifp[i]->FstName, FstName) &&
                _modifp[i]->FstIndex == FstIndex && _modifp[i]->FstRes == FstRes &&
                !strcmp(_modifp[i]->SndAtom, SndAtom) &&
                !strcmp(_modifp[i]->SndName, SndName) &&
                _modifp[i]->SndIndex == SndIndex && _modifp[i]->SndRes == SndRes &&
                !strcmp(_modifp[i]->Replaced_Atom, Replaced_Atom)) break;
       }
       if (i < _n_modifp) return;

       if (_n_modifp >= _alloc_modifp) {
            _alloc_modifp += 5;
            tmp = _modifp;
            _modifp = new _MODIFP*[_alloc_modifp];
            for (i = 0; i < _n_modifp; i++) _modifp[i] = tmp[i];
            for (i = _n_modifp; i < _alloc_modifp; i++) _modifp[i] = NULL;
            delete [] tmp;
       }
       _modifp[_n_modifp] = new _MODIFP;
       strcpy(_modifp[_n_modifp]->Name, Name);
       strcpy(_modifp[_n_modifp]->FstAtom, FstAtom);
       strcpy(_modifp[_n_modifp]->FstName, FstName);
       _modifp[_n_modifp]->FstIndex = FstIndex;
       _modifp[_n_modifp]->FstRes = FstRes;
       strcpy(_modifp[_n_modifp]->SndAtom, SndAtom);
       strcpy(_modifp[_n_modifp]->SndName, SndName);
       _modifp[_n_modifp]->SndIndex = SndIndex;
       _modifp[_n_modifp]->SndRes = SndRes;
       strcpy(_modifp[_n_modifp]->Replaced_Atom, Replaced_Atom);
       _n_modifp++;
}

void _MAXIT::Insert_a_Modifs(const char *Name, const char *FstAtom, const char *FstName,
           const int FstIndex, const int FstRes, const char *SndAtom, const char *SndName,
     const int SndIndex, const int SndRes, const char *Replaced_Atom, const char *Sugar_Name)
{
       int i;
       _MODIFS** tmp = NULL;
       for (i = 0; i < _n_modifs; i++) {
            if (!strcmp(_modifs[i]->FstAtom, FstAtom) &&
                !strcmp(_modifs[i]->FstName, FstName) &&
                _modifs[i]->FstIndex == FstIndex && _modifs[i]->FstRes == FstRes &&
                !strcmp(_modifs[i]->SndAtom, SndAtom) &&
                !strcmp(_modifs[i]->SndName, SndName) &&
                _modifs[i]->SndIndex == SndIndex && _modifs[i]->SndRes == SndRes &&
                !strcmp(_modifs[i]->Replaced_Atom, Replaced_Atom) &&
                !strcmp(_modifs[i]->Sugar_Name, Sugar_Name)) break;
       }
       if (i < _n_modifs) return;

       if (_n_modifs >= _alloc_modifs) {
            _alloc_modifs += 5;
            tmp = _modifs;
            _modifs = new _MODIFS*[_alloc_modifs];
            for (i = 0; i < _n_modifs; i++) _modifs[i] = tmp[i];
            for (i = _n_modifs; i < _alloc_modifs; i++) _modifs[i] = NULL;
            delete [] tmp;
       }
       _modifs[_n_modifs] = new _MODIFS;
       strcpy(_modifs[_n_modifs]->Name, Name);
       strcpy(_modifs[_n_modifs]->FstAtom, FstAtom);
       strcpy(_modifs[_n_modifs]->FstName, FstName);
       _modifs[_n_modifs]->FstIndex = FstIndex;
       _modifs[_n_modifs]->FstRes = FstRes;
       strcpy(_modifs[_n_modifs]->SndAtom, SndAtom);
       strcpy(_modifs[_n_modifs]->SndName, SndName);
       _modifs[_n_modifs]->SndIndex = SndIndex;
       _modifs[_n_modifs]->SndRes = SndRes;
       strcpy(_modifs[_n_modifs]->Replaced_Atom, Replaced_Atom);
       strcpy(_modifs[_n_modifs]->Sugar_Name, Sugar_Name);
       _n_modifs++;
}

typedef struct {
       CifString modRes;
       CifString stdRes;
       int count;
} _MODRES_COUNT;

void _MAXIT::Insert_a_Modres(const char *PDB_ID, const char *FstName, const int FstIndex,
            const int FstRes, const char *StdName, const char *details)
{
       int i;
       char Standard_Name[50];
       memset(Standard_Name, 0, 50);
       if (StdName && strcmp(StdName, "")) strcpy(Standard_Name, StdName);

       ReVarCifArray<_MODRES_COUNT> modres_count_array;
       _MODRES_COUNT modres_count;
       modres_count_array.Clear();

       ConnectFormat *drug = find_drug(FstName);
       if (drug && drug->parent_residue() && strcmp(drug->parent_residue(), "")) {
            int len = strlen(drug->parent_residue()) + 1;
            char *buff = new char[len];
            memset(buff, 0, len);
            strcpy(buff, drug->parent_residue());
            for (i = 0; i < len - 1; i++) {
                 if (buff[i] == ',' || buff[i] == ';') buff[i] = ' ';
            }
            ndb_clean_string(buff);
            char *buffer = new char[len];
            memset(buffer, 0, len);
            char *p = buff;
            while (p) {
                 p = get_word(p, buffer);
                 if (!strcmp(buffer, "")) continue;
                 for (i = 0; i < (int) modres_count_array.Length(); i++) {
                      if (modres_count_array[i].modRes == FstName &&
                          modres_count_array[i].stdRes == buffer) {
                           modres_count_array[i].count++;
                           break;
                      }
                 }
                 if (i == (int) modres_count_array.Length()) {
                      modres_count.modRes.Copy(FstName);
                      modres_count.stdRes.Copy(buffer);
                      modres_count.count = 1;
                      modres_count_array.Add(modres_count);
                 }
            }
            delete [] buff;
            delete [] buffer;
       } 

       if (!modres_count_array.Length()) {
            modres_count.modRes.Copy(FstName);
            modres_count.stdRes.Copy(Standard_Name);
            modres_count.count = 1;
            modres_count_array.Add(modres_count);
       }

       int count = 0;
       for (i = 0; i < _n_modres; i++) {
            if (!strcmp(_modres[i]->PDB_ID, PDB_ID) &&
                !strcmp(_modres[i]->FstName, FstName) &&
                _modres[i]->FstIndex == FstIndex &&
                _modres[i]->FstRes == FstRes && 
                !strcmp(_modres[i]->Standard_Name, Standard_Name)) {
                 count++;
            }
       }
       for (unsigned int j = 0; j < modres_count_array.Length(); j++) {
            if (modres_count_array[j].modRes == FstName &&
                modres_count_array[j].stdRes == Standard_Name) {
                 if (count >= modres_count_array[j].count) return;
            }
       }

       for (i = 0; i < _n_modres; i++) {
            if (!strcmp(_modres[i]->PDB_ID, PDB_ID) &&
                !strcmp(_modres[i]->FstName, FstName) &&
                _modres[i]->FstIndex == FstIndex &&
                _modres[i]->FstRes == FstRes && 
               ((!strcmp(_modres[i]->Standard_Name, "DA") && !strcmp(Standard_Name, "A")) ||
                (!strcmp(_modres[i]->Standard_Name, "DC") && !strcmp(Standard_Name, "C")) ||
                (!strcmp(_modres[i]->Standard_Name, "DG") && !strcmp(Standard_Name, "G")) ||
                (!strcmp(_modres[i]->Standard_Name, "DT") && !strcmp(Standard_Name, "T")) ||
                (!strcmp(_modres[i]->Standard_Name, "DU") && !strcmp(Standard_Name, "U")))) {
                 return;
            }
       }

       for (i = 0; i < _n_modres; i++) {
            if (!strcmp(_modres[i]->PDB_ID, PDB_ID) &&
                !strcmp(_modres[i]->FstName, FstName) &&
                _modres[i]->FstIndex == FstIndex &&
                _modres[i]->FstRes == FstRes && 
               ((!strcmp(_modres[i]->Standard_Name, "A") && !strcmp(Standard_Name, "DA")) ||
                (!strcmp(_modres[i]->Standard_Name, "C") && !strcmp(Standard_Name, "DC")) ||
                (!strcmp(_modres[i]->Standard_Name, "G") && !strcmp(Standard_Name, "DG")) ||
                (!strcmp(_modres[i]->Standard_Name, "T") && !strcmp(Standard_Name, "DT")) ||
                (!strcmp(_modres[i]->Standard_Name, "U") && !strcmp(Standard_Name, "DU")))) {
                 strcpy(_modres[i]->Standard_Name, Standard_Name);
                 return;
            }
       }

       if (_n_modres >= _alloc_modres) {
            _alloc_modres += 5;
            _MODRES **tmp = _modres;
            _modres = new _MODRES*[_alloc_modres];
            for (i = 0; i < _n_modres; i++) _modres[i] = tmp[i];
            for (i = _n_modres; i < _alloc_modres; i++) _modres[i] = NULL;
            delete [] tmp;
       }
       _modres[_n_modres] = new _MODRES;
       strcpy(_modres[_n_modres]->PDB_ID, PDB_ID);
       strcpy(_modres[_n_modres]->FstName, FstName);
       _modres[_n_modres]->FstIndex = FstIndex;
       _modres[_n_modres]->FstRes = FstRes;
       if (strcmp(Standard_Name, ""))
            strcpy(_modres[_n_modres]->Standard_Name, Standard_Name);
       else strcpy(_modres[_n_modres]->Standard_Name, "");
       int len = strlen(details) + 1;
       _modres[_n_modres]->details = new char[len];
       strcpy(_modres[_n_modres]->details, details);
       _n_modres++;
}

void _MAXIT::input_structural_features(const int format)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       _MOLECULE *mol = _molecules[0];

       int i = 0, initIndex = -1, initRes = -1, endIndex = -1, endRes = -1;
       int curIndex = -1, curRes = -1, prevIndex = -1, prevRes = -1;
       int FstIndex = -1, FstRes = -1, SndIndex = -1, SndRes = -1;
       int H_Index = -1, H_Res = -1, snd_resnum = 0;
       char snd_chainid[5], symmetry[10];
       _CHAIN *chain = NULL;
       _FIELD *init = NULL, *end = NULL;
       One_Card_Info_Format *pCard = _CardInfo[HELIX_TOKEN-1].Head();

       while (pCard) {
            chain = mol->GetFirstChain();
            while (chain) {
                 if ((!strcmp(chain->PDB_ChainID(), pCard->Field(4)) && format == NDB_FILE_FORMAT_PDB) ||
                     (!strcmp(chain->ChainID(), pCard->Field(4)) && format == NDB_FILE_FORMAT_NDB)) break;
                 chain = mol->GetNextChain();
            }
            for (i = 0; i < mol->Num_Chain(); i++) {
            }
            if (chain) {
                 init = chain->find_Residue(format, pCard->Field(3),
                                 atoi(pCard->Field(5)), pCard->Field(6));
                 end = chain->find_Residue(format, pCard->Field(7),
                                 atoi(pCard->Field(9)), pCard->Field(10));
                 if (init && end) {
                      chain->Insert_a_Helix(pCard->Field(2), init->ResIndex,
                            end->ResIndex, atoi(pCard->Field(11)), pCard->Field(12));
                      _exist_helix = 1;
                 }
            }
            pCard = pCard->Next();
       }
       _CardInfo[HELIX_TOKEN-1].clear_Cards();

       pCard = _CardInfo[TURN_TOKEN-1].Head();
       while (pCard) {
            chain = mol->GetFirstChain();
            while (chain) {
                 if (((!strcmp(chain->PDB_ChainID(), pCard->Field(4)) && format == NDB_FILE_FORMAT_PDB) ||
                      (!strcmp(chain->ChainID(), pCard->Field(4)) && format == NDB_FILE_FORMAT_NDB)) &&
                       chain->chain_type() == ATOMP_TOKEN) break;
                 chain = mol->GetNextChain();
            }
            if (chain) {
                 init = chain->find_Residue(format, pCard->Field(3),
                                  atoi(pCard->Field(5)), pCard->Field(6));
                 end  = chain->find_Residue(format, pCard->Field(7),
                                  atoi(pCard->Field(9)), pCard->Field(10));
                 if (init && end) {
                      chain->Insert_a_Turn(pCard->Field(2), init->ResIndex,
                            end->ResIndex, pCard->Field(11));
                 }
            }
            pCard = pCard->Next();
       }
       _CardInfo[TURN_TOKEN-1].clear_Cards();

       pCard = _CardInfo[SHEET_TOKEN-1].Head();
       while (pCard) {
            mol->find_residue(format, pCard->Field(15), pCard->Field(14),
                  atoi(pCard->Field(16)), pCard->Field(17), curIndex, curRes);
            mol->find_residue(format, pCard->Field(20), pCard->Field(19),
                  atoi(pCard->Field(21)), pCard->Field(22), prevIndex, prevRes);
            if (mol->find_residue(format, pCard->Field(5), pCard->Field(4),
                  atoi(pCard->Field(6)), pCard->Field(7), initIndex, initRes) &&
                mol->find_residue(format, pCard->Field(9), pCard->Field(8),
                  atoi(pCard->Field(10)), pCard->Field(11), endIndex, endRes)) {
                 Insert_a_Sheet(pCard->Field(2), atoi(pCard->Field(3)), initIndex,
                      initRes, endIndex, endRes, atoi(pCard->Field(12)),
                      pCard->Field(13), curIndex, curRes, pCard->Field(18),
                      prevIndex, prevRes);
                 _exist_sheet = 1;
            }
            pCard = pCard->Next();
       }
       _CardInfo[SHEET_TOKEN-1].clear_Cards();

       Input_SSBond(format);
       if (_n_ssbond > 0) _exist_ssbond = 1;
       Input_Link(format);

       pCard = _CardInfo[HYDBND_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_residue(format, pCard->Field(4), pCard->Field(3),
                   atoi(pCard->Field(5)), pCard->Field(6), FstIndex, FstRes) &&
                mol->find_residue(format, pCard->Field(15), pCard->Field(14),
                   atoi(pCard->Field(16)), pCard->Field(17), SndIndex, SndRes)) {
                 if (strcmp(pCard->Field(7), "")) {
                      if (!strcmp(pCard->Field(4), pCard->Field(9)) &&
                          !strcmp(pCard->Field(5), pCard->Field(10)) &&
                          !strcmp(pCard->Field(6), pCard->Field(11))) {
                           H_Index = FstIndex; H_Res = FstRes;
                      } else if (!strcmp(pCard->Field(4), pCard->Field(15)) &&
                          !strcmp(pCard->Field(5), pCard->Field(16)) &&
                          !strcmp(pCard->Field(6), pCard->Field(17))) {
                           H_Index = SndIndex; H_Res = SndRes;
                      }       
                 }
                 Insert_a_Hydbnd(pCard->Field(1), pCard->Field(2),  pCard->Field(3),
                     FstIndex, FstRes, pCard->Field(7), pCard->Field(8), pCard->Field(10),
                     H_Index, H_Res, pCard->Field(12), pCard->Field(13), pCard->Field(14),
                     SndIndex, SndRes, pCard->Field(18), pCard->Field(19));
            }
            pCard = pCard->Next();
       }
       _CardInfo[HYDBND_TOKEN-1].clear_Cards();

       pCard = _CardInfo[SLTBRG_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_residue(format, pCard->Field(4), pCard->Field(3),
                   atoi(pCard->Field(5)), pCard->Field(6), FstIndex, FstRes) &&
                mol->find_residue(format, pCard->Field(10), pCard->Field(9),
                   atoi(pCard->Field(11)), pCard->Field(12), SndIndex, SndRes)) { 
                 Insert_a_Sltbrg(pCard->Field(1), pCard->Field(2), pCard->Field(3),
                        FstIndex, FstRes, pCard->Field(7), pCard->Field(8),
                        pCard->Field(9), SndIndex, SndRes,
                        pCard->Field(13), pCard->Field(14));
            }
            pCard = pCard->Next();
       }
       _CardInfo[SLTBRG_TOKEN-1].clear_Cards();


       pCard = _CardInfo[MODRES_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(pCard->Field(2), "") && strcmp(pCard->Field(3), "") &&
                strcmp(pCard->Field(4), "") && mol->find_residue(format,
                       pCard->Field(3), pCard->Field(2), atoi(pCard->Field(4)),
                       pCard->Field(5), FstIndex, FstRes)) {
                 Insert_a_Modres(pCard->Field(1), pCard->Field(2), FstIndex, FstRes,
                       pCard->Field(6), pCard->Field(7));
            }
            pCard = pCard->Next();
       }
       _CardInfo[MODRES_TOKEN-1].clear_Cards();

       input_site_record(format);

       if (format != NDB_FILE_FORMAT_NDB) return;

       pCard = _CardInfo[MISMAT_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_ndb_residue(pCard->Field(2), pCard->Field(3),
                   atoi(pCard->Field(4)), "", FstIndex, FstRes) &&
                mol->find_ndb_residue(pCard->Field(5), pCard->Field(6),
                   atoi(pCard->Field(7)), "", SndIndex, SndRes)) {
                 Insert_a_Mismat(pCard->Field(3), FstIndex, FstRes,
                                 pCard->Field(6), SndIndex, SndRes);
            }
            pCard = pCard->Next();
       }
       _CardInfo[MISMAT_TOKEN-1].clear_Cards();

       pCard = _CardInfo[BSPAIR_TOKEN-1].Head();
       while (pCard) {
            get_bspair_second_chain(pCard->Field(6), pCard->Field(8), pCard->Field(10),
                    snd_chainid, snd_resnum, symmetry);
            if (mol->find_ndb_residue(pCard->Field(2), pCard->Field(3),
                   atoi(pCard->Field(4)), "", FstIndex, FstRes) &&
                mol->find_ndb_residue(snd_chainid, pCard->Field(7),
                   snd_resnum, "", SndIndex, SndRes)) {
                 Insert_a_Bspair(pCard->Field(5), pCard->Field(3), FstIndex, FstRes,
                       pCard->Field(9), pCard->Field(7), SndIndex, SndRes,
                       symmetry, pCard->Field(11));
            }
            pCard = pCard->Next();
       }
       _CardInfo[BSPAIR_TOKEN-1].clear_Cards();

       pCard = _CardInfo[MODIFB_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_ndb_residue(pCard->Field(4), pCard->Field(5),
                   atoi(pCard->Field(6)), "", FstIndex, FstRes) &&
                mol->find_ndb_residue(pCard->Field(8), pCard->Field(9),
                   atoi(pCard->Field(10)), "", SndIndex, SndRes)) {
                 Insert_a_Modifb(pCard->Field(2), pCard->Field(3), pCard->Field(5),
                     FstIndex, FstRes, pCard->Field(7), pCard->Field(9), SndIndex,
                     SndRes, pCard->Field(11));
            }
            pCard = pCard->Next();
       }
       _CardInfo[MODIFB_TOKEN-1].clear_Cards();

       pCard = _CardInfo[MODIFP_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_ndb_residue(pCard->Field(4), pCard->Field(5),
                   atoi(pCard->Field(6)), "", FstIndex, FstRes) &&
                mol->find_ndb_residue(pCard->Field(8), pCard->Field(9),
                   atoi(pCard->Field(10)), "", SndIndex, SndRes)) {
                 Insert_a_Modifp(pCard->Field(2), pCard->Field(3), pCard->Field(5),
                     FstIndex, FstRes, pCard->Field(7), pCard->Field(9), SndIndex,
                     SndRes, pCard->Field(11));
            }
            pCard = pCard->Next();
       }
       _CardInfo[MODIFP_TOKEN-1].clear_Cards();

       pCard = _CardInfo[MODIFS_TOKEN-1].Head();
       while (pCard) {
            if (mol->find_ndb_residue(pCard->Field(4), pCard->Field(5),
                   atoi(pCard->Field(6)), "", FstIndex, FstRes) &&
                mol->find_ndb_residue(pCard->Field(8), pCard->Field(9),
                   atoi(pCard->Field(10)), "", SndIndex, SndRes)) {
                 Insert_a_Modifs(pCard->Field(2), pCard->Field(3), pCard->Field(5),
                     FstIndex, FstRes, pCard->Field(7), pCard->Field(9), SndIndex,
                     SndRes, pCard->Field(11), pCard->Field(12));
            }
            pCard = pCard->Next();
       }
       _CardInfo[MODIFS_TOKEN-1].clear_Cards();
}

void _MAXIT::input_site_record(const int format)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       _MOLECULE *mol = _molecules[0];

       int FstIndex = -1, FstRes = -1;
       One_Card_Info_Format * pCard = _CardInfo[SITE_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(pCard->Field(4), "") && strcmp(pCard->Field(6), "") &&
                mol->find_residue(format, pCard->Field(5), pCard->Field(4),
                atoi(pCard->Field(6)), pCard->Field(7), FstIndex, FstRes)) {
                 Insert_a_Site(pCard->Field(2), FstIndex, FstRes, "1_555");
            }
            if (strcmp(pCard->Field(8), "") && strcmp(pCard->Field(10), "") &&
                mol->find_residue(format, pCard->Field(9), pCard->Field(8),
                atoi(pCard->Field(10)), pCard->Field(11), FstIndex, FstRes)) {
                 Insert_a_Site(pCard->Field(2), FstIndex, FstRes, "1_555");
            }
            if (strcmp(pCard->Field(12), "") && strcmp(pCard->Field(14), "") &&
                mol->find_residue(format, pCard->Field(13), pCard->Field(12),
                atoi(pCard->Field(14)), pCard->Field(15), FstIndex, FstRes)) {
                 Insert_a_Site(pCard->Field(2), FstIndex, FstRes, "1_555");
            }
            if (strcmp(pCard->Field(16), "") && strcmp(pCard->Field(18), "") &&
                mol->find_residue(format, pCard->Field(17), pCard->Field(16),
                atoi(pCard->Field(18)), pCard->Field(19), FstIndex, FstRes)) {
                 Insert_a_Site(pCard->Field(2), FstIndex, FstRes, "1_555");
            }
            pCard = pCard->Next();
       }
       _CardInfo[SITE_TOKEN-1].clear_Cards();
}

void _MAXIT::output_structural_features(const int format)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       _MOLECULE *mol = _molecules[0];
       char sheetID[4], SiteID[10];
       char SymOP[10];
       int serial_helix = 0, serial_turn = 0, serial_sheet = 1;
       int i = 0, j = 0, length = 0, serial_no = 0, siteId = 0, numRes;
       int *index = NULL, n_index = 0;
       One_Card_Info_Format *pCard = NULL;
       _HELIX *helix = NULL;
       _TURN *turn = NULL;
       _FIELD *Init = NULL, *End = NULL, *Cur = NULL, *Prev = NULL;
       _CHAIN *ch1 = NULL, *ch2 = NULL, *ch3 = NULL, *ch4 = NULL;
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            for (j = 0; j < chain->n_helix(); j++) {
                 pCard = _CardInfo[HELIX_TOKEN-1].
                         add_a_new_card(NdbToken[HELIX_TOKEN-1], 0);
                 serial_helix++;
                 pCard->AddElementToField(serial_helix, 1);
                 helix = chain->helix(j);
                 pCard->AddElementToField(helix->helixID, 2);
                 pCard->AddElementToField(helix->helixClass, 11);
                 pCard->AddElementToField(helix->comment, 12);
                 length = helix->endRes - helix->initRes + 1;
                 pCard->AddElementToField(length, 13);
                 Init = chain->SeqResIndex(helix->initRes);
                 End  = chain->SeqResIndex(helix->endRes);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(chain->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[3], 7);
                      pCard->AddElementToField(chain->PDB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[4], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(Init->Field[0], 3);
                      pCard->AddElementToField(chain->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[0], 7);
                      pCard->AddElementToField(chain->ChainID(), 8);
                      pCard->AddElementToField(End->Field[2], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 }
            }

            for (j = 0; j < chain->n_turn(); j++) {
                 pCard = _CardInfo[TURN_TOKEN-1].
                         add_a_new_card(NdbToken[TURN_TOKEN-1], 0);
                 serial_turn++;
                 pCard->AddElementToField(serial_turn, 1);
                 turn = chain->turn(j);
                 pCard->AddElementToField(turn->turnID, 2);
                 pCard->AddElementToField(turn->comment, 11);
                 Init = chain->SeqResIndex(turn->initRes);
                 End  = chain->SeqResIndex(turn->endRes);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(chain->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[3], 7);
                      pCard->AddElementToField(chain->PDB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[4], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(Init->Field[0], 3);
                      pCard->AddElementToField(chain->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[0], 7);
                      pCard->AddElementToField(chain->ChainID(), 8);
                      pCard->AddElementToField(End->Field[2], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 }
            }
            chain = mol->GetNextChain();
       }

       strcpy(sheetID, "xxx");
       for (i = 0; i < _n_sheet; i++) {
            ch1 = mol->GetIndexChain(_sheet[i]->initIndex);
            ch2 = mol->GetIndexChain(_sheet[i]->endIndex);
            ch3 = mol->GetIndexChain(_sheet[i]->curIndex);
            ch4 = mol->GetIndexChain(_sheet[i]->prevIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_sheet[i]->initRes);
            End  = ch2->SeqResIndex(_sheet[i]->endRes);
            if (!Init || !End) continue;

            Cur = NULL; Prev = NULL;
            if (ch3) Cur  = ch3->SeqResIndex(_sheet[i]->curRes);
            if (ch4) Prev = ch4->SeqResIndex(_sheet[i]->prevRes);

            pCard = _CardInfo[SHEET_TOKEN-1].
                    add_a_new_card(NdbToken[SHEET_TOKEN-1], 0);
            if (strcmp(_sheet[i]->sheetID, sheetID)) {
                 strcpy(sheetID, _sheet[i]->sheetID);
                 serial_sheet = 1;
            } else serial_sheet++;
            pCard->AddElementToField(serial_sheet, 1);
            pCard->AddElementToField(_sheet[i]->sheetID, 2);
            pCard->AddElementToField(_sheet[i]->numStrands, 3);
            pCard->AddElementToField(_sheet[i]->sense, 12);
            if (Cur && Prev) {
                 pCard->AddElementToField(_sheet[i]->curAtom, 13);
                 pCard->AddElementToField(_sheet[i]->prevAtom, 18);
            }
            if (format == NDB_FILE_FORMAT_PDB) {
                 pCard->AddElementToField(Init->Field[3], 4);
                 pCard->AddElementToField(ch1->PDB_ChainID(), 5);
                 pCard->AddElementToField(Init->Field[4], 6);
                 pCard->AddElementToField(Init->InsCode, 7);
                 pCard->AddElementToField(End->Field[3], 8);
                 pCard->AddElementToField(ch2->PDB_ChainID(), 9);
                 pCard->AddElementToField(End->Field[4], 10);
                 pCard->AddElementToField(End->InsCode, 11);
                 if (Cur && Prev) {
                      pCard->AddElementToField(Cur->Field[3], 14);
                      pCard->AddElementToField(ch3->PDB_ChainID(), 15);
                      pCard->AddElementToField(Cur->Field[4], 16);
                      pCard->AddElementToField(Cur->InsCode, 17);
                      pCard->AddElementToField(Prev->Field[3], 19);
                      pCard->AddElementToField(ch4->PDB_ChainID(), 20);
                      pCard->AddElementToField(Prev->Field[4], 21);
                      pCard->AddElementToField(Prev->InsCode, 22);
                 }
            } else if (format == NDB_FILE_FORMAT_NDB) {
                 pCard->AddElementToField(Init->Field[0], 4);
                 pCard->AddElementToField(ch1->ChainID(), 5);
                 pCard->AddElementToField(Init->Field[2], 6);
                 pCard->AddElementToField(Init->InsCode, 7);
                 pCard->AddElementToField(End->Field[0], 8);
                 pCard->AddElementToField(ch2->ChainID(), 9);
                 pCard->AddElementToField(End->Field[2], 10);
                 pCard->AddElementToField(End->InsCode, 11);
                 if (Cur && Prev) {
                      pCard->AddElementToField(Cur->Field[0], 14);
                      pCard->AddElementToField(ch3->ChainID(), 15);
                      pCard->AddElementToField(Cur->Field[2], 16);
                      pCard->AddElementToField(Cur->InsCode, 17);
                      pCard->AddElementToField(Prev->Field[0], 19);
                      pCard->AddElementToField(ch4->ChainID(), 20);
                      pCard->AddElementToField(Prev->Field[2], 21);
                      pCard->AddElementToField(Prev->InsCode, 22);
                 }
            }
       }

       if (_n_ssbond > 0) {
            int max_length = 0;
            for (i = 0; i < _n_ssbond; i++) {
                 ch1 = mol->GetIndexChain(_ssbond[i]->FstIndex);
                 if (!ch1) continue;
                 if (ch1->SeqLen() > max_length) max_length = ch1->SeqLen();
            }
            _IIndex _idx;
            BTreeObj<_IIndex> _Index;
            _Index.Clear();
            index = new int[_n_ssbond];
            for (i = 0; i < _n_ssbond; i++) {
                 ch1 = mol->GetIndexChain(_ssbond[i]->FstIndex);
                 if (!ch1) continue;
                 for (j = 0; j < n_index; j++) {
                      if (_ssbond[i]->FstIndex == _ssbond[index[j]]->FstIndex &&
                          _ssbond[i]->SndIndex == _ssbond[index[j]]->SndIndex &&
                          _ssbond[i]->FstRes == _ssbond[index[j]]->FstRes &&
                          _ssbond[i]->SndRes == _ssbond[index[j]]->SndRes) break;
                 }
                 if (j == n_index) {
                      index[j] = i;
                      n_index++;
                      int ii = _ssbond[i]->FstIndex * max_length + _ssbond[i]->FstRes;
                      _idx.set_value(ii);
                      _idx.set_pos(i);
                      _Index.Add(_idx);
                 }
            }
/*
            for (j = 0; j < n_index; j++) {
                 i = index[j];
*/
            _Index.GoFirst();
            int ret = 1;
            while (ret >= 0) {
                 _idx = _Index.GetCurrData();
                 i = _idx.pos();
                 ret = _Index.GoToNext();
                 ch1 = mol->GetIndexChain(_ssbond[i]->FstIndex);
                 ch2 = mol->GetIndexChain(_ssbond[i]->SndIndex);
                 if (!ch1 || !ch2) continue;
                 Init = ch1->SeqResIndex(_ssbond[i]->FstRes);
                 End  = ch2->SeqResIndex(_ssbond[i]->SndRes);
                 if (Init && End) {
                      pCard = _CardInfo[SSBOND_TOKEN-1].
                              add_a_new_card(NdbToken[SSBOND_TOKEN-1], 0);
                      serial_no++;
                      pCard->AddElementToField(serial_no, 1);
                      pCard->AddElementToField(_ssbond[i]->FstName, 2);
                      pCard->AddElementToField(_ssbond[i]->SndName, 6);
                      get_pdbformat_symmetry(SymOP, _ssbond[i]->SymOP_1);
                      pCard->AddElementToField(SymOP, 10);
                      get_pdbformat_symmetry(SymOP, _ssbond[i]->SymOP_2);
                      pCard->AddElementToField(SymOP, 11);
if (12 < pCard->NumField())
                      pCard->AddElementToField(_ssbond[i]->dist.Text(), 12);
                      if (format == NDB_FILE_FORMAT_PDB) {
                           pCard->AddElementToField(ch1->PDB_ChainID(), 3);
                           pCard->AddElementToField(Init->Field[4], 4);
                           pCard->AddElementToField(Init->InsCode, 5);
                           pCard->AddElementToField(ch2->PDB_ChainID(), 7);
                           pCard->AddElementToField(End->Field[4], 8);
                           pCard->AddElementToField(End->InsCode, 9);
                      } else if (format == NDB_FILE_FORMAT_NDB) {
                           pCard->AddElementToField(ch1->ChainID(), 3);
                           pCard->AddElementToField(Init->Field[2], 4);
                           pCard->AddElementToField(Init->InsCode, 5);
                           pCard->AddElementToField(ch2->ChainID(), 7);
                           pCard->AddElementToField(End->Field[2], 8);
                           pCard->AddElementToField(End->InsCode, 9);
                      }
                 }
            }
            delete [] index;
       }

       output_link(mol);
/*
       for (i = 0; i < _n_link; i++) {
            ch1 = mol->GetIndexChain(_link[i]->FstIndex);
            ch2 = mol->GetIndexChain(_link[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_link[i]->FstRes);
            End  = ch2->SeqResIndex(_link[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[LINK_TOKEN-1].
                         add_a_new_card(NdbToken[LINK_TOKEN-1], 0);
                 pCard->AddElementToField(_link[i]->FstAtom, 1);
                 pCard->AddElementToField(_link[i]->altLoc1, 2);
                 pCard->AddElementToField(_link[i]->FstName, 3);
                 pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                 pCard->AddElementToField(Init->Field[4], 5);
                 pCard->AddElementToField(Init->InsCode, 6);
                 pCard->AddElementToField(_link[i]->SndAtom, 7);
                 pCard->AddElementToField(_link[i]->altLoc2, 8);
                 pCard->AddElementToField(_link[i]->SndName, 9);
                 pCard->AddElementToField(ch2->PDB_ChainID(), 10);
                 pCard->AddElementToField(End->Field[4], 11);
                 pCard->AddElementToField(End->InsCode, 12);
                 get_pdbformat_symmetry(SymOP, _link[i]->SymOP_1);
                 pCard->AddElementToField(SymOP, 13);
                 get_pdbformat_symmetry(SymOP, _link[i]->SymOP_2);
                 pCard->AddElementToField(SymOP, 14);
if (15 < pCard->NumField())
                 pCard->AddElementToField(_link[i]->dist.Text(), 15);
            }
       }
*/
       for (i = 0; i < _n_hydbnd; i++) {
            ch1 = mol->GetIndexChain(_hydbnd[i]->FstIndex);
            ch2 = mol->GetIndexChain(_hydbnd[i]->SndIndex);
            ch3 = mol->GetIndexChain(_hydbnd[i]->H_Index);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_hydbnd[i]->FstRes);
            End  = ch2->SeqResIndex(_hydbnd[i]->SndRes);
            Cur  = NULL;
            if (ch3) Cur = ch3->SeqResIndex(_hydbnd[i]->H_Res);
            if (Init && End) {
                 pCard = _CardInfo[HYDBND_TOKEN-1].
                         add_a_new_card(NdbToken[HYDBND_TOKEN-1], 0);
                 pCard->AddElementToField(_hydbnd[i]->FstAtom, 1);
                 pCard->AddElementToField(_hydbnd[i]->altLoc1, 2);
                 pCard->AddElementToField(_hydbnd[i]->FstName, 3);
                 pCard->AddElementToField(_hydbnd[i]->H_Atom, 7);
                 pCard->AddElementToField(_hydbnd[i]->altLocH, 8);
                 pCard->AddElementToField(_hydbnd[i]->H_Name, 9);
                 pCard->AddElementToField(_hydbnd[i]->SndAtom, 12);
                 pCard->AddElementToField(_hydbnd[i]->altLoc2, 13);
                 pCard->AddElementToField(_hydbnd[i]->SndName, 14);
                 get_pdbformat_symmetry(SymOP, _hydbnd[i]->SymOP_1);
                 pCard->AddElementToField(SymOP, 18);
                 get_pdbformat_symmetry(SymOP, _hydbnd[i]->SymOP_2);
                 pCard->AddElementToField(SymOP, 19);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      if (Cur) {
                           pCard->AddElementToField(ch3->PDB_ChainID(), 10);
                           pCard->AddElementToField(Cur->Field[4], 11);
                           pCard->AddElementToField(Cur->InsCode, 12);
                      }
                      pCard->AddElementToField(ch2->PDB_ChainID(), 15);
                      pCard->AddElementToField(End->Field[4], 16);
                      pCard->AddElementToField(End->InsCode, 17);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      if (Cur) {
                           pCard->AddElementToField(ch3->ChainID(), 10);
                           pCard->AddElementToField(Cur->Field[2], 11);
                           pCard->AddElementToField(Cur->InsCode, 12);
                      }
                      pCard->AddElementToField(ch2->ChainID(), 15);
                      pCard->AddElementToField(End->Field[2], 16);
                      pCard->AddElementToField(End->InsCode, 17);
                 }
            }
       }

       for (i = 0; i < _n_sltbrg; i++) {
            ch1 = mol->GetIndexChain(_sltbrg[i]->FstIndex);
            ch2 = mol->GetIndexChain(_sltbrg[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_sltbrg[i]->FstRes);
            End  = ch2->SeqResIndex(_sltbrg[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[SLTBRG_TOKEN-1].
                         add_a_new_card(NdbToken[SLTBRG_TOKEN-1], 0);
                 pCard->AddElementToField(_sltbrg[i]->FstAtom, 1);
                 pCard->AddElementToField(_sltbrg[i]->altLoc1, 2);
                 pCard->AddElementToField(_sltbrg[i]->FstName, 3);
                 pCard->AddElementToField(_sltbrg[i]->SndAtom, 7);
                 pCard->AddElementToField(_sltbrg[i]->altLoc2, 8);
                 pCard->AddElementToField(_sltbrg[i]->SndName, 9);
                 get_pdbformat_symmetry(SymOP, _sltbrg[i]->SymOP_1);
                 pCard->AddElementToField(SymOP, 13);
                 get_pdbformat_symmetry(SymOP, _sltbrg[i]->SymOP_2);
                 pCard->AddElementToField(SymOP, 14);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 10);
                      pCard->AddElementToField(End->Field[4], 11);
                      pCard->AddElementToField(End->InsCode, 12);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(ch2->ChainID(), 10);
                      pCard->AddElementToField(End->Field[2], 11);
                      pCard->AddElementToField(End->InsCode, 12);
                 }
            }
       }

       serial_no = 0;
       strcpy(SiteID, "XXXXX");
       TblIndexObj site_index;
       CifString cs;
       for (i = 0; i < _n_site; i++) {
            siteId = 0;
            pCard = NULL;
            numRes = 0;
            site_index.Clear();
            for (j = 0; j < _site[i]->numRes; j++) {
                 ch1 = mol->GetIndexChain(_site[i]->initIndex[j]);
                 if (!ch1) continue;
                 Init = ch1->SeqResIndex(_site[i]->initRes[j]);
                 if (!Init) continue;
                 cs.Clear();
                 cs += _site[i]->initIndex[j];
                 cs += "_";
                 cs += _site[i]->initRes[j];
                 if (site_index.Seek(cs) >= 0) continue;
                 site_index.Add(cs);
                 // if (strcmp(_site[i]->SymOP[j], "1_555")) continue;

                 numRes++;
            }

            site_index.Clear();
            for (j = 0; j < _site[i]->numRes; j++) {
                 ch1 = mol->GetIndexChain(_site[i]->initIndex[j]);
                 if (!ch1) continue;
                 Init = ch1->SeqResIndex(_site[i]->initRes[j]);
                 if (!Init) continue;
                 cs.Clear();
                 cs += _site[i]->initIndex[j];
                 cs += "_";
                 cs += _site[i]->initRes[j];
                 if (site_index.Seek(cs) >= 0) continue;
                 site_index.Add(cs);
                 // if (strcmp(_site[i]->SymOP[j], "1_555")) continue;

                 if (strcmp(SiteID, _site[i]->SiteID)) {
                      serial_no = 0;
                      strcpy(SiteID, _site[i]->SiteID);
                 }

                 if (siteId == 0) {
                      pCard = _CardInfo[SITE_TOKEN-1].
                              add_a_new_card(NdbToken[SITE_TOKEN-1], 0);
                      serial_no++;
                      pCard->AddElementToField(serial_no, 1);
                      pCard->AddElementToField(_site[i]->SiteID, 2);
                      // pCard->AddElementToField(_site[i]->numRes, 3);
                      pCard->AddElementToField(numRes, 3);
                 }
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(Init->Field[3], siteId * 4 + 4);
                      pCard->AddElementToField(ch1->PDB_ChainID(), siteId * 4 + 5);
                      pCard->AddElementToField(Init->Field[4], siteId * 4 + 6);
                      pCard->AddElementToField(Init->InsCode, siteId * 4 + 7);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(Init->Field[0], siteId * 4 + 4);
                      pCard->AddElementToField(ch1->ChainID(), siteId * 4 + 5);
                      pCard->AddElementToField(Init->Field[2], siteId * 4 + 6);
                      pCard->AddElementToField(Init->InsCode, siteId * 4 + 7);
                 }
                 siteId++;
                 if (siteId == 4) siteId = 0;
            }
       }

       serial_no = 0;
       for (i = 0; i < _n_mismat; i++) {
            ch1 = mol->GetIndexChain(_mismat[i]->FstIndex);
            ch2 = mol->GetIndexChain(_mismat[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_mismat[i]->FstRes);
            End  = ch2->SeqResIndex(_mismat[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[MISMAT_TOKEN-1].
                         add_a_new_card(NdbToken[MISMAT_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_mismat[i]->FstName, 3);
                 pCard->AddElementToField(_mismat[i]->SndName, 6);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 2);
                      pCard->AddElementToField(Init->Field[4], 4);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 5);
                      pCard->AddElementToField(End->Field[4], 7);

                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 2);
                      pCard->AddElementToField(Init->Field[2], 4);
                      pCard->AddElementToField(ch2->ChainID(), 5);
                      pCard->AddElementToField(End->Field[2], 7);
                 }
            }
       }
       
       serial_no = 0;
       for (i = 0; i < _n_bspair; i++) {
            ch1 = mol->GetIndexChain(_bspair[i]->FstIndex);
            ch2 = mol->GetIndexChain(_bspair[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_bspair[i]->FstRes);
            End  = ch2->SeqResIndex(_bspair[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[BSPAIR_TOKEN-1].
                         add_a_new_card(NdbToken[BSPAIR_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_bspair[i]->FstName, 3);
                 pCard->AddElementToField(_bspair[i]->FstAtom, 5);
                 pCard->AddElementToField(_bspair[i]->SndName, 7);
                 pCard->AddElementToField(_bspair[i]->SndAtom, 9);
                 pCard->AddElementToField("A", 10);
                 pCard->AddElementToField(_bspair[i]->Type, 11);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 2);
                      pCard->AddElementToField(Init->Field[4], 4);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 6);
                      pCard->AddElementToField(End->Field[4], 8);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 2);
                      pCard->AddElementToField(Init->Field[2], 4);
                      pCard->AddElementToField(ch2->ChainID(), 6);
                      pCard->AddElementToField(End->Field[2], 8);
                 }

                 if (strcmp(_bspair[i]->Asym_or_Gen, "")) {
                      pCard->AddElementToField("S", 10);
                      for (j = 0; j < _n_stbiol; j++) {
                           if (!_stbiol[j]) continue;
                           if (_stbiol[j]->Asym_ID == ch2->ChainID() &&
                               !strcmp(_stbiol[j]->Symmetry, _bspair[i]->Asym_or_Gen)) {
                                pCard->AddElementToField(_stbiol[j]->NDB_Strand_ID.Text(), 6);
                                break;
                           }
                      }
                 }
            }
       }
       
       serial_no = 0;
       for (i = 0; i < _n_modifb; i++) {
            ch1 = mol->GetIndexChain(_modifb[i]->FstIndex);
            ch2 = mol->GetIndexChain(_modifb[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_modifb[i]->FstRes);
            End  = ch2->SeqResIndex(_modifb[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[MODIFB_TOKEN-1].
                         add_a_new_card(NdbToken[MODIFB_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_modifb[i]->Name, 2);
                 pCard->AddElementToField(_modifb[i]->FstAtom, 3);
                 pCard->AddElementToField(_modifb[i]->FstName, 5);
                 pCard->AddElementToField(_modifb[i]->SndAtom, 7);
                 pCard->AddElementToField(_modifb[i]->SndName, 9);
                 pCard->AddElementToField(_modifb[i]->Replaced_Atom, 11);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 6);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[4], 10);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 6);
                      pCard->AddElementToField(ch2->ChainID(), 8);
                      pCard->AddElementToField(End->Field[2], 10);
                 }
            }
       }
       
       serial_no = 0;
       for (i = 0; i < _n_modifp; i++) {
            ch1 = mol->GetIndexChain(_modifp[i]->FstIndex);
            ch2 = mol->GetIndexChain(_modifp[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_modifp[i]->FstRes);
            End  = ch2->SeqResIndex(_modifp[i]->SndRes);
            if (Init && End) {
                pCard = _CardInfo[MODIFP_TOKEN-1].
                        add_a_new_card(NdbToken[MODIFP_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_modifp[i]->Name, 2);
                 pCard->AddElementToField(_modifp[i]->FstAtom, 3);
                 pCard->AddElementToField(_modifp[i]->FstName, 5);
                 pCard->AddElementToField(_modifp[i]->SndAtom, 7);
                 pCard->AddElementToField(_modifp[i]->SndName, 9);
                 pCard->AddElementToField(_modifp[i]->Replaced_Atom, 11);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 6);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[4], 10);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 6);
                      pCard->AddElementToField(ch2->ChainID(), 8);
                      pCard->AddElementToField(End->Field[2], 10);
                 }
            }
       }
       
       serial_no = 0;
       for (i = 0; i < _n_modifs; i++) {
            ch1 = mol->GetIndexChain(_modifs[i]->FstIndex);
            ch2 = mol->GetIndexChain(_modifs[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_modifs[i]->FstRes);
            End  = ch2->SeqResIndex(_modifs[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[MODIFS_TOKEN-1].
                         add_a_new_card(NdbToken[MODIFS_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_modifs[i]->Name, 2);
                 pCard->AddElementToField(_modifs[i]->FstAtom, 3);
                 pCard->AddElementToField(_modifs[i]->FstName, 5);
                 pCard->AddElementToField(_modifs[i]->SndAtom, 7);
                 pCard->AddElementToField(_modifs[i]->SndName, 9);
                 pCard->AddElementToField(_modifs[i]->Replaced_Atom, 11);
                 pCard->AddElementToField(_modifs[i]->Sugar_Name, 12);
                 if (format == NDB_FILE_FORMAT_PDB) {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[4], 6);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[4], 10);
                 } else if (format == NDB_FILE_FORMAT_NDB) {
                      pCard->AddElementToField(ch1->ChainID(), 4);
                      pCard->AddElementToField(Init->Field[2], 6);
                      pCard->AddElementToField(ch2->ChainID(), 8);
                      pCard->AddElementToField(End->Field[2], 10);
                 }
            }
       }

       output_het_and_modres(format);
}

void _MAXIT::output_het_and_modres(const int format)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       One_Card_Info_Format *pCard = NULL;

       for (int i = 0; i < _n_het; i++) {
            _CHAIN *ch1 = _molecules[_rep_mol]->GetIndexChain(_het[i]->initIndex);
            if (!ch1) continue;
            _FIELD *Init = ch1->SeqResIndex(_het[i]->initRes);
            if (!Init) continue;

            pCard = _CardInfo[HET_TOKEN-1].
                    add_a_new_card(NdbToken[HET_TOKEN-1], 0);
            if (format == NDB_FILE_FORMAT_PDB) {
                 pCard->AddElementToField(Init->Field[3], 1);
                 pCard->AddElementToField(ch1->PDB_ChainID(), 2);
                 pCard->AddElementToField(Init->Field[4], 3);
                 pCard->AddElementToField(Init->InsCode, 4);
            } else if (format == NDB_FILE_FORMAT_NDB) {
                 pCard->AddElementToField(Init->Field[0], 1);
                 pCard->AddElementToField(ch1->ChainID(), 2);
                 pCard->AddElementToField(Init->Field[2], 3);
                 pCard->AddElementToField(Init->InsCode, 4);
            }
            pCard->AddElementToField(_het[i]->numAtoms, 5);
       }

       int serial_no = 0;
       for (int i = 0; i < _n_modres; i++) { 
            _CHAIN *ch1 = _molecules[0]->GetIndexChain(_modres[i]->FstIndex);
            if (!ch1) continue;
            _FIELD *Init = ch1->SeqResIndex(_modres[i]->FstRes);
            if (!Init) continue;

            pCard = _CardInfo[MODRES_TOKEN-1].
                    add_a_new_card(NdbToken[MODRES_TOKEN-1], 0);
            pCard->AddElementToField(_modres[i]->PDB_ID, 1);
            pCard->AddElementToField(_modres[i]->FstName, 2);
            pCard->AddElementToField(_modres[i]->Standard_Name, 6);
            pCard->AddElementToField(_modres[i]->details, 7);
            serial_no++;
            pCard->AddElementToField(serial_no, 8);
            if (format == NDB_FILE_FORMAT_PDB) {
                 pCard->AddElementToField(ch1->PDB_ChainID(), 3);
                 pCard->AddElementToField(Init->Field[4], 4);
                 pCard->AddElementToField(Init->InsCode, 5);
            } else if (format == NDB_FILE_FORMAT_NDB) {
                 pCard->AddElementToField(ch1->ChainID(), 3);
                 pCard->AddElementToField(Init->Field[2], 4);
                 pCard->AddElementToField(Init->InsCode, 5);
            }
       }
}

void _MAXIT::get_refn_card()
{
       int i, ref_id, refn_id, is_epublished;
       char buffer[1000];
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;
       ref_trans_format *ref = NULL;

       ref_id = find_jrnl_id("REF");
       refn_id = find_jrnl_id("REFN");
       
       qCard = _JrnlCardInfo[ref_id].Head();
       while (qCard) {
            if (!strcasecmp(qCard->Field(4), "Proc.Nat.Acad.Sci.USA")) {
                 qCard->AddElementToField("Proc.Natl.Acad.Sci.USA", 4);
            }
            qCard = qCard->Next();
       }

       qCard = _JrnlCardInfo[ref_id].Head();     
       while (qCard) {
            pCard = _JrnlCardInfo[refn_id].Head();
            while (pCard) {
                 if (!strcmp(pCard->Field(1), qCard->Field(1))) break;
                 pCard = pCard->Next();
            }
            if (pCard) {
                 if (strcmp(pCard->Field(3), "")) {
                      qCard = qCard->Next();
                      continue;
                 }
            } else {
                 pCard = _JrnlCardInfo[refn_id].add_a_new_card(JrnlToken[refn_id], 1);
                 pCard->AddElementToField(qCard->Field(1), 1);
            }

            is_epublished = 1;
            if (strcmp(qCard->Field(6), "") && strcmp(qCard->Field(7), ""))
                 is_epublished = 0;
            ref = find_ref(qCard->Field(4), is_epublished);
            if (ref) {
                 for (i = 3; i < 9; i++) {
                      // if (!strcmp(pCard->Field(i), "")) {
                           pCard->AddElementToField(ref->refn[i-3], i);
                      // }
                 }
                 if (strcasecmp(qCard->Field(4), ref->ref)) {
                      sprintf(buffer, "There are space differences between '%s' and '%s'",
                                 qCard->Field(4), ref->ref);
                      ndb_log_message(NDB_MSG_WARN, "%s\n", buffer);
                 }
            } else if (strcasecmp(qCard->Field(4), "To be Published")) {
                 ndb_log_message(NDB_MSG_WARN, "Journal %s did not exist in dictionary\n", qCard->Field(4));
            }
            qCard = qCard->Next();
       }
}

void _MAXIT::get_bspair_second_chain(const char *old_chnid, const char *old_num,
         const char *old_sym, char *new_chnid, int &new_num, char *new_sym)
{
       int i;
       strcpy(new_chnid, old_chnid);
       new_num = atoi(old_num);
       strcpy(new_sym, "");
       if (!strcmp(old_sym, "A")) return;

       for (i = 0; i < _n_stbiol; i++) {
            if (!_stbiol[i]) continue;
            if (_stbiol[i]->NDB_Strand_ID.Text() == old_chnid &&
                 strcmp(_stbiol[i]->Symmetry, "1_555")) break;
       }
       if (i < _n_stbiol) {
            strcpy(new_chnid, _stbiol[i]->Asym_ID.Text());
            strcpy(new_sym, _stbiol[i]->Symmetry);
            new_num = atoi(old_num) - _stbiol[i]->abrn + _stbiol[i]->bbrn;
       }
}

void _MAXIT::generate_ncs_strands()
{
       int card_id, successful;
       double mat[3][4];
       One_Card_Info_Format *pCard = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;

       card_id = find_card_id("NCSGEN");
       if (_CardInfo[card_id].Head() == NULL ||
           _CardInfo[MTRIX_TOKEN-1].Head() == NULL) return;

       pCard = _CardInfo[MTRIX_TOKEN-1].Head();
       while (pCard) {
            if (!strcmp(pCard->Field(6), "")) break;
            pCard = pCard->Next();
       }
       if (pCard == NULL) return;

       successful = 0;
       pCard = _CardInfo[card_id].Head();
       while (pCard) {
            if (strcmp(pCard->Field(2), "") && strcmp(pCard->Field(5), "")) {
                 if (get_matrix(mat, atoi(pCard->Field(2)))) {
                      if (_molecules[0]->symmetry_related_chain(atoi(pCard->Field(7)),
                            pCard->Field(3), pCard->Field(5), pCard->Field(6), mat))
                           successful = 1;
                 }
            }
            pCard = pCard->Next();
       }
       if (successful) {
            if (_stbiol) {
                 delete [] _stbiol;
                 _stbiol = NULL;
            }
            _n_stbiol = 0;
            match_stbiol_with_molecules();
       }
}

int _MAXIT::get_matrix(double m[3][4], const int serial_no)
{
       int i = 0, j;
       One_Card_Info_Format *pCard = _CardInfo[MTRIX_TOKEN-1].Head();
       while (pCard) {
            if (atoi(pCard->Field(1)) == serial_no) {
                 if (i >= 3) return 0;
                 for (j = 0; j < 4; j++)
                      m[i][j] = atof(pCard->Field(j+2));
                 i++;
            }
            pCard = pCard->Next();
       }
       if (i != 3) return 0;
       else        return 1;
}

void _MAXIT::delete_atoms(const char *filename)
{
       char command[400], line[200], chnid[10], resname[6], inscode[10], *p;
       int i, resnum;
       _RESIDUE *residue = NULL;

       FILE *fp = fopen(filename, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", filename);
            return;
       }
       command[0] = '\0';
       while (!feof(fp)) {
            if (fgets(line, 199, fp));
            if (feof(fp)) break;
            line[strlen(line)-1] = '\0';
            resnum = strlen(line);
            for (i = 0; i < resnum; i++) {
                 if (line[i] == '=' || line[i] == ',') line[i] = ' ';
            }
            ndb_set_string_to_upper(line);
            strcat(command, " ");
            strcat(command, line);
            ndb_clean_string(command);
            if (strstr(command, "RESIDUE") && strstr(command, "ATOMLIST")) {
                 sscanf(command, "%s %s %s %d %s", line, chnid, resname, &resnum,
                                         inscode);
                 if (!strcmp(chnid, "\"\"") || !strcmp(chnid, "''"))
                      strcpy(chnid, "");
                 if (!strcmp(inscode, "\"\"") || !strcmp(inscode, "''"))
                      strcpy(inscode, "");
                 for (i = 0; i < _num_mol; i++) {
                      residue = _molecules[i]->find_pdb_residue(chnid,
                                           resname, resnum, inscode);
                      p = strstr(command, "ATOMLIST");
                      if (strlen(p) > 8 && residue) {
                           p += 8;
                           while (p) {
                                p = get_word(p, line);
                                residue->delete_a_atom(line);
                           }
                           residue->reorder_atoms();
                      }
                 }
                 strcpy(command, "");
            }
       }
       fclose (fp);
}

void _MAXIT::insert_atoms(const char *filename)
{
       if (_num_mol == 0) return;

       FILE *fp = fopen(filename, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", filename);
            return;
       }
       char CardString[10], PrevSeqNo[10], FieldInfo[MxField][MxLen];
       int i, PrevTokenId, JrnlId, lineno, TokenId;
       _RESIDUE *residue = NULL, *exist = NULL;
       _ATOM *atom = NULL, *tmp = NULL;

       for (i = 0; i < MxField; i++) memset(FieldInfo[i], 0, MxLen);
       residue = new _RESIDUE;
       PrevTokenId = -1;
       JrnlId = -1;
       lineno = 0;
       TokenId = -1;
       strcpy(PrevSeqNo, "xxx");
       do {
            TokenId = get_card(fp, CardString, lineno);
            if (TokenId == -2) break;
            if (!is_atom_card(TokenId) && strcmp(CardString, "ATOM") &&
                 strcmp(CardString, "HETATM")) {
                 if (!strcmp(CardString, "END")) break;
                 end_line(fp); continue;
            }
            if (TokenId == -1) TokenId = ATOMN_TOKEN - 1;
            get_one_line_info(fp, TokenId, PrevTokenId, PrevSeqNo, JrnlId, -1,
                    CardString, FieldInfo, "", "", "", NDB_FILE_FORMAT_PDB, 0);
            residue->insert_a_atom(TokenId , FieldInfo, 1, -1, _error_message);
            for (i = 0; i < 23; i++) FieldInfo[i][0] = '\0';
       } while (!feof(fp));
       fclose (fp);

       exist = NULL;
       atom = residue->GetFirstAtom();
       while (atom) {
            exist = _molecules[0]->find_pdb_residue(atom->chnid(),
                          atom->restype(), atom->resnum(), atom->ins_code());
            if (exist) {
                 tmp = new _ATOM;
                 *tmp = *atom;
                 exist->insert_a_atom(tmp, -1, _error_message);
                 exist->reorder_atoms();
            }
            atom = residue->GetNextAtom();
       }
       delete residue;
}

void _MAXIT::delete_residues(const char *filename)
{
       FILE *fp = fopen(filename, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", filename);
            return;
       }

       char line[200], buffer[10], chnid[10], resname[6], inscode[10];
       int resnum = 0, i = 0;
       _CHAIN *chain = NULL;

       while (!feof(fp)) {
            if (fgets(line, 199, fp));
            if (feof(fp)) break;
            line[strlen(line)-1] = '\0';
            resnum = strlen(line);
            for (i = 0; i < resnum; i++) {
                 if (line[i] == '=' || line[i] == ',') line[i] = ' ';
            }
            ndb_set_string_to_upper(line);
            ndb_clean_string(line);
            if (strstr(line, "RESIDUE")) {
                 sscanf(line, "%s %s %s %d %s", buffer, chnid, resname, &resnum,
                                         inscode);
                 if (!strcmp(chnid, "\"\"") || !strcmp(chnid, "''"))
                      strcpy(chnid, "");
                 if (!strcmp(inscode, "\"\"") || !strcmp(inscode, "''"))
                      strcpy(inscode, "");
                 for (i = 0; i < _num_mol; i++) {
                      chain = _molecules[i]->GetFirstChain();
                      while (chain) {
                           if (!strcmp(chnid, chain->PDB_ChainID())) {
                                if (chain->delete_a_residue(resname, resnum,
                                      inscode)) break;
                           }
                           chain = _molecules[i]->GetNextChain();
                      }
                 }
            }
       }
       fclose (fp);
}

void _MAXIT::insert_residues(const char *filename)
{
       if (_num_mol == 0) return;

       FILE *fp = fopen(filename, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", filename);
            return;
       }
       char CardString[10], PrevSeqNo[10], FieldInfo[MxField][MxLen], line[200];
       char chnid[10], resname[6], inscode[10], ID[3], NAME[5], CODE[3];
       char ResidueNo[10], ChainId[10], ResidueName[10], InsCode[10];
       int  resnum, Num_Residue, Alloc_Residue, PrevTokenId, JrnlId;
       int  i, lineno, is_first, TokenId, pos, NUM, at_begin;
       _ATOM *atom = NULL;
       _RESIDUE **residue = NULL, **tmp = NULL;
       _CHAIN *chain = NULL;

       for (i = 0; i < MxField; i++) memset(FieldInfo[i], 0, MxLen);
       Num_Residue = 0;
       Alloc_Residue = 0;
       PrevTokenId = -1;
       JrnlId = -1;
       lineno = 0;
       is_first = 0;
       TokenId = -1;
       pos = 0; NUM = 0; at_begin = 0; resnum = 0;

       strcpy(chnid, "");
       strcpy(inscode, "");
       strcpy(inscode, "");
       strcpy(NAME, "");
       strcpy(CODE, "");

       strcpy(PrevSeqNo, "xxx");
       strcpy(ResidueNo, "xxx");
       strcpy(ChainId, "xxx");
       strcpy(ResidueName, "xxx");
       strcpy(InsCode, "xxx");

       do {
            TokenId = get_card(fp, CardString, lineno);
            if (TokenId == -2) break;
            if (!strcmp(CardString, "BEFORE") ||
                !strcmp(CardString, "AFTER")) {
                 if (Num_Residue) {
                      chain = _molecules[0]->GetFirstChain();
                      while (chain) {
                           if (!strcmp(chnid, chain->PDB_ChainID()) &&
                                chain->find_PDB_residue(resname, resnum,
                                inscode, pos)) {
                                for (i = 0; i < Num_Residue; i++) {
                                     if (!i) {
                                          strcpy(ID, chnid); strcpy(NAME, resname);
                                          NUM = resnum; strcpy(CODE, inscode);
                                          at_begin = is_first;
                                     }
                                     if (chain->find_PDB_residue(NAME, NUM, CODE, pos)) {
                                          if (at_begin)
                                               chain->insert_a_residue(residue[i], -1);
                                          else chain->insert_a_residue(residue[i], pos);
                                          atom = residue[i]->GetFirstAtom();
                                          strcpy(ID, atom->chnid());
                                          strcpy(NAME, atom->restype());
                                          NUM = atom->resnum();
                                          strcpy(CODE, atom->ins_code());
                                          at_begin = 0;
                                          residue[i] = NULL;
                                     }
                                }
                                break;
                           }
                           chain = _molecules[0]->GetNextChain();
                      }
                      for (i = 0; i < Num_Residue; i++) {
                           if (residue[i]) {
                                delete residue[i];
                                residue[i] = NULL;
                           }
                      }
                 }
                 Num_Residue = 0;

                 get_one_line(line, fp);
                 line[strlen(line)-1] = '\0';
                 resnum = strlen(line);
                 for (i = 0; i < resnum; i++) {
                      if (line[i] == '=' || line[i] == ',') line[i] = ' ';
                 }
                 ndb_set_string_to_upper(line);
                 ndb_clean_string(line);
                 sscanf(line, "%s %s %d %s", chnid, resname, &resnum, inscode);
                 if (!strcmp(chnid, "\"\"") || !strcmp(chnid, "''"))
                      strcpy(chnid, "");
                 if (!strcmp(inscode, "\"\"") || !strcmp(inscode, "''"))
                      strcpy(inscode, "");
                 if (!strcmp(CardString, "BEFORE")) is_first = 1;
                 if (!strcmp(CardString, "AFTER")) is_first = 0;
                 continue;
            }
            if (!is_atom_card(TokenId) && strcmp(CardString, "ATOM") &&
                 strcmp(CardString, "HETATM")) {
                 if (!strcmp(CardString, "END")) break;
                 end_line(fp); continue;
            }
            if (TokenId == -1) TokenId = ATOMN_TOKEN - 1;
            get_one_line_info(fp, TokenId, PrevTokenId, PrevSeqNo, JrnlId, -1,
                    CardString, FieldInfo, "", "", "", NDB_FILE_FORMAT_PDB, 0);

            if (strcmp(FieldInfo[5], ChainId) ||
                strcmp(FieldInfo[6], ResidueNo) ||
                strcmp(FieldInfo[4], ResidueName) ||
                strcmp(FieldInfo[7], InsCode)) {
                 if (Num_Residue >= Alloc_Residue) {
                      Alloc_Residue +=  10;
                      tmp = residue;
                      residue = new _RESIDUE*[Alloc_Residue];
                      for (i = 0; i < Num_Residue; i++)
                           residue[i] = tmp[i];
                      for (i = Num_Residue; i < Alloc_Residue; i++)
                           residue[i] = NULL;
                      delete [] tmp;
                 }
                 residue[Num_Residue] = new _RESIDUE;
                 residue[Num_Residue]->set_ResName(FieldInfo[4]);
                 residue[Num_Residue]->set_res_no(atoi(FieldInfo[6]));
                 if (strcmp(FieldInfo[15], ""))
                      residue[Num_Residue]->set_pdb_res_no(atoi(FieldInfo[15]));
                 else residue[Num_Residue]->set_pdb_res_no(atoi(FieldInfo[6]));
                 Num_Residue++;
            }
            residue[Num_Residue-1]->insert_a_atom(TokenId , FieldInfo, 1,
                                 -1, _error_message);

            strcpy(ChainId, FieldInfo[5]);
            strcpy(ResidueNo, FieldInfo[6]);
            strcpy(ResidueName, FieldInfo[4]);
            strcpy(InsCode, FieldInfo[7]);
            for (i = 0; i < 23; i++) FieldInfo[i][0] = '\0';
       } while (!feof(fp));
       fclose (fp);

       if (Num_Residue) {
            chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (!strcmp(chnid, chain->PDB_ChainID()) &&
                      chain->find_PDB_residue(resname, resnum,
                      inscode, pos)) {
                      for (i = 0; i < Num_Residue; i++) {
                           if (!i) {
                                strcpy(ID, chnid); strcpy(NAME, resname);
                                NUM = resnum; strcpy(CODE, inscode);
                                at_begin = is_first;
                           }
                           if (chain->find_PDB_residue(NAME, NUM, CODE, pos)) {
                                if (at_begin)
                                     chain->insert_a_residue(residue[i], -1);
                                else chain->insert_a_residue(residue[i], pos);
                                atom = residue[i]->GetFirstAtom();
                                strcpy(ID, atom->chnid());
                                strcpy(NAME, atom->restype());
                                NUM = atom->resnum();
                                strcpy(CODE, atom->ins_code());
                                at_begin = 0;
                                residue[i] = NULL;
                           }
                      }
                      break;
                 }
                 chain = _molecules[0]->GetNextChain();
            }
            for (i = 0; i < Num_Residue; i++) {
                 if (residue[i]) {
                      delete residue[i];
                      residue[i] = NULL;
                 }
            }
       }
       delete [] residue;
}

void _MAXIT::replace_residues(const char *filename)
{
       if (_num_mol == 0) return;

       FILE *fp = fopen(filename, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", filename);
            return;
       }
       char CardString[10], PrevSeqNo[10], FieldInfo[MxField][MxLen];
       char ResidueNo[10], ChainId[10], ResidueName[10], InsCode[10];
       int  i, PrevTokenId, JrnlId, lineno, TokenId;
       _ATOM *atom = NULL;
       _RESIDUE *residue = NULL, *exist = NULL;
       
       for (i = 0; i < MxField; i++) memset(FieldInfo[i], 0, MxLen);
       residue = new _RESIDUE;
       PrevTokenId = -1;
       JrnlId = -1;
       lineno = 0;
       TokenId = -1;
       strcpy(PrevSeqNo, "xxx");
       strcpy(ResidueNo, "xxx");
       strcpy(ChainId, "xxx");
       strcpy(ResidueName, "xxx");
       strcpy(InsCode, "xxx");
       do {
            TokenId = get_card(fp, CardString, lineno);
            if (TokenId == -2) break;
            if (!is_atom_card(TokenId) && strcmp(CardString, "ATOM") &&
                 strcmp(CardString, "HETATM")) {
                 if (!strcmp(CardString, "END")) break;
                 end_line(fp); continue;
            }
            if (TokenId == -1) TokenId = ATOMN_TOKEN - 1;
            get_one_line_info(fp, TokenId, PrevTokenId, PrevSeqNo, JrnlId, -1,
                    CardString, FieldInfo, "", "", "", NDB_FILE_FORMAT_PDB, 0);

            if (strcmp(FieldInfo[5], ChainId) ||
                strcmp(FieldInfo[6], ResidueNo) ||
                strcmp(FieldInfo[4], ResidueName) ||
                strcmp(FieldInfo[7], InsCode)) {
                 atom = residue->GetFirstAtom();
                 if (atom) {
                      residue->change_nomenclature(NDB_FILE_FORMAT_PDB, 1, 0, 0, _error_message);
                      residue->correction_atom_name(1, _error_message);
                      exist = _molecules[0]->find_pdb_residue(atom->chnid(),
                           atom->restype(), atom->resnum(), atom->ins_code());
                      if (exist) exist->replace_residue(residue);
                 }
                 residue->set_ResName(FieldInfo[4]);
                 residue->set_res_no(atoi(FieldInfo[6]));
                 if (strcmp(FieldInfo[15], ""))
                      residue->set_pdb_res_no(atoi(FieldInfo[15]));
                 else residue->set_pdb_res_no(atoi(FieldInfo[6]));
            }
            residue->insert_a_atom(TokenId , FieldInfo, 1, -1, _error_message);

            strcpy(ChainId, FieldInfo[5]);
            strcpy(ResidueNo, FieldInfo[6]);
            strcpy(ResidueName, FieldInfo[4]);
            strcpy(InsCode, FieldInfo[7]);
            for (i = 0; i < 23; i++) FieldInfo[i][0] = '\0';
       } while (!feof(fp));
       fclose (fp);

       atom = residue->GetFirstAtom();
       if (atom) {
            residue->change_nomenclature(NDB_FILE_FORMAT_PDB, 1, 0, 0, _error_message);
            residue->correction_atom_name(1, _error_message);
            exist = _molecules[0]->find_pdb_residue(atom->chnid(),
                    atom->restype(), atom->resnum(), atom->ins_code());
            if (exist) exist->replace_residue(residue);
       }

       delete residue;
}

void _MAXIT::set_refine_program(_MAXIT *obj)
{
       One_Card_Info_Format *qCard = obj->_CardInfo[REFMET_TOKEN-1].Head();
       if (qCard && strcmp(qCard->Field(3), "")) {
            One_Card_Info_Format *pCard = _CardInfo[REFMET_TOKEN-1].Head();
            if (!pCard) pCard = _CardInfo[REFMET_TOKEN-1].
                         add_a_new_card(NdbToken[REFMET_TOKEN-1], 0);
            pCard->AddElementToField(qCard->Field(3), 3);
       }
}

void _MAXIT::merge_remarks(_MAXIT *rmk)
{
       int i, j, card_id, temp_id;
       One_Card_Info_Format *pCard, *qCard;

       temp_id = find_card_id("TMLPTL"); 
       for (i = 0; i < NUM_TOKENS; i++) {
            card_id = find_card_id(merge_tokens[i]);
            if (card_id < 0) continue;
            qCard = rmk->_CardInfo[card_id].Head();
            if (!qCard) continue;

            pCard = _CardInfo[card_id].Head();
            while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);
            if (!strcmp(merge_tokens[i], "RFTPOS") ||
                !strcmp(merge_tokens[i], "RFMPOS") ||
                !strcmp(merge_tokens[i], "RFLPOS") ||
                !strcmp(merge_tokens[i], "RFTTHR") ||
                !strcmp(merge_tokens[i], "RFMTHR") ||
                !strcmp(merge_tokens[i], "RFLTHR")) {
                 if (temp_id >= 0) {
                      pCard = _CardInfo[temp_id].Head();
                      while (pCard) pCard = _CardInfo[temp_id].delete_a_card(pCard);
                 }
            }

            qCard = rmk->_CardInfo[card_id].Head();
          while (qCard) {
            pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);

            for (j = 1; j < pCard->NumField(); j++) {
                 if (strcmp(qCard->Field(j), ""))
                      pCard->AddElementToField(qCard->Field(j), j);
            }
            qCard = qCard->Next();
          }
       }
}

void _MAXIT::switch_residue_name_from_sequences()
{
       int i;
       _CHAIN *chain = NULL;
       for (i = 0; i < _num_mol; i++) {
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMP_TOKEN)
                      chain->correction_residues_name();
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

void _MAXIT::moving_waters(const int no_chain_id_residue_only)
{
       int i, prev_n_moving_waters = 0;
       _RESIDUE *residue = NULL;

       while (1) {
            calculate_all_contact(WATER_LIMIT, 1, 2);
            if (!_n_moving_waters) break;
            for (i = 0; i < _n_moving_waters; i++) {
                 residue = _molecules[0]->find_ndb_residue(_moving_waters[i]->a_chnid,
                           _moving_waters[i]->a_resnam, _moving_waters[i]->a_resnum, "");
                 if (!residue) continue;
                 if (no_chain_id_residue_only && 
                     strcmp(residue->GetFirstAtom()->pdb_chnid(), "")) continue;

                 residue->convert_between_orthogonal_and_fractional(_cell,
                            NDB_TRANS_ORTHO_TO_FRAC);
                 residue->symmetry_operations(_moving_waters[i]->transl[0],
                               _moving_waters[i]->transl[1], _moving_waters[i]->transl[2],
                               &(_cell->symops[_moving_waters[i]->sym]), 1);
                 residue->convert_between_orthogonal_and_fractional(_cell,
                            NDB_TRANS_FRAC_TO_ORTHO);
            }
            if (prev_n_moving_waters == _n_moving_waters) break;
            prev_n_moving_waters = _n_moving_waters;
       }

       assign_chain_id_to_non_polymer(1);

       if (_CifObj->IsTablePresent(_StructureId, "pdbx_validate_close_contact"))
            _CifObj->DeleteTable(_StructureId, "pdbx_validate_close_contact");

       if (_CifObj->IsTablePresent(_StructureId, "pdbx_validate_symm_contact"))
            _CifObj->DeleteTable(_StructureId, "pdbx_validate_symm_contact");

       calculate_all_contact(WATER_LIMIT, 0, 1);
       get_pdbx_validate_close_contact();
       get_pdbx_validate_symm_contact();
}

void _MAXIT::checking_moving_waters()
{
       COORD coord, sym_coord;
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       float xmax = -1000.0;
       float xmin = 1000.0; 
       float ymax = -1000.0;
       float ymin = 1000.0; 
       float zmax = -1000.0;
       float zmin = 1000.0; 
       while (chain) {
/*
            if (chain->chain_type() != HETAS_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }
*/
            _RESIDUE *residue = chain->GetFirstResidue();
            while (residue) {
                 if (strcmp(residue->GetFirstAtom()->pdb_chnid(), "")) {
                      coord = residue->GetFirstAtom()->orig();
                      if (coord.x > xmax) xmax = coord.x;
                      if (coord.x < xmin) xmin = coord.x;
                      if (coord.y > ymax) ymax = coord.y;
                      if (coord.y < ymin) ymin = coord.y;
                      if (coord.z > zmax) zmax = coord.z;
                      if (coord.z < zmin) zmin = coord.z;
                 }
                 residue = chain->GetNextResidue();
            }
            chain = _molecules[0]->GetNextChain();
       }
       xmin -= 5.0;
       xmax += 5.0;
       ymin -= 5.0;
       ymax += 5.0;
       zmin -= 5.0;
       zmax += 5.0;
/*
printf("xmin=%.3f xmax=%.3f ymin=%.3f ymax=%.3f zmin=%.3f zmax=%.3f\n",
xmin, xmax, ymin, ymax, zmin, zmax);
*/

       chain = _molecules[0]->GetFirstChain();
       while (chain) {
/*
            if (chain->chain_type() != HETAS_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }
*/
            _RESIDUE *residue = chain->GetFirstResidue();
            while (residue) {
                 if (strcmp(residue->GetFirstAtom()->pdb_chnid(), "")) {
                      residue = chain->GetNextResidue();
                      continue;
                 }
                 coord = residue->GetFirstAtom()->orig();
                 printf("%.3f %.3f %.3f\n", coord.x, coord.y, coord.z);
                 ndb_trans_coord_3by3(coord, _cell, NDB_TRANS_ORTHO_TO_FRAC);
                 for (int lx = -3; lx <= 3; lx++)
                 for (int ly = -3; ly <= 3; ly++)
                 for (int lz = -3; lz <= 3; lz++)
                 for (int i = 0; i < _cell->nop; i++) {
                      do_symmetry_operation(sym_coord, coord, &(_cell->symops[i]));
                      sym_coord.x += (float) lx;
                      sym_coord.y += (float) ly;
                      sym_coord.z += (float) lz;
                      ndb_trans_coord_3by3(sym_coord, _cell, NDB_TRANS_FRAC_TO_ORTHO);
/*
int x = sym_coord.x > xmin && sym_coord.x < xmax;
int y = sym_coord.y > ymin && sym_coord.y < ymax;
int z = sym_coord.z > zmin && sym_coord.z < zmax;
printf("x=%.3f (%.3f %.3f) %d\n", sym_coord.x, xmin, xmax,x);
printf("y=%.3f (%.3f %.3f) %d\n", sym_coord.y, ymin, ymax,y);
printf("z=%.3f (%.3f %.3f) %d\n", sym_coord.z, zmin, zmax,z);
printf("%d %d %d\n", x, y, z);
*/
                      if (sym_coord.x > xmin && sym_coord.x < xmax &&
                          sym_coord.y > ymin && sym_coord.y < ymax &&
                          sym_coord.z > zmin && sym_coord.z < zmax) {
                            printf("\t %d_%d%d%d %.3f %.3f %.3f\n",
                               i + 1, lx + 5, ly + 5, lz + 5,
                               sym_coord.x, sym_coord.y, sym_coord.z);
                      }
                 }
                 residue = chain->GetNextResidue();
            }
            chain = _molecules[0]->GetNextChain();
       }
}

int _MAXIT::moving_models(const int mol_id)
{
       int i, index, j;
       CifString cifstring;
       ISTable *Table = NULL;
       _MOLECULE *tmp = NULL;

       if (_num_mol < 2) {
            fprintf(stdout, "There are no multiple models.\n");
            return 1;
       }
       for (i = 0; i < _num_mol; i++) {
            if (_molecules[i]->Mol_ID() == mol_id) break;
       }
       if (i == _num_mol) {
            fprintf(stdout, "The best representative conformer ID %d did not exist in the file.\n", mol_id);
	    return 1;
       }
       tmp = _molecules[i];
       for (j = i; j > 0; j--) _molecules[j] = _molecules[j - 1];
       _molecules[0] = tmp;
       for (i = 0; i < _num_mol; i++) _molecules[i]->set_Mol_ID(i+1);

       Table = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_representative");
       if (is_empty_table(Table)) {
            _CifObj->DeleteTable(_StructureId, "rcsb_nmr_representative");
            Table = new ISTable("rcsb_nmr_representative");
            Table->AddColumn("entry_id");
            Table->AddColumn("conformer_id");
            Table->AddColumn("selection_criteria");
            Table->AddRow();
            cifstring.Copy(_StructureId);
            Table->UpdateCell(cifstring, 0, 0);
       }
       index = Table->GetColumnIndex("conformer_id");
       cifstring.Copy("1");
       Table->UpdateCell(cifstring, index, 0);
       _CifObj->WriteTable(Table, _StructureId, "rcsb_nmr_representative");

       return 0;
}

void _MAXIT::check_chnids()
{
       int i, num = 0;
       char chnid[5];
       _CHAIN *chain = NULL;
       One_Card_Info_Format *pCard = NULL;

       strcpy(chnid, "XXX");
       pCard = _CardInfo[SEQRES_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(chnid, pCard->Field(2))) {
                 strcpy(chnid, pCard->Field(2));
                 num++;
            }
            pCard = pCard->Next();
       }
       if (num != 1 || !strcmp(chnid, "")) return;

       for (i = 0; i < _num_mol; i++) {
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 if (!strcmp(chain->ChainID(), ""))
                      chain->set_ChainID(chnid);
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

void _MAXIT::change_coordinates_symm(const char *support_file)
{
       unsigned int i;
       int op = 0, lx = 0, ly = 0, lz = 0, flag_asym = 0;
       char asym_id[500], line[200], symmetry[20];
       FILE *fp = NULL;

       memset(asym_id, 0, 500);
       memset(symmetry, 0, 20);
       fp = fopen(support_file, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", support_file);
            return;
       }

       ReVarCifArray<ReVarCifArray<CifString> > lists;
       ReVarCifArray<CifString> list;
       lists.Clear();
       CifString cs;

       while (!feof(fp)) {
            if (fgets(line, 199, fp));
            if (feof(fp)) break;
            ndb_clean_string(line);
            if (!strcmp(line, "")) continue;

            if (!strncmp(line, "ASYMID:", 7)) {
                 strcpy(asym_id, &line[7]);
                 flag_asym = 1;
            } else if (!strncmp(line, "SYMMETRY:", 9)) {
                 strcpy(symmetry, &line[9]);
                 ndb_clean_string(asym_id);
                 ndb_clean_string(symmetry);
                 if (strcmp(asym_id, "") && strcmp(symmetry, "")) {
                      list.Clear();
                      cs.Copy(asym_id);
                      list.Add(cs);
                      cs.Copy(symmetry);
                      list.Add(cs);
                      lists.Add(list);
                      memset(asym_id, 0, 500);
                      memset(symmetry, 0, 20);
                 }
            } else if (!strncmp(line, "END", 3))
                 flag_asym = 0;
            else if (flag_asym) {
                 strcat(asym_id, " ");
                 strcat(asym_id, line);
            }
       }
       fclose (fp);

       if (lists.Length() == 0) return;

       for (unsigned int j = 0; j < lists.Length(); ++j) {
            strcpy(asym_id, lists[j][0].Text());
            strcpy(symmetry, lists[j][1].Text());
            op = atoi(symmetry) - 1;
            for (i = 0; i < strlen(symmetry); i++)
                 if (symmetry[i] == '_') break;
            lx = symmetry[i + 1] - '5';
            ly = symmetry[i + 2] - '5';
            lz = symmetry[i + 3] - '5';

            _molecules[0]->change_coordinates(asym_id, _cell, op, lx, ly, lz);
       }
}

void _MAXIT::change_coordinates_non_symm(const char *support_file)
{
       int flag_asym = 0, flag_matx = 0, i = 0;
       char asym_id[500], line[200], buf0[20], buf1[20], buf2[20], buf3[20];
       double m[3][4];
       FILE *fp = NULL;

       memset(asym_id, 0, 500);
       m[0][0] = m[1][1] = m[2][2] = 1.0;
       m[0][1] = m[0][2] = m[0][3] = 0.0;
       m[1][0] = m[1][2] = m[1][3] = 0.0;
       m[2][0] = m[2][1] = m[2][3] = 0.0;

       fp = fopen(support_file, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", support_file);
            return;
       }
       while (!feof(fp)) {
            if (fgets(line, 199, fp));
            if (feof(fp)) break;
            ndb_clean_string(line);
            if (!strcmp(line, "")) continue;

            if (!strncmp(line, "ASYMID:", 7)) {
                 strcpy(asym_id, &line[7]);
                 flag_asym = 1;
            } else if (!strncmp(line, "MATRIX:", 7))
                 flag_matx = 1;
            else if (!strncmp(line, "END", 3)) {
                 if (flag_asym) flag_asym = 0;
                 if (flag_matx) flag_matx = 0;
            } else if (flag_asym) {
                 strcat(asym_id, " ");
                 strcat(asym_id, line);
            } else if (flag_matx) {
                 if (i < 3) {
                      sscanf(line, "%s %s %s %s", buf0, buf1, buf2, buf3);
                      m[i][0] = atof(buf0);
                      m[i][1] = atof(buf1);
                      m[i][2] = atof(buf2);
                      m[i][3] = atof(buf3);
                      i++;
                 }
            }
       }
       fclose (fp);

       ndb_clean_string(asym_id);
       if (!strcmp(asym_id, "")) return;

       _molecules[0]->change_coordinates(asym_id, m);
}

void _MAXIT::remove_secondary_structures()
{
       int i;
       _exist_helix = _exist_sheet = _exist_ssbond = 0;

       _CardInfo[HELIX_TOKEN-1].clear_Cards();
       _CardInfo[SHEET_TOKEN-1].clear_Cards();
       _CardInfo[SSBOND_TOKEN-1].clear_Cards();

       Clear_Structural_Feature();
       _alloc_sheet = 0; _n_sheet = 0; _sheet = NULL;
       _alloc_ssbond = 0; _n_ssbond = 0; _ssbond = NULL;
       _alloc_hydbnd = 0; _n_hydbnd = 0; _hydbnd = NULL;
       _alloc_sltbrg = 0; _n_sltbrg = 0; _sltbrg = NULL;
       _alloc_site = 0; _n_site = 0; _site = NULL;
       _alloc_het = 0; _n_het = 0; _het = NULL;
       _alloc_modifb = 0; _n_modifb = 0; _modifb = NULL;
       _alloc_modifp = 0; _n_modifp = 0; _modifp = NULL;
       _alloc_modifs = 0; _n_modifs = 0; _modifs = NULL;
       _alloc_modres = 0; _n_modres = 0; _modres = NULL;
       if (_link) {
/*
            for (i = 0; i < _n_link; i++) {
                 if (_link[i]) {
                      if (_link[i]->details) delete [] _link[i]->details;
                      delete _link[i];
                 }
            }
*/
            delete [] _link;
       }
       if (_mismat) {
            for (i = 0; i < _n_mismat; i++)
                 if (_mismat[i]) delete _mismat[i];
            delete [] _mismat;
       }
       if (_bspair) {
            for (i = 0; i < _n_bspair; i++)
                 if (_bspair[i]) delete _bspair[i];
            delete [] _bspair;
       }
       _alloc_link = 0; _n_link = 0; _link = NULL;
       _alloc_mismat = 0; _n_mismat = 0; _mismat = NULL;
       _alloc_bspair = 0; _n_bspair = 0; _bspair = NULL;
       
       if (_num_mol == 0 || _molecules == NULL) return;
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN)
                 chain->remove_helix();
            chain = _molecules[0]->GetNextChain();
       }
       if (_CifObj) {
            _CifObj->DeleteTable(_StructureId, "refine_hist");
            _CifObj->DeleteTable(_StructureId, "ndb_nonstandard_list");
            // _CifObj->DeleteTable(_StructureId, "rcsb_missing_residue_list");
            remove_missing_residue();
       }
       delete_atom_count();

       One_Card_Info_Format *pCard = _CardInfo[MISRES_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[MISRES_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[LINK_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[LINK_TOKEN-1].delete_a_card(pCard);
}

void _MAXIT::remove_protein_secondary_structures()
{
       int i;
       _CHAIN *chain = NULL;

       _exist_helix = _exist_sheet = _exist_ssbond = 0;

       _CardInfo[HELIX_TOKEN-1].clear_Cards();
       _CardInfo[SHEET_TOKEN-1].clear_Cards();
       _CardInfo[SSBOND_TOKEN-1].clear_Cards();

       if (_sheet) {
            for (i = 0; i < _n_sheet; i++)
                 if (_sheet[i]) delete _sheet[i];
            delete [] _sheet;
       }
       if (_ssbond) {
            for (i = 0; i < _n_ssbond; i++)
                 if (_ssbond[i]) delete _ssbond[i];
            delete [] _ssbond;
       }
       _alloc_sheet = 0; _n_sheet = 0; _sheet = NULL;
       _alloc_ssbond = 0; _n_ssbond = 0; _ssbond = NULL;
       
       if (_num_mol == 0 || _molecules == NULL) return;

       chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN)
                 chain->remove_helix();
            chain = _molecules[0]->GetNextChain();
       }
}

void _MAXIT::convert_between_fractional_and_orthogonal(const int iflag)
{
       get_cell();

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "atom_site");
       if (is_empty_table(Table)) return;

       int x_index = Table->GetColumnIndex("Cartn_x");
       int y_index = Table->GetColumnIndex("Cartn_y");
       int z_index = Table->GetColumnIndex("Cartn_z");

       int x_esd_index = Table->GetColumnIndex("Cartn_x_esd");
       int y_esd_index = Table->GetColumnIndex("Cartn_y_esd");
       int z_esd_index = Table->GetColumnIndex("Cartn_z_esd");

       int rowNo = Table->GetNumRows();
 
       char value[15]; CifString cifstring;
       double x, y, z;
       for (int irow = 0; irow < rowNo; irow++) {
            if (x_index >= 0 && y_index >= 0 && z_index >= 0) {
                 get_value(cifstring, Table, x_index, irow);
                 if (cifstring != "") {
                      x = atof(cifstring.Text());
                      get_value(cifstring, Table, y_index, irow);
                      if (cifstring != "") {
                           y = atof(cifstring.Text());
                           get_value(cifstring, Table, z_index, irow);
                           if (cifstring != "") {
                                z = atof(cifstring.Text());
                                ndb_trans_coord_3by3(&x, &y, &z, _cell, iflag);
                                sprintf(value, "%.3f", x);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, x_index, irow);
                                sprintf(value, "%.3f", y);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, y_index, irow);
                                sprintf(value, "%.3f", z);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, z_index, irow);
                           }
                      }
                 }
            }
            if (x_esd_index >= 0 && y_esd_index >= 0 && z_esd_index >= 0) {
                 get_value(cifstring, Table, x_esd_index, irow);
                 if (cifstring != "") {
                      x = atof(cifstring.Text());
                      get_value(cifstring, Table, y_esd_index, irow);
                      if (cifstring != "") {
                           y = atof(cifstring.Text());
                           get_value(cifstring, Table, z_esd_index, irow);
                           if (cifstring != "") {
                                z = atof(cifstring.Text());
                                ndb_trans_coord_3by3(&x, &y, &z, _cell, iflag);
                                sprintf(value, "%.3f", x);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, x_esd_index, irow);
                                sprintf(value, "%.3f", y);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, y_esd_index, irow);
                                sprintf(value, "%.3f", z);
                                ndb_clean_string(value);
                                cifstring.Copy(value);
                                Table->UpdateCell(cifstring, z_esd_index, irow);
                           }
                      }
                 }
            }
       }
       _CifObj->WriteTable(Table, _StructureId, "atom_site");
}

#define MAX_LENGTH 60

void _MAXIT::get_one_letter_code_from_coordinates()
{
       int j = 0, len = 0, len1 = 0;
       char *code = NULL;
       _CHAIN *chain = NULL;
       CifString sequence;
       _FIELD *SeqRes = NULL;
       FILE *fp = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;

       fp = fopen(_output_filename, "w");
       chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMN_TOKEN ||
                chain->chain_type() == ATOMP_TOKEN) {
                 sequence.Copy("");
                 len = 0;
                 for (j = 0; j < chain->SeqLen(); j++) {
                      SeqRes = chain->SeqRes(j);
                      if (SeqRes->Deleted) continue;
                      code = get_oneletter_code(SeqRes->Field[0]);
                      len1 = strlen(code);
                      if (len1 == 1) {
                           if (len + len1 > MAX_LENGTH) {
                                sequence += "\n";
                                len = 0;
                           }
                           sequence += code;
                           len += len1;
                      } else if (len1 == 2 && code[0] == '+') {
                           if (len + 1 > MAX_LENGTH) {
                                sequence += "\n";
                                len = 0;
                           }
                           sequence += &code[1];
                           len += 1;
                      } else {
                           if (len + len1 + 2 > MAX_LENGTH) {
                                sequence += "\n";
                                len = 0;
                           }
                           sequence += "(";
                           sequence += code;
                           sequence += ")";
                           len += (len1 + 2);
                      }
                 }
                 fprintf(fp, "CHAIN: %s\n", chain->PDB_ChainID());
                 fprintf(fp, "%s\n", sequence.Text());
            }
            chain = _molecules[0]->GetNextChain();
       }
       fclose (fp);
}

void _MAXIT::check_cryst1_card()
{
       if (!_CardInfo[CRYST1_TOKEN-1].Head()) return;

       if (_cell) {
            if (_cell->symops) delete [] _cell->symops;
            delete _cell;
            _cell = NULL;
       }
       if (_CifObj) {
            _CifObj->DeleteTable(_StructureId, "symmetry");
            _CifObj->DeleteTable(_StructureId, "cell");
            _CifObj->DeleteTable(_StructureId, "atom_sites");
            One_Card_Info_Format *pCard = _CardInfo[SCALE_TOKEN-1].Head();
            while (pCard) pCard = _CardInfo[SCALE_TOKEN-1].delete_a_card(pCard);

            CifString string;
            ISTable *Table = new ISTable("cell");
            Table->AddColumn("entry_id");
            Table->AddColumn("length_a");
            Table->AddColumn("length_b");
            Table->AddColumn("length_c");
            Table->AddColumn("angle_alpha");
            Table->AddColumn("angle_beta");
            Table->AddColumn("angle_gamma");
            Table->AddColumn("Z_PDB");
            Table->AddColumn("ndb_unique_axis");
            Table->AddRow();
            string.Copy(_StructureId);
            Table->UpdateCell(string, 0, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(1));
            Table->UpdateCell(string, 1, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(2));
            Table->UpdateCell(string, 2, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(3));
            Table->UpdateCell(string, 3, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(4));
            Table->UpdateCell(string, 4, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(5));
            Table->UpdateCell(string, 5, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(6));
            Table->UpdateCell(string, 6, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(8));
            Table->UpdateCell(string, 7, 0);
            _CifObj->WriteTable(Table, _StructureId, "cell");

            Table = new ISTable("symmetry");
            Table->AddColumn("entry_id");
            Table->AddColumn("space_group_name_H-M");
            Table->AddColumn("ndb_full_space_group_name_H-M");
            Table->AddColumn("cell_setting");
            Table->AddColumn("Int_Tables_number");
            Table->AddRow();
            string.Copy(_StructureId);
            Table->UpdateCell(string, 0, 0);
            string.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(7));
            Table->UpdateCell(string, 1, 0);
            _CifObj->WriteTable(Table, _StructureId, "symmetry");
       }
       get_cell();
}

void _MAXIT::replace_cryst_info(ISTable *Table1, ISTable *Table2)
{
       if (_cell) {
            if (_cell->symops) delete [] _cell->symops;
            delete _cell;
            _cell = NULL;
       }
       if (_CifObj) {
            _CifObj->DeleteTable(_StructureId, "symmetry");
            _CifObj->DeleteTable(_StructureId, "cell");
            _CifObj->DeleteTable(_StructureId, "atom_sites");
            _CifObj->WriteTable(Table1, _StructureId, "cell");
            _CifObj->WriteTable(Table2, _StructureId, "symmetry");
       }
       get_cell();
}

void _MAXIT::replace_category(ISTable *Table, const char *category)
{
       int i, index, rowNo;
       CifString cifstring;

       if (_CifObj) {
            index = Table->GetColumnIndex("entry_id");
            if (index >= 0) {
                 rowNo = Table->GetNumRows();
                 cifstring.Copy(_StructureId);
                 for (i = 0; i < rowNo; i++) {
                      Table->UpdateCell(cifstring, index, i);
                 }
            }
            _CifObj->DeleteTable(_StructureId, category);
            _CifObj->WriteTable(Table, _StructureId, category);
       }
}

void _MAXIT::replace_coordinates(ISTable *Table, ISTable *Table1, const int assign_flag)
{
       if (_CifObj) {
            _CifObj->DeleteTable(_StructureId, "atom_site");
            _CifObj->WriteTable(Table, _StructureId, "atom_site");
            _CifObj->DeleteTable(_StructureId, "atom_site_anisotrop");
            if (Table1) _CifObj->WriteTable(Table1, _StructureId, "atom_site_anisotrop");
            put_coordinates_into_molecules(0, 0, 0, 0, 1);
            assign_chain_id_to_non_polymer(assign_flag);
       }
}

ISTable* _MAXIT::get_atom_site_category()
{
       ISTable *t = NULL;
       if (_CifObj) {
            t = _CifObj->GetTable(_StructureId, "atom_site");
            if (t) {
                 if (is_empty_table(t)) {
                      delete t; t = NULL;
                 }
            }
       }
       return t;
}

void _MAXIT::merge_original_author_nomenclature(ISTable *old_atom)
{
       if (is_empty_table(old_atom)) return;
       if (!_CifObj) return;

       ISTable *new_atom = _CifObj->GetTablePtr(_StructureId, "atom_site");
       if (is_empty_table(new_atom)) return;

       int old_rowNo = old_atom->GetNumRows();
       int old_idx1  = old_atom->GetColumnIndex("ndb_model");
       int old_idx2  = old_atom->GetColumnIndex("Cartn_x");
       int old_idx3  = old_atom->GetColumnIndex("Cartn_y");
       int old_idx4  = old_atom->GetColumnIndex("Cartn_z");
       int old_idx5  = old_atom->GetColumnIndex("B_iso_or_equiv");
       int old_idx6  = old_atom->GetColumnIndex("ndb_auth_seq_id");
       int old_idx7  = old_atom->GetColumnIndex("ndb_auth_comp_id");
       int old_idx8  = old_atom->GetColumnIndex("ndb_auth_asym_id");
       int old_idx9  = old_atom->GetColumnIndex("ndb_auth_atom_name");
       if (old_idx1 < 0 || old_idx2 < 0 || old_idx3 < 0 || old_idx4 < 0 ||
           old_idx5 < 0 || old_idx6 < 0 || old_idx7 < 0 || old_idx8 < 0 ||
           old_idx9 < 0) return;

       std::map<std::string, std::vector<std::string> > old_mapping;
       old_mapping.clear();
       std::vector<std::string> data;
       std::string scs;
       CifString cs;
       for (int i = 0; i < old_rowNo; ++i) {
            get_value(cs, old_atom, old_idx1, i);
            scs = cs.Text();
            get_value(cs, old_atom, old_idx2, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, old_atom, old_idx3, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, old_atom, old_idx4, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, old_atom, old_idx5, i);
            scs += "_"; scs += cs.Text();

            data.clear();
            get_value(cs, old_atom, old_idx6, i);
            data.push_back(cs.Text());
            get_value(cs, old_atom, old_idx7, i);
            data.push_back(cs.Text());
            get_value(cs, old_atom, old_idx8, i);
            data.push_back(cs.Text());
            get_value(cs, old_atom, old_idx9, i);
            data.push_back(cs.Text());

            old_mapping.insert(std::make_pair(scs, data));
       }

       if (old_mapping.empty()) return;

       int new_rowNo = new_atom->GetNumRows();
       int new_idx1  = new_atom->GetColumnIndex("ndb_model");
       int new_idx2  = new_atom->GetColumnIndex("Cartn_x");
       int new_idx3  = new_atom->GetColumnIndex("Cartn_y");
       int new_idx4  = new_atom->GetColumnIndex("Cartn_z");
       int new_idx5  = new_atom->GetColumnIndex("B_iso_or_equiv");
       int new_idx6  = new_atom->GetColumnIndex("ndb_auth_seq_id");
       int new_idx7  = new_atom->GetColumnIndex("ndb_auth_comp_id");
       int new_idx8  = new_atom->GetColumnIndex("ndb_auth_asym_id");
       int new_idx9  = new_atom->GetColumnIndex("ndb_auth_atom_name");
       if (new_idx1 < 0 || new_idx2 < 0 || new_idx3 < 0 || new_idx4 < 0 ||
           new_idx5 < 0 || new_idx6 < 0 || new_idx7 < 0 || new_idx8 < 0 ||
           new_idx9 < 0) return;

       std::set<std::string> new_coord_set;
       new_coord_set.clear();
       for (int i = 0; i < new_rowNo; ++i) {
            get_value(cs, new_atom, new_idx1, i);
            scs = cs.Text();
            get_value(cs, new_atom, new_idx2, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx3, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx4, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx5, i);
            scs += "_"; scs += cs.Text();

            new_coord_set.insert(scs);
       }

       if (new_coord_set.empty()) return;

       int count = 0;
       std::map<std::string, std::vector<std::string> >::iterator mpos;
       for (mpos = old_mapping.begin(); mpos != old_mapping.end(); ++mpos) {
            if (new_coord_set.find(mpos->first) != new_coord_set.end()) count++;
       }
       if ((float) count / (float) old_mapping.size() < 0.95) return;

       count = 0;
       std::set<std::string>::iterator spos;
       for (spos = new_coord_set.begin(); spos != new_coord_set.end(); ++spos) {
            if (old_mapping.find(*spos) != old_mapping.end()) count++;
       }
       if ((float) count / (float) new_coord_set.size() < 0.95) return;

       for (int i = 0; i < new_rowNo; ++i) {
            get_value(cs, new_atom, new_idx1, i);
            scs = cs.Text();
            get_value(cs, new_atom, new_idx2, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx3, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx4, i);
            scs += "_"; scs += cs.Text();
            get_value(cs, new_atom, new_idx5, i);
            scs += "_"; scs += cs.Text();

            mpos = old_mapping.find(scs);
            if (mpos == old_mapping.end()) continue;

            cs.Copy(mpos->second[0].c_str());
            new_atom->UpdateCell(cs, new_idx6, i);
            cs.Copy(mpos->second[1].c_str());
            new_atom->UpdateCell(cs, new_idx7, i);
            cs.Copy(mpos->second[2].c_str());
            new_atom->UpdateCell(cs, new_idx8, i);
            cs.Copy(mpos->second[3].c_str());
            new_atom->UpdateCell(cs, new_idx9, i);
       }
       _CifObj->WriteTable(new_atom, _StructureId, "atom_site");
}

void _MAXIT::get_biol_unit_information()
{
       int i, j, k;
       _CHAIN *chain = NULL;
       _STBIOL **stbiol = NULL;

       if (!_n_stbiol || !_stbiol) return;

       _biol_unit_info = new _BIOL_UNIT_INFO*[_n_stbiol];
       for (i = 0; i < _n_stbiol; i++) _biol_unit_info[i] = NULL;

       _n_biol_unit_info = 0;
       for (i = 0; i < _n_stbiol; i++) {
            if (!strcmp(_stbiol[i]->Symmetry, "1_555")) {
                 chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) {
                      if (_stbiol[i]->Asym_ID == chain->ChainID()) break;
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
                 if (!chain || chain->empty_chain()) {
                      ndb_log_message(NDB_MSG_ERR, "Can not find asymmetric chain %s\n",
                            _stbiol[i]->Asym_ID.Text());
                      continue;
                 }
                 j = _n_biol_unit_info;
                 _n_biol_unit_info++;
                 _biol_unit_info[j] = new _BIOL_UNIT_INFO;
                 _biol_unit_info[j]->n_stbiol = 1;
                 _biol_unit_info[j]->stbiol = new _STBIOL*[_biol_unit_info[j]->n_stbiol];
                 _biol_unit_info[j]->stbiol[0] = new _STBIOL;
                 *(_biol_unit_info[j]->stbiol[0]) = *(_stbiol[i]);
                 _biol_unit_info[j]->Asym_ID.Copy(_stbiol[i]->Asym_ID);
                 strcpy(_biol_unit_info[j]->PDB_Strand_ID, chain->PDB_ChainID());
                 _biol_unit_info[j]->chain_type = chain->chain_type();
                 strcpy(_biol_unit_info[j]->First_Residue_Name, chain->GetFirstResidue()->ResName());
                 _biol_unit_info[j]->First_Residue_Number = chain->GetFirstResidue()->pdb_res_no();
            } else {
                 for (j = 0; j < _n_biol_unit_info; j++) {
                      if (_biol_unit_info[j]->Asym_ID == _stbiol[i]->Asym_ID) break;
                 }
                 if (j == _n_biol_unit_info) {
                      ndb_log_message(NDB_MSG_ERR, "Can not find asymmetric chain for %s (%s)\n",
                               _stbiol[i]->Asym_ID.Text(), _stbiol[i]->Symmetry);
                      continue;
                 }
                 stbiol = _biol_unit_info[j]->stbiol;
                 _biol_unit_info[j]->n_stbiol++;
                 _biol_unit_info[j]->stbiol = new _STBIOL*[_biol_unit_info[j]->n_stbiol];
                 for (k = 0; k < _biol_unit_info[j]->n_stbiol - 1; k++)
                      _biol_unit_info[j]->stbiol[k] = stbiol[k];
                 delete [] stbiol;
                 _biol_unit_info[j]->stbiol[k] = new _STBIOL;
                 *(_biol_unit_info[j]->stbiol[k]) = *(_stbiol[i]);
            }
       }
       if (!_n_biol_unit_info) {
            delete [] _biol_unit_info;
            _biol_unit_info = NULL;
       }
}

void _MAXIT::put_biol_unit_information()
{
       int i, j, k, l, n_stbiol;
       _CHAIN *chain = NULL;
       _STBIOL **stbiol = NULL;

       if (!_biol_unit_info || !_n_biol_unit_info) return;

       if (_stbiol) {
            for (i = 0; i < _n_stbiol; i++) if (_stbiol[i]) delete _stbiol[i];
            delete [] _stbiol;
       }
       _stbiol = NULL; _n_stbiol = 0;

       match_stbiol_with_molecules();

       n_stbiol = _n_stbiol;
       for (i = 0; i < n_stbiol; i++) {
            if (!strcmp(_stbiol[i]->Symmetry, "1_555")) {
                 chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) {
                      if (_stbiol[i]->Asym_ID == chain->ChainID()) {
                           for (j = 0; j < _n_biol_unit_info; j++) {
                                if (!strcmp(_biol_unit_info[j]->PDB_Strand_ID, chain->PDB_ChainID()) &&
                                    !strcmp(_biol_unit_info[j]->First_Residue_Name, chain->GetFirstResidue()->ResName()) &&
                                    _biol_unit_info[j]->First_Residue_Number == chain->GetFirstResidue()->pdb_res_no() && 
                                    _biol_unit_info[j]->chain_type == chain->chain_type()) {
                                     for (k = 0; k < _biol_unit_info[j]->n_stbiol; k++) {
                                          if (!strcmp(_biol_unit_info[j]->stbiol[k]->Symmetry, "1_555")) {
                                               _stbiol[i]->Biol_ID = _biol_unit_info[j]->stbiol[k]->Biol_ID;
                                               _stbiol[i]->order = _biol_unit_info[j]->stbiol[k]->order;
                                               strcpy(_stbiol[i]->PDB_Strand_ID_Flag,
                                                       _biol_unit_info[j]->stbiol[k]->PDB_Strand_ID_Flag);
                                          } else {
                                               stbiol = _stbiol;
                                               _stbiol = new _STBIOL*[_n_stbiol + 1];
                                               for (l = 0; l < _n_stbiol; l++) _stbiol[l] = stbiol[l];
                                               delete [] stbiol;
                                               _n_stbiol++;
                                               _stbiol[l] = new _STBIOL;
                                               _stbiol[l]->Biol_ID = _biol_unit_info[j]->stbiol[k]->Biol_ID;
                                               _stbiol[l]->order = _biol_unit_info[j]->stbiol[k]->order;
                                               _stbiol[l]->Asym_ID.Copy(_stbiol[i]->Asym_ID);
                                               _stbiol[l]->NDB_Strand_ID.Copy(_stbiol[i]->NDB_Strand_ID);
                                               strcpy(_stbiol[l]->PDB_Strand_ID,
                                                      _biol_unit_info[j]->stbiol[k]->PDB_Strand_ID);
                                               strcpy(_stbiol[l]->PDB_Strand_ID_Flag,
                                                      _biol_unit_info[j]->stbiol[k]->PDB_Strand_ID_Flag);
                                               strcpy(_stbiol[l]->Symmetry,
                                                      _biol_unit_info[j]->stbiol[k]->Symmetry);
                                          }
                                     }
                                     break;
                                }
                           }
                           break;
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
            }
       }
}

void _MAXIT::processing_struct_biol_gen(const char *support_file)
{
       int symmetry_flag = 0, i, j, k, l, found_parenthesis, rowNo, *label = NULL;
       int biol_index, asym_index, symmetry_index, irow, row, colNo;
       char *buffer = NULL, *p = NULL, *q = NULL, buf[10];
       FILE *fp = NULL;
       _ASYM_SYMMETRY _asym_symm; 
       _ASYM_SYMM_PAIR _asym_symm_pair;
       _BIOL_ASYM_SYMMETRY _biol_asym_symmetry;
       CifString line, asym_ids, cifstring;
       ISTable *Table = NULL;
       ReVarCifArray<_BIOL_ASYM_SYMMETRY> _biol_array;
       ReVarCifArray<CifString> symmetry_array;
       ReVarCifArray<_STRUC_BIOL> _struct_biol_array;
       _STRUC_BIOL _struct_biol;
       
       fp = fopen(support_file, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", support_file);
            fprintf(stdout, "Can not read file %s!\n", support_file);
            exit (1);
       }

       _asym_symm_pair.asym_ids.Clear();
       _asym_symm_pair.symmetries.Clear();
       _biol_array.Clear();
       _biol_asym_symmetry.biol_id = 0;
       _biol_asym_symmetry.asym_symm_array.Clear();
       while (!feof(fp)) {
            get_line_from_file(fp, line);
            if (feof(fp)) break;
            line.ToUpper();
            if ((p = strstr(line.Text(), "BIOL_ID:"))) {
                 _biol_asym_symmetry.biol_id = atoi(p + 8);
                 _biol_asym_symmetry.asym_symm_array.Clear();
            } else if (strstr(line.Text(), "END_BIOL")) {
                 _biol_array.Add(_biol_asym_symmetry);
            } else if (strstr(line.Text(), "ASYM_IDS:")) {
                 symmetry_flag = 0;
            } else if (strstr(line.Text(), "WITH_SYMM_OPER:")) {
                 symmetry_flag = 1;
            } else if (strstr(line.Text(), "END_ASYM")) {
                 _biol_asym_symmetry.asym_symm_array.Add(_asym_symm_pair);
                 _asym_symm_pair.asym_ids.Clear();
                 _asym_symm_pair.symmetries.Clear();
            } else {
                 if (symmetry_flag)
                      _asym_symm_pair.symmetries += line;
                 else _asym_symm_pair.asym_ids += line;
            }
       }
       fclose (fp);

       _struct_biol_array.Clear();
       for (i = 0; i < (int) _biol_array.Length(); i++) {
            _struct_biol.biol_id = _biol_array[i].biol_id;
            _struct_biol.asym_array.Clear();
            for (j = 0; j < (int) _biol_array[i].asym_symm_array.Length(); j++) {
                 symmetry_array.Clear();
                 asym_ids = _biol_array[i].asym_symm_array[j].symmetries;
                 if (asym_ids != "") {
                      buffer = new char[strlen(asym_ids.Text())+1];
                      memset(buffer, 0, strlen(asym_ids.Text())+1);
                      strcpy(buffer, asym_ids.Text());
                      ndb_clean_string(buffer);
                      if (buffer[strlen(buffer)-1] == ',') {
                           buffer[strlen(buffer)-1] = '\0';
                      }
                      q = strtok(buffer, ",");
                      while (q) {
                           ndb_clean_string(q);
                           cifstring.Copy(q);
                           symmetry_array.Add(cifstring);
                           q = strtok(NULL, ",");
                      }
                      delete [] buffer;
                 }
                 asym_ids = _biol_array[i].asym_symm_array[j].asym_ids;
                 if (asym_ids != "") {
                      buffer = new char[strlen(asym_ids.Text())+1];
                      memset(buffer, 0, strlen(asym_ids.Text())+1);
                      strcpy(buffer, asym_ids.Text());
                      ndb_clean_string(buffer);
                      if (buffer[strlen(buffer)-1] == ',') {
                           buffer[strlen(buffer)-1] = '\0';
                      }
                      q = strtok(buffer, ",");
                      while (q) {
                           ndb_clean_string(q);
                           _asym_symm.asym_id.Clear();
                           _asym_symm.symmetry.Clear();
                           found_parenthesis = 0;
                           cifstring.Clear();
                           for (k = 0; k < (int) strlen(q); k++) {
                                if (q[k] == ' ') {
                                     if (cifstring != "") {
                                          if (!found_parenthesis)
                                               _asym_symm.asym_id = cifstring;
                                          else _asym_symm.symmetry.Add(cifstring);
                                     }
                                     cifstring.Clear();
                                     continue;
                                } else if (q[k] == '(') {
                                     if (cifstring != "") _asym_symm.asym_id = cifstring;
                                     found_parenthesis = 1;
                                     cifstring.Clear();
                                     continue;
                                } else if (q[k] == ')') {
                                     _asym_symm.symmetry.Add(cifstring);
                                     cifstring.Clear();
                                     break;
                                } else cifstring += q[k];
                           }
                           if (cifstring != "") {
                                if (!found_parenthesis)
                                     _asym_symm.asym_id = cifstring;
                                else _asym_symm.symmetry.Add(cifstring);
                           }
                           for (k = 0; k < (int) symmetry_array.Length(); k++) {
                                for (l = 0; l < (int) _asym_symm.symmetry.Length(); l++) {
                                     if (symmetry_array[k] == _asym_symm.symmetry[l]) break;
                                }
                                if (l == (int) _asym_symm.symmetry.Length()) {
                                     _asym_symm.symmetry.Add(symmetry_array[k]);
                                }
                           }
                           _struct_biol.asym_array.Add(_asym_symm);
                           q = strtok(NULL, ",");
                      }
                      delete [] buffer;
                 }
            }
            _struct_biol_array.Add(_struct_biol);
       }

       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       if (is_empty_table(Table)) {
            fprintf(stdout, "Empty struct_biol_gen category\n");
            exit (1);
       }

       rowNo = Table->GetNumRows();
       colNo = Table->GetNumColumns();
       row = rowNo;
       biol_index = Table->GetColumnIndex("biol_id");
       asym_index = Table->GetColumnIndex("asym_id");
       symmetry_index = Table->GetColumnIndex("symmetry");
       label = new int[rowNo];
       for (i = 0; i < rowNo; i++) label[i] = 0;

       line.Clear();
       for (i = 0; i < (int) _struct_biol_array.Length(); i++) {
            fprintf(stdout, "Biol. Unit: %d\n", _struct_biol_array[i].biol_id);
            for (j = 0; j < (int) _struct_biol_array[i].asym_array.Length(); j++) {
                 fprintf(stdout, "%s 1_555 ", _struct_biol_array[i].asym_array[j].asym_id.Text());
                 irow = -1;
                 for (k = 0; k < rowNo; k++) {
                      Table->GetCell(cifstring, asym_index, k);
                      if (cifstring == _struct_biol_array[i].asym_array[j].asym_id) {
                           irow = k;
                           label[k] = 1;
                           sprintf(buf, "%d", _struct_biol_array[i].biol_id);
                           cifstring.Copy(buf);
                           Table->UpdateCell(cifstring, biol_index, k);
                           break;
                      }
                 }
                 if (irow < 0) {
                      line += " ";
                      line += _struct_biol_array[i].asym_array[j].asym_id;
                 }
                 for (k = 0; k < (int) _struct_biol_array[i].asym_array[j].symmetry.Length(); k++) {
                      fprintf(stdout, "%s ", _struct_biol_array[i].asym_array[j].symmetry[k].Text());
                      if (irow >= 0) {
                           Table->AddRow();
                           for (l = 0; l < colNo; l++) {
                                Table->GetCell(cifstring, l, irow);
                                Table->UpdateCell(cifstring, l, row);
                           }
                           Table->UpdateCell(_struct_biol_array[i].asym_array[j].symmetry[k], symmetry_index, row);
                           row++;
                      }
                 }
                 fprintf(stdout, "\n");
            }
       }
       cifstring.Clear();
       if (line != "") {
            cifstring += "The following asym id(s) do(es) not exist in struct_biol_gen category:\n";
            cifstring += line;
            cifstring += "\n";
       }
       line.Clear();
       for (i = 0; i < rowNo; i++) {
            if (!label[i]) {
                 Table->GetCell(asym_ids, asym_index, i);
                 line += asym_ids;
                 line += " ";
            }
       }
       if (line != "") {
            cifstring += "Support file does not include the following asym id(s):\n";
            cifstring += line;
            cifstring += "\n";
       }
       if (cifstring != "") {
            fprintf(stdout, "%s\n", cifstring.Text());
            exit (1);
       }
       if (label) delete [] label;
       _CifObj->WriteTable(Table, _StructureId, "struct_biol_gen");
}

void _MAXIT::processing_ndb_poly_seq_scheme(const char *support_file)
{
       int i, j, terminal_flag = 0, rowNo, colNo, id_index, entity_index;
       int mon_index, num_index, pdb_index, row = 0, found_error = 0; 
       char *p = NULL;
       FILE *fp = NULL;
       _ENTITY_TERMINAL_SEQ _entity_terminal;
       CifString line, last_id, last_entity, entity, c_seq, n_seq, sequence, cifstring;
       CifString id, pdb_id, pdb_num, last_pdb_id, last_pdb_num;
       ISTable *Table = NULL, *Table1 = NULL;
       ReVarCifArray<CifString> *ColumnNames = NULL;
       ReVarCifArray<_ENTITY_TERMINAL_SEQ> _entity_terminal_array;

       fp = fopen(support_file, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", support_file);
            fprintf(stdout, "Can not read file %s!\n", support_file);
            exit (1);
       }

       entity.Clear();
       sequence.Clear();
       c_seq.Clear();
       n_seq.Clear();
       _entity_terminal_array.Clear();
       while (!feof(fp)) {
            get_line_from_file(fp, line);
            if (feof(fp)) break;

            line.RemoveBlanks();

            if ((p = strstr(line.Text(), "ENTITY_ID:"))) {
                 if (terminal_flag == 1)
                      n_seq = sequence;
                 else if (terminal_flag == 2)
                      c_seq = sequence;
                 if (entity != "" && (c_seq != "" || n_seq != "")) {
                      _entity_terminal.entity_id = entity;
                      _entity_terminal.n_terminal = n_seq;
                      _entity_terminal.c_terminal = c_seq;
                      _entity_terminal_array.Add(_entity_terminal);
                 }
                 entity.Clear();
                 p += 10;
                 for (i = 0; i < (int) strlen(p); i++) {
                      if (p[i] != ' ' && p[i] != '\0') {
                           entity += p[i];
                      }
                 }
                 sequence.Clear();
                 c_seq.Clear();
                 n_seq.Clear();
                 terminal_flag = 0;
            } else if (strstr(line.Text(), "N-T:")) {
                 if (terminal_flag == 2 && sequence != "") {
                      c_seq = sequence;
                      sequence.Clear();
                 }
                 terminal_flag = 1;
            } else if (strstr(line.Text(), "C-T:")) {
                 if (terminal_flag == 1 && sequence != "") {
                      n_seq = sequence;
                      sequence.Clear();
                 }
                 terminal_flag = 2;
            // } else if (strstr(line.Text(), "END")) {
            } else if (line == "END") {
                 if (terminal_flag == 1)
                      n_seq = sequence;
                 else if (terminal_flag == 2)
                      c_seq = sequence;
                 if (entity != "" && (c_seq != "" || n_seq != "")) {
                      _entity_terminal.entity_id  = entity;
                      _entity_terminal.n_terminal = n_seq;
                      _entity_terminal.c_terminal = c_seq;
                      _entity_terminal.is_protein = 1;
                      _entity_terminal_array.Add(_entity_terminal);
                 }
                 break;
            } else {
                 for (i = 0; i < (int) strlen(line.Text()); i++) {
                      if (line[i] != ' ' && line[i] != '\t' && line[i] != '\n') {
                           sequence += line[i];
                      }
                 }
            }
       }
       fclose (fp);

       Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            entity_index = Table->GetColumnIndex("entity_id");
            id_index = Table->GetColumnIndex("type");
            for (i = 0; i < (int) _entity_terminal_array.Length(); i++) {
                 for (j = 0; j < rowNo; j++) {
                      Table->GetCell(cifstring, entity_index, j);
                      if (cifstring == _entity_terminal_array[i].entity_id) {
                           Table->GetCell(cifstring, id_index, j);
                           if (cifstring == "polydeoxyribonucleotide" ||
                               cifstring == "polyribonucleotide" ||
                               cifstring == "polydeoxyribonucleotide/polyribonucleotide hybrid") {
                                _entity_terminal_array[i].is_protein = 0;    
                           }
                           break;
                      }
                 }
            }
       }

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_poly_seq_scheme");
       if (is_empty_table(Table)) {
            fprintf(stdout, "Empty ndb_poly_seq_scheme category\n");
            exit (1);
       }

       found_error = 0;
       for (i = 0; i < (int) _entity_terminal_array.Length(); i++) {
            fprintf(stdout, "Entity: %s\n", _entity_terminal_array[i].entity_id.Text());
            if (_entity_terminal_array[i].n_terminal != "") {
                 fprintf(stdout, "N-terminal: %s\n", _entity_terminal_array[i].n_terminal.Text());
            }
            if (_entity_terminal_array[i].c_terminal != "") {
                 fprintf(stdout, "C-terminal: %s\n", _entity_terminal_array[i].c_terminal.Text());
            }
            if (_entity_terminal_array[i].n_terminal == "" &&
                _entity_terminal_array[i].c_terminal == "") found_error = 1;
       }
       if (found_error) exit (1);

       rowNo = Table->GetNumRows();
       colNo = Table->GetNumColumns();
       ColumnNames = Table->GetColumnNames();
       Table1 = new ISTable("ndb_poly_seq_scheme");
       for (i = 0; i < colNo; i++) {
            Table1->AddColumn((*ColumnNames)[i].Text());
       }
       delete ColumnNames;
       id_index = Table->GetColumnIndex("id");
       entity_index = Table->GetColumnIndex("entity_id");
       mon_index = Table->GetColumnIndex("mon_id");
       num_index = Table->GetColumnIndex("pdb_num");
       pdb_index = Table->GetColumnIndex("pdb_id");
       last_id.Copy("___");
       last_entity.Copy("___");
       last_pdb_num.Copy("___");
       last_pdb_id.Copy("___");
       row = 0;
       for (i = 0; i < rowNo; i++) {
            Table->GetCell(id, id_index, i);
            Table->GetCell(entity, entity_index, i);
            Table->GetCell(pdb_num, num_index, i);
            Table->GetCell(pdb_id, pdb_index, i);
            if (id != last_id) {
                 if (last_entity != "___") {
                      add_terminal_residues(Table1, row, last_id, last_entity, last_pdb_num,
                               last_pdb_id, id_index, entity_index, mon_index, num_index,
                               pdb_index, 1, _entity_terminal_array);
                 }
                 add_terminal_residues(Table1, row, id, entity, pdb_num, pdb_id, id_index,
                               entity_index, mon_index, num_index, pdb_index, 0,
                              _entity_terminal_array);
            }
            Table1->AddRow();
            for (j = 0; j < colNo; j++) {
                 Table->GetCell(cifstring, j, i);
                 Table1->UpdateCell(cifstring, j, row);
            }
            row++;
            last_id = id;
            last_entity = entity;
            last_pdb_num = pdb_num;
            last_pdb_id = pdb_id;
       }
       if (last_entity != "___") {
            add_terminal_residues(Table1, row, last_id, last_entity, last_pdb_num,
                     last_pdb_id, id_index, entity_index, mon_index, num_index,
                     pdb_index, 1, _entity_terminal_array);
       }

       _CifObj->DeleteTable(_StructureId, "ndb_poly_seq_scheme");
       // _CifObj->DeleteTable(_StructureId, "rcsb_missing_residue_list");
       remove_missing_residue();
       _CifObj->WriteTable(Table1, _StructureId, "ndb_poly_seq_scheme");

       _re_assign_pdb_number = 1;
}

void _MAXIT::get_virus_matrix_from_file(const char *support_file)
{
       int i, j, found_index = 0, m_index = 0, i_index = 0, index[12], total_number = 0;
       int rowNo, asym_id_idx, pdb_id_idx;
       const char *name_index[12] = { "M11", "M12", "M13", "M21", "M22", "M23", "M31", "M32", "M33",
                                "V1", "V2", "V3" };
       char buffer[10];
       CifString line, cifstring, matrices[1000][12];
       FILE *fp = NULL;
       ISTable *Table = NULL;
       ReVarCifArray<CifString> wordlist, *asym_id = NULL, *pdb_id = NULL;

       fp = fopen(support_file, "r");
       if (fp == (FILE *) NULL) {
            ndb_log_message(NDB_MSG_ERR, "Can not read file %s!\n", support_file);
            fprintf(stdout, "Can not read file %s!\n", support_file);
            exit (1);
       }

       for (i = 0; i < 12; i++) index[i] = -1;
       while (!feof(fp)) {
            get_line_from_file(fp, line);
            if (feof(fp)) break;
            if (line == "") continue;
            line.ToUpper(); 
            if (strstr(line.Text(), "M11")) {
                 get_wordlist_from_string(wordlist, line.Text());
                 found_index = 0;
                 if (wordlist.Length() == 12) {
                      for (i = 0; i < 12; i++) {
                           for (j = 0; j < 12; j++) {
                                if (wordlist[j] == name_index[i]) {
                                     index[j] = i;
                                     break;
                                }
                           }
                           if (j < 12) found_index++;
                      }
                 }
                 if (found_index == 12)
                      found_index = 1;
                 else {
                      for (i = 0; i < 12; i++) {
                           if (index[i] < 0) fprintf(stdout, "Can not find index %s\n", name_index[i]);
                      }
                      found_index = 0;
                      exit (1);
                 }
            } else if (found_index) {
                 get_wordlist_from_string(wordlist, line.Text());
                 for (i = 0; i < (int) wordlist.Length(); i++) {
                      matrices[m_index][index[i_index]] = wordlist[i];
                      total_number++;
                      i_index++;
                      if (i_index == 12) {
                           i_index = 0;
                           m_index++;
                      }
                 }
            }
       }
       fclose (fp);

       // if (m_index != 60 || total_number != 720) {
/*
       if (m_index != 60 && m_index != 48) {
            fprintf(stdout, "It should have 60 matrices, but got %d matrices:\n", m_index);
            for (i = 0; i < m_index; i++) {
                 for (j = 0; j < 12; j++) {
                      if (!(j % 3)) fprintf(stdout, "\n\t");
                      fprintf(stdout, "%12s ", matrices[i][j].Text());
                 }
                 fprintf(stdout, "\n");
            }
            exit (1);
       }
*/
       if (_CifObj->IsTablePresent(_StructureId, "struct_ncs_oper")) {
            _CifObj->DeleteTable(_StructureId, "struct_ncs_oper");
       }

       Table = new ISTable("struct_ncs_oper");
       Table->AddColumn("id");
       Table->AddColumn("code");
       Table->AddColumn("details");
       Table->AddColumn("matrix[1][1]");
       Table->AddColumn("matrix[1][2]");
       Table->AddColumn("matrix[1][3]");
       Table->AddColumn("matrix[2][1]");
       Table->AddColumn("matrix[2][2]");
       Table->AddColumn("matrix[2][3]");
       Table->AddColumn("matrix[3][1]");
       Table->AddColumn("matrix[3][2]");
       Table->AddColumn("matrix[3][3]");
       Table->AddColumn("vector[1]");
       Table->AddColumn("vector[2]");
       Table->AddColumn("vector[3]");

       for (i = 0; i < m_index; i++) {
            Table->AddRow();
            sprintf(buffer, "%d", i + 1);
            cifstring.Copy(buffer);
            Table->UpdateCell(cifstring, 0, i);
            if (i == 0)
                 cifstring.Copy("given");
            else cifstring.Copy("generate");
            Table->UpdateCell(cifstring, 1, i);
            for (j = 0; j < 12; j++) {
                 Table->UpdateCell(matrices[i][j], j + 3, i);
            }
       }
       _CifObj->WriteTable(Table, _StructureId, "struct_ncs_oper");

       if (_CifObj->IsTablePresent(_StructureId, "rcsb_struct_ncs_virus_gen")) {
            _CifObj->DeleteTable(_StructureId, "rcsb_struct_ncs_virus_gen");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       rowNo = Table->GetNumRows();
       asym_id_idx = Table->GetColumnIndex("asym_id");
       pdb_id_idx  = Table->GetColumnIndex("ndb_new_pdb_asym_id");
       asym_id = Table->GetColumn(asym_id_idx);
       pdb_id  = Table->GetColumn(pdb_id_idx);

       Table = new ISTable("rcsb_struct_ncs_virus_gen");
       Table->AddColumn("id");
       Table->AddColumn("oper_id");
       Table->AddColumn("asym_id");
       Table->AddColumn("pdb_chain_id");

       total_number = 0;
       for (i = 0; i < m_index; i++) {
            for (j = 0; j < rowNo; j++) {
                 Table->AddRow();
                 sprintf(buffer, "%d", total_number + 1);
                 cifstring.Copy(buffer);
                 Table->UpdateCell(cifstring, 0, total_number);
                 sprintf(buffer, "%d", i + 1);
                 cifstring.Copy(buffer);
                 Table->UpdateCell(cifstring, 1, total_number);
                 Table->UpdateCell((*asym_id)[j], 2, total_number);
                 Table->UpdateCell((*pdb_id)[j], 3, total_number);
                 total_number++;
            }
       }

       _CifObj->WriteTable(Table, _StructureId, "rcsb_struct_ncs_virus_gen");

       if (asym_id) delete asym_id;
       if (pdb_id) delete pdb_id;
}

static void get_line_from_file(FILE *fp, CifString &string)
{
       int c;
       char t;

       string.Copy("");
       while (!feof(fp)) {
            c = fgetc(fp);
            if (c == '\n' || c == EOF) break;
            t = c;
            string += t;
       }
}

static void add_terminal_residues(ISTable *Table, int &row, CifString &id, CifString &entity,
               CifString &pdb_num, CifString &pdb_id, const int id_index, const int entity_index,
               const int mon_index, const int num_index, const int pdb_index, const int type,
               ReVarCifArray<_ENTITY_TERMINAL_SEQ> &_array)
{
       int i, j, alloc_num, num, start_number = 0;
       char **seqs = NULL, buffer[10];
       CifString sequence;

       int is_protein = 0;
       sequence.Clear();
       for (i = 0; i < (int) _array.Length(); i++) {
            if (entity == _array[i].entity_id) {
                 if (type == 0) sequence = _array[i].n_terminal;
                 else if (type == 1) sequence = _array[i].c_terminal;
                 is_protein = _array[i].is_protein;
                 break;
            }
       }
       if (sequence == "") return;

       alloc_num = strlen(sequence.Text());
       seqs = new char*[alloc_num];
       for (i = 0; i < alloc_num; i++) seqs[i] = new char[4];
       num = 0;
       for (i = 0; i < alloc_num; i++) {
            if (sequence[i] == ' ') continue;
            if (sequence[i] != '(') {
                 sprintf(seqs[num], "%c", sequence[i]);
                 ndb_set_string_to_upper(seqs[num]);
                 num++;
            } else {
                 i++;
                 j = 0;
                 while (sequence[i] != ')') {
                      seqs[num][j] = sequence[i];
                      i++; j++;
                 }
                 seqs[num][j] = '\0';
                 ndb_set_string_to_upper(seqs[num]);
                 num++;
            }
       }
       for (i = 0; i < num; i++) {
            if (strlen(seqs[i]) == 1 && is_protein) {
                 strcpy(buffer, get_threeletter_code(seqs[i][0]));
                 ndb_set_string_to_upper(buffer);
                 strcpy(seqs[i], buffer);
            }
       }
       if (type == 0)
            start_number = atoi(pdb_num.Text()) - num;
       else if (type == 1)
            start_number = atoi(pdb_num.Text()) + 1;

       for (i = 0; i < num; i++) {
            Table->AddRow();
            Table->UpdateCell(id, id_index, row);
            Table->UpdateCell(entity, entity_index, row);
            Table->UpdateCell(pdb_id, pdb_index, row);
            sequence.Copy(seqs[i]);
            Table->UpdateCell(sequence, mon_index, row);
            sprintf(buffer, "%d", start_number + i);
            sequence.Copy(buffer);
            Table->UpdateCell(sequence, num_index, row);
            row++;
       }

       for (i = 0; i < alloc_num; i++) delete [] seqs[i];
       delete [] seqs;
}
/*
static void get_pdbformat_symmetry(char *symm, const char *old_symm)
{
       *symm = '\0';
       if (!strcmp(old_symm, "")) strcpy(symm, "1555");
       else {
           for (unsigned int i = 0; i < strlen(old_symm); i++) {
                if (isdigit(old_symm[i])) {
                     *symm = old_symm[i];
                     symm++;
                }
           }
           *symm = '\0';
       }
}
*/
