#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

#include "_fileobj_ioutil.h"
#include "_maxit.h"
#include "erf_category.h"
#include "erf_util.h"
#include "util.h"

void _MAXIT::regenerate_site_record()
{
       int is_software_site_record = 0;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_site");
       if (!is_empty_table(Table)) {
            int index = Table->GetColumnIndex("pdbx_evidence_code");
            int rowNo = Table->GetNumRows();
            int i;
            CifString cifstring;
            for (i = 0; i < rowNo; i++) {
                 get_value(cifstring, Table, index, i);
                 cifstring.ToUpper();
                 if (cifstring != "SOFTWARE") break;
            }
            if (i == rowNo) is_software_site_record = 1;
       }

       if (!is_software_site_record) return;

       One_Card_Info_Format *pCard = _CardInfo[SITE_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[SITE_TOKEN-1].delete_a_card(pCard);
       Clear_Site();
       if (_CifObj) {
            _CifObj->DeleteTable(_StructureId, "struct_site");
            _CifObj->DeleteTable(_StructureId, "struct_site_gen");
       }
       get_site_record();
}

void _MAXIT::get_site_record()
{
       if (_n_het == 0 || _n_site) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
       if (!is_empty_table(Table)) {
            CifString cifstring;
            int index = Table->GetColumnIndex("skip_PDB_REMARK");
            get_value(cifstring, Table, index, 0);
            cifstring.ToLower();
            if (strstr(cifstring.Text(), "site")) return;
       }

       CifString site_path;
       site_path.Copy(root_dir);
       site_path += "/site/get_site.csh";
 
       struct stat statbuf;
       if (stat(site_path.Text(), &statbuf) != 0) return;

       ndb_to_pdb_get_remark_290();

       char pdb_code[10];
       memset(pdb_code, 0, 10);
       Table = _CifObj->GetTablePtr(_StructureId, "database_2");
       if (!is_empty_table(Table)) get_pdb_id(Table, pdb_code);

       ndb_set_string_to_lower(pdb_code);

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
       if (!is_empty_table(Table)) {
            for (int i = 0; i < MxCategory; i++) {
                 if (!strcmp(CifCategory[i].category, "struct_ref_seq_dif")) {
                      cifndb_process_general_category(Table, i, 0);
                      break;
                 }
            }
       }

       One_Card_Info_Format *pCard = _CardInfo[HET_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[HET_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[MODRES_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[MODRES_TOKEN-1].delete_a_card(pCard);

       output_het_and_modres(NDB_FILE_FORMAT_PDB);

       int SEQADV_TOKEN = find_card_id("SEQADV");

       char filename[20];
       sprintf(filename, "pdb%s.ent", pdb_code);
       FILE *fp = fopen(filename, "w");
       int numRemark = 0;
       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       if (!pCard) {
            Table = _CifObj->GetTablePtr(_StructureId, "exptl");
            if (!is_empty_table(Table)) {
                 pCard = _CardInfo[EXPDTA_TOKEN- 1].add_a_new_card(NdbToken[EXPDTA_TOKEN-1], 0);
                 int rowNo = Table->GetNumRows();
                 int idx = Table->GetColumnIndex("method");
                 CifString cs, cs1;
                 cs.Clear();
                 for (int i = 0; i < rowNo; ++i) {
                      get_value(cs1, Table, idx, i);
                      if (cs1 != "") {
                           if (cs != "") cs += "; ";
                           cs += cs1;
                      }
                 }
                 pCard->AddElementToField(cs.Text(), 2);
            }
       }

       get_seqres_from_mapping(NDB_FILE_FORMAT_PDB);

       print_general_cards(fp, EXPDTA_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, REMARK_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, SEQADV_TOKEN, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, MODRES_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, SEQRES_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, HET_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, CRYST1_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, ORIGX_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_general_cards(fp, SCALE_TOKEN - 1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       print_molecules(fp, NDB_FILE_FORMAT_PDB, 0, 0, 0, 1);
       fclose (fp);

       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[EXPDTA_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[SEQRES_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[SEQRES_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[REMARK_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[REMARK_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[HET_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[HET_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[MODRES_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[MODRES_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[SEQADV_TOKEN].Head();
       while (pCard) pCard = _CardInfo[SEQADV_TOKEN].delete_a_card(pCard);

       CifString script;
       script.Copy("setenv RCSBROOT ");
       script += root_dir; script += "; "; script += site_path;
       script += " "; script += pdb_code; script += " >& site.record; ";
       ndb_execute_command(script.Text());

       script.Copy(pdb_code);
       script += "_site.cif";
       if (stat(script.Text(), &statbuf) != 0) {
            script.Copy("rm site.record pdb");
            script += pdb_code; script += ".ent ";
            ndb_execute_command(script.Text());
            return;
       }

       char odbfile[100], *diags = NULL;
       strcpy(odbfile, "./XXXXXX");
       int fd = mkstemp(odbfile);
       close (fd);

       CifFileObj *fobj = new CifFileObj(odbfile, WRITE_MODE);
       fobj->Read(script.Text(), diags);
       if (diags && strlen(diags)) {
            ndb_log_message(NDB_MSG_ERR, "%s: Read cif file %s failed\n%s\n",
                    _input_filename, script.Text(), diags);
            if (fobj) delete fobj;
            remove(odbfile);

            script.Copy("rm site.record");
            script += " site."; script += pdb_code;
            script += " rem800."; script += pdb_code;
            script += " contacts."; script += pdb_code;
            script += " pdb"; script += pdb_code; script += ".ent ";
            script += pdb_code; script += "_site.cif";
            ndb_execute_command(script.Text());
            return;
       }

       int numBlock = 0;
       char **blockName = NULL;
       blockName = fobj->GetBlockNames(numBlock);
       ISTable *t1 = fobj->GetTablePtr(blockName[0], "struct_site");
       ISTable *t = new ISTable("struct_site");
       t->AddColumn("id");
       t->AddColumn("details");
       t->AddColumn("pdbx_evidence_code");
       
       int rowNo = t1->GetNumRows();
       int idx1 = t1->GetColumnIndex("id");
       int idx2 = t1->GetColumnIndex("pdbx_auth_comp_id");
       int idx3 = t1->GetColumnIndex("pdbx_auth_asym_id");
       int idx4 = t1->GetColumnIndex("pdbx_auth_seq_id");
       int idx5 = t1->GetColumnIndex("pdbx_auth_ins_code");
       int idx6 = t1->GetColumnIndex("details");
       CifString cs1, cs2, cs3, cs4, cs5, cs6;
       for (int i = 0; i < rowNo; ++i) {
            t->AddRow();
            get_value(cs1, t1, idx1, i);
            t->UpdateCell(cs1, 0, i);
            get_value(cs2, t1, idx2, i);
            get_value(cs3, t1, idx3, i);
            get_value(cs4, t1, idx4, i);
            get_value(cs5, t1, idx5, i);
            if (cs2 == "" && cs3 == "" && cs4 == "" && cs5 == "")
                 get_value(cs1, t1, idx6, i);
            else {
                 cs1.Copy("BINDING SITE FOR RESIDUE ");
                 cs1 += cs2; cs1 += " "; cs1 += cs3; cs1 += " ";
                 cs1 += cs4; cs1 += cs5;
            }
            t->UpdateCell(cs1, 1, i);
            cs1.Copy("SOFTWARE");
            t->UpdateCell(cs1, 2, i);
       }
       _CifObj->WriteTable(t, _StructureId, "struct_site");

       _MOLECULE *mol = _molecules[0];
       t1 = fobj->GetTablePtr(blockName[0], "struct_site_gen");
       rowNo = t1->GetNumRows();
       idx1 = t1->GetColumnIndex("site_id");
       idx2 = t1->GetColumnIndex("auth_comp_id");
       idx3 = t1->GetColumnIndex("auth_asym_id");
       idx4 = t1->GetColumnIndex("auth_seq_id");
       idx5 = t1->GetColumnIndex("pdbx_auth_ins_code");
       idx6 = t1->GetColumnIndex("symmetry");
       int FstIndex = -1, FstRes = -1;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs1, t1, idx1, i);
            get_value(cs2, t1, idx2, i);
            get_value(cs3, t1, idx3, i);
            get_value(cs4, t1, idx4, i);
            get_value(cs5, t1, idx5, i);
            get_value(cs6, t1, idx6, i);
            if (cs6 == "1_333") cs6.Copy("1_555");
            if (mol->find_pdb_residue(cs3.Text(), cs2.Text(), atoi(cs4.Text()),
                      cs5.Text(), FstIndex, FstRes)) {
                 Insert_a_Site(cs1.Text(), FstIndex, FstRes, cs6.Text());
            }
       }

       input_site_record(NDB_FILE_FORMAT_PDB);

       delete fobj;
       remove(odbfile);
       for (int i = 0; i < numBlock; i++) delete [] blockName[i];
       delete [] blockName;

       script.Copy("rm site.record");
       script += " site."; script += pdb_code;
       script += " rem800."; script += pdb_code;
       script += " contacts."; script += pdb_code;
       script += " pdb"; script += pdb_code; script += ".ent ";
       script += pdb_code; script += "_site.cif";
       ndb_execute_command(script.Text());
}

void _MAXIT::merge_site_record(const char *filename)
{
       struct stat statbuf;
       if (stat(filename, &statbuf) != 0) return;

       Clear_Site();

       char odbfile[100], *diags = NULL;
       strcpy(odbfile, "./XXXXXX");
       int fd = mkstemp(odbfile);
       close (fd);

       CifFileObj *fobj = new CifFileObj(odbfile, WRITE_MODE);
       fobj->Read(filename, diags);
       if (diags && strlen(diags)) {
            ndb_log_message(NDB_MSG_ERR, "%s: Read cif file %s failed\n%s\n",
                    _input_filename, filename, diags);
            exit (1);
       }

       int numBlock = 0;
       char **blockName = NULL;
       blockName = fobj->GetBlockNames(numBlock);
       ISTable *t1 = fobj->GetTablePtr(blockName[0], "struct_site");
       ISTable *t = new ISTable("struct_site");
       t->AddColumn("id");
       t->AddColumn("details");
       t->AddColumn("pdbx_evidence_code");
       
       int rowNo = t1->GetNumRows();
       int idx1 = t1->GetColumnIndex("id");
       int idx2 = t1->GetColumnIndex("pdbx_auth_comp_id");
       int idx3 = t1->GetColumnIndex("pdbx_auth_asym_id");
       int idx4 = t1->GetColumnIndex("pdbx_auth_seq_id");
       int idx5 = t1->GetColumnIndex("pdbx_auth_ins_code");
       int idx6 = t1->GetColumnIndex("details");
       CifString cs1, cs2, cs3, cs4, cs5, cs6;
       for (int i = 0; i < rowNo; ++i) {
            t->AddRow();
            get_value(cs1, t1, idx1, i);
            t->UpdateCell(cs1, 0, i);
            get_value(cs2, t1, idx2, i);
            get_value(cs3, t1, idx3, i);
            get_value(cs4, t1, idx4, i);
            get_value(cs5, t1, idx5, i);
            if (cs2 == "" && cs3 == "" && cs4 == "" && cs5 == "")
                 get_value(cs1, t1, idx6, i);
            else {
                 cs1.Copy("BINDING SITE FOR RESIDUE ");
                 cs1 += cs2; cs1 += " "; cs1 += cs3; cs1 += " ";
                 cs1 += cs4; cs1 += cs5;
            }
            t->UpdateCell(cs1, 1, i);
            cs1.Copy("SOFTWARE");
            t->UpdateCell(cs1, 2, i);
       }
       _CifObj->WriteTable(t, _StructureId, "struct_site");

       _MOLECULE *mol = _molecules[0];
       t1 = fobj->GetTablePtr(blockName[0], "struct_site_gen");
       rowNo = t1->GetNumRows();
       idx1 = t1->GetColumnIndex("site_id");
       idx2 = t1->GetColumnIndex("auth_comp_id");
       idx3 = t1->GetColumnIndex("auth_asym_id");
       idx4 = t1->GetColumnIndex("auth_seq_id");
       idx5 = t1->GetColumnIndex("pdbx_auth_ins_code");
       idx6 = t1->GetColumnIndex("symmetry");
       int FstIndex = -1, FstRes = -1;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs1, t1, idx1, i);
            get_value(cs2, t1, idx2, i);
            get_value(cs3, t1, idx3, i);
            get_value(cs4, t1, idx4, i);
            get_value(cs5, t1, idx5, i);
            get_value(cs6, t1, idx6, i);
            if (cs6 == "1_333") cs6.Copy("1_555");
            if (mol->find_pdb_residue(cs3.Text(), cs2.Text(), atoi(cs4.Text()),
                      cs5.Text(), FstIndex, FstRes)) {
                 Insert_a_Site(cs1.Text(), FstIndex, FstRes, cs6.Text());
            }
       }

       input_site_record(NDB_FILE_FORMAT_PDB);

       delete fobj;
       remove(odbfile);
       for (int i = 0; i < numBlock; i++) delete blockName[i];
       delete blockName;
}
