/*
FILE:     stereochemistry.C
*/
/*
VERSION:  8.200
*/
/*
DATE:     3/18/2015
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2015 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <string>
#include <map>
#include <vector>

#include "_validation.h"
#include "bond_angle_torsion.h"
#include "stereochemistry.h"
#include "util.h"

extern void ndb_writing_missing_words(FILE*, FILE*);

static void check_first_flag(FILE *template_fp, FILE *fp, int& first_flag,
                             const int model_num);
static void print_chiral_error(FILE *fp, const int model_num, const char* ResName,
                               const int pdb_res_no, const char *pdb_chnid, _ATOM *fst,
                               _ATOM *snd, _ATOM *trd, _ATOM *fth, _ATOM *f5h);
static void get_atom_name_with_alt_loc(std::string& name, _ATOM *atom);
static void ndb_check_E_Z(_RESIDUE *residue, const int Chn_Index, const int n_link,
             _LINK **link);
static void ndb_check_E_Z_switch(_RESIDUE *residue, EZ_CHECK *ez_check,
             const int Chn_Index, const int n_link, _LINK **link);
static void switch_link_atoms(const char *old_name, const char *alt_loc,
          const char *new_name, const int ChnIndex, const int ResIndex,
          const int n_link, _LINK **link);
static void get_defined_links(std::map<std::string, std::vector<std::string> >& links,
                              const int n_link, _LINK **link, _MOLECULE *mol);
static void additional_chiral_checking(FILE *template_fp, FILE *fp, int &first_flag,
                         const int _num_mol, _MOLECULE **_molecules,
                         std::map<std::string, std::vector<std::string> >& links,
                         CifString& error);
static void switch_other_O1P_O2P_atoms(_CHAIN *chain, const int n_link, _LINK **link);
static void get_exist_links(std::map<std::string, _ATOM*>& exist_links, _MOLECULE* mol,
                            std::map<std::string, std::vector<std::string> >& links);
static void get_chiral_center_mapping(std::map<std::string,
                std::vector<std::vector<std::string> > >& mapping, const char *ResName);
static void get_P_chiral_centers(std::vector<std::vector<std::string> >& chiral_centers,
                                 const char *ResName);
static void checking_chiral_center(FILE *template_fp, FILE *fp, int& flag,
                 const int model_num, const std::vector<std::vector<std::string> >&
                 chiral_list, const char *ResName, _RESIDUE *residue,
                 std::map<std::string, _ATOM*>& links, CifString& error);
static void checking_P_chiral_center(const std::vector<std::vector<std::string> >&
                 chiral_list, const int chn_index, _RESIDUE *residue, const int n_link,
                 _LINK **link);
static void get_chiral_info(const std::vector<AtomFormat*>& RefAtoms,
                            std::vector<std::string>& data);
static _ATOM **get_atom_list(_ATOM **c_atom, const int c_num,  const std::string& name,
                             int& num, const char *ResName, _RESIDUE *residue,
                             std::map<std::string, _ATOM*>& links);
void static reorder_chiral_atoms(std::vector<std::string>& list, ConnectFormat *drug);
static bool is_right_oxygen(ConnectFormat *drug, const std::string& p_atom,
                            const std::string& o_atom, std::string& h_atom);
static void write_chiral_error(CifString& error, const int model_num,
                               const char *ResName, const int pdb_res_no,
                               const char *pdb_chnid, _ATOM *fst, _ATOM *snd,
                               _ATOM *trd, _ATOM *fth);
// static void ndb_check_methyl_protons(_RESIDUE *residue);
// static void find_new_atom_labeling(MP_TORSION *torsion);
// static void get_j_k(int i, int &j, int &k);

void _VALIDATION::ndb_check_stereochemistry(CifString &message, int &count)
{
       double ang;
       char buffer[200], buffer1[20], chain_id[5], ins_code[5];
       CifString message1, message2, Atom1, Atom2, Atom3, Atom4;

       if (_num_mol > 1)
            message1.Copy(_chirality_label_nmr);
       else message1.Copy(_chirality_label);

       message2.Clear();

       _chiralIndex_init();
       ReVarCifArray<ReVarCifArray<CifString> > atom_lists;
       ReVarCifArray<CifString> atom_list;
       for (int i = 0; i < _num_mol; ++i) {
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                       CHIRALITYS_POOL *ext = find_chiral(residue->ResName());
/*
                      if (!ext && chain->chain_type() == ATOMN_TOKEN) {
                           ext = &phosphate_sugar;
                           ConnectFormat *drug = find_drug(residue->ResName());
                           if (drug) {
                                CifString cs;
                                cs.Copy(drug->chemical_name());
                                cs.ToUpper();
                                if (strstr(cs.Text(), "ARABINOSE")) {
                                     residue = chain->GetNextResidue();
                                     continue;
                                }
                           }
                      }
*/
                      if (!ext) {
                           residue = chain->GetNextResidue();
                           continue;
                      }
                      atom_lists.Clear();
                      strcpy(ins_code, " ");
                      for (int k = 0; k < ext->num_residues; ++k) {
                           if (ext->lists[k].num != 1) continue;
                           int num_a = 0;
                           int num_b = 0;
                           int num_c = 0;
                           int num_d = 0;
                           _ATOM **atom1 = residue->find_atom(ext->lists[k].chiral[0].
                                                        center_atom, num_a);
                           _ATOM **atom2 = residue->find_atom(ext->lists[k].chiral[0].
                                                        first_atom, num_b);
                           _ATOM **atom3 = residue->find_atom(ext->lists[k].chiral[0].
                                                        second_atom, num_c);
                           _ATOM **atom4 = residue->find_atom(ext->lists[k].chiral[0].
                                                        check_atom, num_d);
                           if (!atom1 || !atom2 || !atom3 || !atom4) {
                                if (atom1) delete [] atom1;
                                if (atom2) delete [] atom2;
                                if (atom3) delete [] atom3;
                                if (atom4) delete [] atom4;
                                continue;
                           }
                           int num = 0;
                           _FOUR_ATOMS *four_atoms = find_four_atoms(num, atom1, num_a,
                                            atom2, num_b, atom3, num_c, atom4, num_d);
                           for (int j = 0; j < num; ++j) {
                                double volume = calc_chiral_volume(four_atoms[j].fst,
                                            four_atoms[j].snd, four_atoms[j].trd,
                                            four_atoms[j].fth, ang);
                                if (volume * ext->lists[k].chiral[0].correct_value < 0.0) {
                                     Atom1.Copy(four_atoms[j].fst->pdb_atmnam());
                                     if (strcmp(four_atoms[j].fst->alt_loc(), "")) {
                                          Atom1 += "(";
                                          Atom1 += four_atoms[j].fst->alt_loc();
                                          Atom1 += ")";
                                     }
                                     Atom2.Copy(four_atoms[j].snd->pdb_atmnam());
                                     if (strcmp(four_atoms[j].snd->alt_loc(), "")) {
                                          Atom2 += "(";
                                          Atom2 += four_atoms[j].snd->alt_loc();
                                          Atom2 += ")";
                                     }
                                     Atom3.Copy(four_atoms[j].trd->pdb_atmnam());
                                     if (strcmp(four_atoms[j].trd->alt_loc(), "")) {
                                          Atom3 += "(";
                                          Atom3 += four_atoms[j].trd->alt_loc();
                                          Atom3 += ")";
                                     }
                                     Atom4.Copy(four_atoms[j].fth->pdb_atmnam());
                                     if (strcmp(four_atoms[j].fth->alt_loc(), "")) {
                                          Atom4 += "(";
                                          Atom4 += four_atoms[j].fth->alt_loc();
                                          Atom4 += ")";
                                     }
                                     bool found1 = false;
                                     bool found2 = false;
                                     bool found3 = false;
                                     bool found4 = false;
                                     for (unsigned int l = 0; l < atom_lists.Length(); ++l) {
                                          if (atom_lists[l][0] != Atom1) continue;
                                          found1 = true;
                                          for (unsigned int m = 1;
                                                        m < atom_lists[l].Length(); ++m) {
                                               if (atom_lists[l][m] == Atom2) found2 = true;
                                               if (atom_lists[l][m] == Atom3) found3 = true;
                                               if (atom_lists[l][m] == Atom4) found4 = true;
                                          }
                                     }
                                     if (found1 && found2 && found3 && found4) continue;

                                     atom_list.Clear();
                                     atom_list.Add(Atom1);
                                     atom_list.Add(Atom2);
                                     atom_list.Add(Atom3);
                                     atom_list.Add(Atom4);
                                     atom_lists.Add(atom_list);

                                     if (strcmp(four_atoms[j].fth->ins_code(), ""))
                                          strcpy(ins_code, four_atoms[j].fth->ins_code());
                                }
                           }
                           if (four_atoms) delete [] four_atoms;
                           if (atom1) delete [] atom1;
                           if (atom2) delete [] atom2;
                           if (atom3) delete [] atom3;
                           if (atom4) delete [] atom4;
                      }
                      if (atom_lists.Length() > 0) {
                           strcpy(chain_id, " ");
                           if (strcmp(chain->PDB_ChainID(), ""))
                                strcpy(chain_id, chain->PDB_ChainID());
                           if (_num_mol > 1)
                                sprintf(buffer, "        %3s      %s     %4d%s    %3d    ",
                                         residue->ResName(), chain_id,
                                         residue->pdb_res_no(), ins_code,
                                         _molecules[i]->Mol_ID());
                           else sprintf(buffer, "        %3s      %s     %4d%s           ",
                                         residue->ResName(), chain_id,
                                         residue->pdb_res_no(), ins_code);
                           for (unsigned int l = 0; l < atom_lists.Length(); ++l) {
                                message2 += buffer;
                                sprintf(buffer1, "     %-10s", atom_lists[l][0].Text());
                                message2 += buffer1;
                                for (unsigned int m = 1; m < atom_lists[l].Length(); ++m) {
                                     message2 += " " + atom_lists[l][m];
                                }
                                message2 += "\n";
                           }
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
       if (message2 != "") {
            message += message1 + message2;
            count++;
       }
}

int _VALIDATION::ndb_check_stereochemistry(FILE *template_fp, FILE *fp, CifString& error)
{
       int first_flag = 1, i = 0, j = 0, k = 0, l = 0, first_error = 1;
       int num = 0, num_a = 0, num_b = 0, num_c = 0, num_d = 0;
       double volume = 0, ang;
       _ATOM **atom1 = NULL, **atom2 = NULL, **atom3 = NULL, **atom4 = NULL;
       _RESIDUE *prev = NULL, *residue = NULL;
       _CHAIN *chain = NULL;
       CifString cifstring;
       _FOUR_ATOMS *four_atoms = NULL;
       CHIRALITYS_POOL *ext = NULL;

       _chiralIndex_init();
       for (i = 0; i < _num_mol; i++) {
            int model_num = -1;
            if (_num_mol > 1) model_num = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 prev = NULL;
                 residue = chain->GetFirstResidue();
                 while (residue) {
                      ext = find_chiral(residue->ResName());
                      bool non_standard_flag = false;
                      if (!ext && chain->chain_type() == ATOMN_TOKEN) {
                           ext = find_chiral("DA");
                           non_standard_flag = true;
                      }
/*
                      if (!ext && chain->chain_type() == ATOMN_TOKEN) {
                           ext = &phosphate_sugar;
                           ConnectFormat *drug = find_drug(residue->ResName());
                           if (drug) {
                                CifString cs;
                                cs.Copy(drug->chemical_name());
                                cs.ToUpper();
                                if (strstr(cs.Text(), "ARABINOSE")) {
                                     prev = residue;
                                     residue = chain->GetNextResidue();
                                     continue;
                                }
                           }
                      }
*/
                      if (!ext) {
                           prev = residue;
                           residue = chain->GetNextResidue();
                           continue;
                      }
                      for (k = 0; k < ext->num_residues; k++) {
                           if (non_standard_flag && ext->lists[k].isPhosphorus == 0) continue;
                           for (l = 0; l < ext->lists[k].num; l++) {
                                atom1 = NULL; atom2 = NULL; atom3 = NULL; atom4 = NULL;
                                num_a = 0; num_b = 0; num_c = 0; num_d = 0;
                                if (ext->lists[k].isPhosphorus) {
                                     if (prev) {
                                          atom1 = residue->find_atom(ext->lists[k].chiral[l].center_atom,
                                                    num_a);
                                          atom2 = residue->find_atom(ext->lists[k].chiral[l].first_atom,
                                                    num_b);
                                          atom3 = prev->find_atom(ext->lists[k].chiral[l].second_atom,
                                                    num_c);
                                          atom4 = residue->find_atom(ext->lists[k].chiral[l].check_atom,
                                                    num_d);
                                          if (atom1 && atom3 && !is_a_bond(atom1[0], atom3[0])) {
                                               if (atom1) delete [] atom1;
                                               if (atom2) delete [] atom2;
                                               if (atom3) delete [] atom3;
                                               if (atom4) delete [] atom4;
                                               continue;
                                          }
                                     }
                                } else {
                                     atom1 = residue->find_atom(ext->lists[k].chiral[l].center_atom, num_a);
                                     atom2 = residue->find_atom(ext->lists[k].chiral[l].first_atom, num_b);
                                     atom3 = residue->find_atom(ext->lists[k].chiral[l].second_atom, num_c);
                                     atom4 = residue->find_atom(ext->lists[k].chiral[l].check_atom, num_d);
                                }
                                if (!atom1 || !atom2 || !atom3 || !atom4) {
                                     if (atom1) delete [] atom1;
                                     if (atom2) delete [] atom2;
                                     if (atom3) delete [] atom3;
                                     if (atom4) delete [] atom4;
                                     continue;
                                }
                                four_atoms = find_four_atoms(num, atom1, num_a, atom2,
                                                num_b, atom3, num_c, atom4, num_d);
                                for (j = 0; j < num; j++) {
                                     volume = calc_chiral_volume(four_atoms[j].fst,
                                                 four_atoms[j].snd, four_atoms[j].trd,
                                                 four_atoms[j].fth, ang);
                                     if (volume * ext->lists[k].chiral[l].correct_value
                                         < 0.0) {
                                          check_first_flag(template_fp, fp, first_flag,
                                                           model_num);
                                          print_chiral_error(fp, model_num,
                                               residue->ResName(), residue->pdb_res_no(),
                                               chain->PDB_ChainID(), four_atoms[j].fst,
                                               four_atoms[j].snd, four_atoms[j].trd,
                                               four_atoms[j].fth, four_atoms[j].fth);
                                          _Error_Check[6]++;
                                          if ((!strcmp(ext->lists[k].chiral[l].check_atom, "O2P") ||
                                               !strcmp(ext->lists[k].chiral[l].check_atom, "O1P")) &&
                                               _Error_Check_Message[6] == "") {
                                               _Error_Check_Message[6] += "O1P,O2P";
                                          }
                                     }
                                }
                                if (four_atoms) delete [] four_atoms;
                                if (atom1) delete [] atom1;
                                if (atom2) delete [] atom2;
                                if (atom3) delete [] atom3;
                                if (atom4) delete [] atom4;
                           }
                      }
                      prev = residue;
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       std::map<std::string, std::vector<std::string> > links;
       get_defined_links(links, _n_link, _link, _molecules[0]);

       additional_chiral_checking(template_fp, fp, first_flag, _num_mol, _molecules,
                                  links, error);
       if (first_flag) {
            ndb_writing_missing_words(template_fp, fp);
            fprintf(fp, "     none\n\n");
       } else _Error_Check[6]++;

       for (i = 0; i < _num_mol; i++) {
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 residue = chain->GetFirstResidue();
                 while (residue) {
                      cifstring.Copy("");
                      residue->cal_hydrogen_bond_distance(_molecules[i]->Mol_ID(), cifstring);
                      if (cifstring != "") {
                           _Error_Check[59]++;
                           if (first_error) {
                                // fprintf(fp, "\nCovalent bond distances involved in hydrogens:\n\n");
                                fprintf(fp, "\n    COVALENT BOND DISTANCES INVOLVED IN HYDROGENS:\n");
                                fprintf(fp, "    ----------------------------------------------\n\n");
                                first_error = 0;
                           }
                           fprintf(fp, "%s\n", cifstring.Text());
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       return (first_flag);
}

int _VALIDATION::ndb_check_E_Z_labeling(FILE *template_fp, FILE *fp)
{
       int first = 1, i = 0, j = 0, k = 0, l = 0, found_all_atoms = 1;
       _CHAIN *chain = NULL;
       _RESIDUE *residue = NULL;
       _ATOM *atom1 = NULL, *atom2 = NULL, *atom3 = NULL, *atom4 = NULL;
       double torsion = 0, tor[2];

       for (i = 0; i < _num_mol; i++) {
            // if (i) break;
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 residue = chain->GetFirstResidue();
                 while (residue) {
                      if (strcmp(residue->ResName(), "A") &&
                          strcmp(residue->ResName(), "ARG") &&
                          strcmp(residue->ResName(), "ASN") &&
/* */
                          strcmp(residue->ResName(), "ASP") &&
/* */
                          strcmp(residue->ResName(), "C") &&
                          strcmp(residue->ResName(), "G") &&
                          strcmp(residue->ResName(), "GLN") /* */ &&
                          strcmp(residue->ResName(), "GLU") &&
                          strcmp(residue->ResName(), "PHE") &&
                          strcmp(residue->ResName(), "TYR") /* */ ) {
                           residue = chain->GetNextResidue();
                           continue;
                      }

                      for (k = 0; k < NUM_EZ_CHECK; k++) {
                           if (strcmp(residue->ResName(), ez_check[k].residue))
                                continue;
                           found_all_atoms = 1;
                           for (l = 0; l < ez_check[k].n_check; l++) {
                                atom1 = residue->find_atom(ez_check[k].check_list[l].atom);
                                if (atom1 == NULL) {
                                     found_all_atoms = 0; break;
                                }
                           }
                           if (!found_all_atoms) continue;

                           if (strcmp(ez_check[k].no_atom, "") &&
                               residue->find_atom(ez_check[k].no_atom)) continue;

                           atom1 = residue->find_atom(ez_check[k].atom1);
                           atom2 = residue->find_atom(ez_check[k].atom2);
                           atom3 = residue->find_atom(ez_check[k].atom3);
                           atom4 = residue->find_atom(ez_check[k].atom4);
                           if (!atom1 || !atom2 || !atom3 || !atom4) continue;
                           torsion = cal_torsion(atom1, atom2, atom3, atom4);
                           torsion = mod(torsion, -180, 180);
                           if (fabs(torsion) > 90.0) {
                                for (j = 0; j < 2; j++) {
                                     atom4 = residue->find_atom(ez_check[k].check_list[j].atom);
                                     tor[j] = cal_torsion(atom1, atom2, atom3, atom4);
                                     tor[j] = mod(tor[j], -180, 180);
                                }
                                if (fabs(tor[0]) > fabs(tor[1])) {
                                     if (first) {
                                          first = 0;
                                          ndb_writing_missing_words(template_fp, fp);
                                          fprintf(fp, "Model  Chain  Residue  Residue  Atom Name  Original\n");
                                          fprintf(fp, "               Name    Number              Atom Name\n");
                                          fprintf(fp, "-----  -----  -------  -------  --------   ---------\n");
                                     }
                                     atom1 = residue->find_atom(ez_check[k].change_list[0].old_name);
                                     atom2 = residue->find_atom(ez_check[k].change_list[0].new_name);
                                     fprintf(fp, "%3d    %s      %4s    %5d     %5s",
                                          _molecules[i]->Mol_ID(), chain->PDB_ChainID(),
                                          residue->ResName(), residue->pdb_res_no(),
                                          atom1->atmtype());
                                     if (strcmp(atom1->atmtype(), atom1->pub_atmnam()))
                                          fprintf(fp, "       %5s\n", atom1->pub_atmnam());
                                     else fprintf(fp, "\n");
                                     fprintf(fp, "%3d    %s      %4s    %5d     %5s",
                                          _molecules[i]->Mol_ID(), chain->PDB_ChainID(),
                                          residue->ResName(), residue->pdb_res_no(),
                                          atom2->atmtype());
                                     if (strcmp(atom2->atmtype(), atom2->pub_atmnam()))
                                          fprintf(fp, "       %5s\n", atom2->pub_atmnam());
                                     else fprintf(fp, "\n");
                                     _Error_Check[6]++;
                                }
                           }
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
       if (first) ndb_writing_missing_words(template_fp, fp);

       return (first);
}

void _VALIDATION::ndb_check_chirality()
{
       int i = 0, j = 0, k = 0, l = 0, m = 0, n = 0, change = 0, exist[2];
       int num[4], num_alt_loc;
       char alt_loc[10];
       double volume = 0, ang;
       _ATOM *atoms[2], **atom[4], *Atom[4];
       _CHAIN *chain = NULL;
       CHIRALITYS_POOL *ext = NULL;
       _RESIDUE *prev = NULL, *residue = NULL;

       _chiralIndex_init();
       for (i = 0; i < _num_mol; i++) {
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 switch_other_O1P_O2P_atoms(chain, _n_link, _link);
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 prev = NULL;
                 residue = chain->GetFirstResidue();
                 while (residue) {
                      change = 0;
                      residue->change_hydrogen_name();
                      if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                          _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) {
                           if (!strcmp(residue->ResName(), "A") ||
                               !strcmp(residue->ResName(), "ARG") ||
                               !strcmp(residue->ResName(), "ASN") ||
/* */
                               !strcmp(residue->ResName(), "ASP") ||
/* */
                               !strcmp(residue->ResName(), "C") ||
                               !strcmp(residue->ResName(), "G") ||
                               !strcmp(residue->ResName(), "GLN") /* */ ||
                               !strcmp(residue->ResName(), "GLU") ||
                               !strcmp(residue->ResName(), "PHE") ||
                               !strcmp(residue->ResName(), "TYR") /* */ )
                                ndb_check_E_Z(residue, chain->index(), _n_link, _link);
/*
                           if (!strcmp(residue->ResName(), "ALA") ||
                               !strcmp(residue->ResName(), "ILE") ||
                               !strcmp(residue->ResName(), "LEU") ||
                               !strcmp(residue->ResName(), "LYS") ||
                               !strcmp(residue->ResName(), "MET") ||
                               !strcmp(residue->ResName(), "T") ||
                               !strcmp(residue->ResName(), "THR") ||
                               !strcmp(residue->ResName(), "VAL"))
                                ndb_check_methyl_protons(residue);
*/
                      }

                      ext = find_chiral(residue->ResName());
                      bool non_standard_flag = false;
                      if (!ext && chain->chain_type() == ATOMN_TOKEN) {
                           ext = find_chiral("DA");
                           non_standard_flag = true;
                      }
/*
                      if (!ext && chain->chain_type() == ATOMN_TOKEN) {
                           ext = &phosphate_sugar;
                           ConnectFormat *drug = find_drug(residue->ResName());
                           if (drug) {
                                CifString cs;
                                cs.Copy(drug->chemical_name());
                                cs.ToUpper();
                                if (strstr(cs.Text(), "ARABINOSE")) {
                                     prev = residue;
                                     residue = chain->GetNextResidue();
                                     continue;
                                }
                           }
                      }
*/
                      if (!ext) {
                           prev = residue;
                           residue = chain->GetNextResidue();
                           continue;
                      }

                      strcpy(alt_loc, residue->alt_loc());
                      num_alt_loc = strlen(alt_loc);
                      if (num_alt_loc == 0) num_alt_loc = 1;

                      for (k = 0; k < ext->num_residues; k++) {
                           if (non_standard_flag && ext->lists[k].isPhosphorus == 0) continue;
                           if (ext->lists[k].num != 2) continue;
                           for (m = 0; m < num_alt_loc; m++) {
                                for (l = 0; l < 2; l++) {
                                     atoms[l] = NULL;
                                     exist[l] = 0;
                                }
                                for (l = 0; l < ext->lists[k].num; l++) {
                                     for (n = 0; n < 4; n++) {
                                          atom[n] = NULL;
                                          num[n] = 0;
                                     }
                                     if (ext->lists[k].isPhosphorus) {
                                          if (prev) {
                                               atom[0] = residue->find_atom(ext->lists[k].chiral[l].center_atom,
                                                         num[0]);
                                               atom[1] = residue->find_atom(ext->lists[k].chiral[l].first_atom,
                                                         num[1]);
                                               atom[2] = prev->find_atom(ext->lists[k].chiral[l].second_atom,
                                                         num[2]);
                                               atom[3] = residue->find_atom(ext->lists[k].chiral[l].check_atom,
                                                         num[3]);
                                               if (atom[0] && atom[2] && !is_a_bond(atom[0][0], atom[2][0])) {
                                                    for (n = 0; n < 4; n++) {
                                                         if (atom[n]) delete [] atom[n];
                                                    }
                                                    continue;
                                               }
                                          }
                                     } else {
                                          atom[0] = residue->find_atom(ext->lists[k].chiral[l].center_atom, num[0]);
                                          atom[1] = residue->find_atom(ext->lists[k].chiral[l].first_atom, num[1]);
                                          atom[2] = residue->find_atom(ext->lists[k].chiral[l].second_atom, num[2]);
                                          atom[3] = residue->find_atom(ext->lists[k].chiral[l].check_atom, num[3]);
                                     }
                                     if (!atom[0] || !atom[1] || !atom[2] || !atom[3]) {
                                          for (n = 0; n < 4; n++) {
                                               if (atom[n]) delete [] atom[n];
                                          }
                                          continue;
                                     }
                                     exist[l] = 1;

                                     for (n = 0; n < 4; n++) {
                                          Atom[n] = NULL;
                                          for (j = 0; j < num[n]; j++) {
                                               if (!strcmp(atom[n][j]->alt_loc(), "") ||
                                                   !strcmp(alt_loc, "") ||
                                                    atom[n][j]->alt_loc()[0] == alt_loc[m]) {
                                                    Atom[n] = atom[n][j];
                                                    break;
                                               }
                                          }
                                     }

                                     if (Atom[0] && Atom[1] && Atom[2] && Atom[3]) {
                                          volume = calc_chiral_volume(Atom[0], Atom[1], Atom[2],
                                                      Atom[3], ang);
                                          if (volume * ext->lists[k].chiral[l].correct_value < 0.0) {
                                               atoms[l] = Atom[3];
                                          }
                                     }
                                     for (n = 0; n < 4; n++) {
                                          if (atom[n]) delete [] atom[n];
                                     }
                                }
                                if (exist[0] && exist[1]) {
                                     if (atoms[0] && atoms[1]) {
                                          change = 1;
                                          for (l = 0; l < 2; l++) {
                                               atom[l] = residue->find_atom(atoms[l]->atmtype(), num[l]);
                                               for (n = 0; n < num[l]; n++) {
                                                    if (!strcmp(atom[l][n]->alt_loc(), "") ||
                                                        !strcmp(alt_loc, "") ||
                                                          atom[l][n]->alt_loc()[0] == alt_loc[m]) {
                                                          switch_link_atoms(atom[l][n]->atmtype(),
                                                            atom[l][n]->alt_loc(),
                                                            ext->lists[k].chiral[l].change_to,
                                                            chain->index(), residue->index(),
                                                            _n_link, _link);
                                                         atom[l][n]->set_atmtype(ext->lists[k].chiral[l].change_to);
                                                         atom[l][n]->set_pdb_atmnam(ext->lists[k].chiral[l].change_to);
                                                    }
                                               }
                                               if (atom[l]) { delete [] atom[l]; atom[l] = NULL; }

                                               for (n = 0; n < ext->lists[k].chiral[l].n_change; n++) {
                                                    atom[0] = residue->find_atom(ext->lists[k].chiral[l].
                                                                 change_list[n].old_name, num[0]);
                                                    if (atom[0]) {
                                                         for (j = 0; j < num[0]; j++) {
                                                              if (!strcmp(atom[0][j]->alt_loc(), "") ||
                                                                  !strcmp(alt_loc, "") ||
                                                                   atom[0][j]->alt_loc()[0] == alt_loc[m]) {
                                                                   switch_link_atoms(atom[0][j]->atmtype(),
                                                                      atom[0][j]->alt_loc(), ext->lists[k].
                                                                      chiral[l].change_list[n].new_name,
                                                                      chain->index(), residue->index(),
                                                                      _n_link, _link);
                                                                   atom[0][j]->set_atmtype(ext->lists[k].chiral[l].
                                                                                 change_list[n].new_name);
                                                                   atom[0][j]->set_pdb_atmnam(ext->lists[k].chiral[l].
                                                                                 change_list[n].new_name);
                                                                   break;
                                                              }
                                                         }
                                                         delete [] atom[0]; atom[0] = NULL;
                                                    }
                                               }
                                          }
                                     }
                                } else if (exist[0]) {
                                     if (atoms[0]) {
                                          change = 1;
                                          // atoms[0]->set_atmtype(ext->lists[k].chiral[0].change_to);
                                          // atoms[0]->set_pdb_atmnam(ext->lists[k].chiral[0].change_to);
                                          atom[0] = residue->find_atom(atoms[0]->atmtype(), num[0]);
                                          for (n = 0; n < num[0]; n++) {
                                               if (!strcmp(atom[0][n]->alt_loc(), "") ||
                                                   !strcmp(alt_loc, "") ||
                                                    atom[0][n]->alt_loc()[0] == alt_loc[m]) {
                                                    switch_link_atoms(atom[0][n]->atmtype(),
                                                      atom[0][n]->alt_loc(),
                                                      ext->lists[k].chiral[0].change_to,
                                                      chain->index(), residue->index(),
                                                      _n_link, _link);
                                                    atom[0][n]->set_atmtype(ext->lists[k].chiral[0].change_to);
                                                    atom[0][n]->set_pdb_atmnam(ext->lists[k].chiral[0].change_to);
                                               }
                                          }
                                          if (atom[0]) { delete [] atom[0]; atom[0] = NULL; }

                                          for (n = 0; n < ext->lists[k].chiral[0].n_change; n++) {
                                               atom[0] = residue->find_atom(ext->lists[k].chiral[0].
                                                            change_list[n].old_name, num[0]);
                                               if (atom[0]) {
                                                    for (j = 0; j < num[0]; j++) {
                                                         if (!strcmp(atom[0][j]->alt_loc(), "") ||
                                                             !strcmp(alt_loc, "") ||
                                                              atom[0][j]->alt_loc()[0] == alt_loc[m]) {
                                                              switch_link_atoms(atom[0][j]->atmtype(),
                                                                 atom[0][j]->alt_loc(), ext->lists[k].
                                                                 chiral[0].change_list[n].new_name,
                                                                 chain->index(), residue->index(),
                                                                 _n_link, _link);
                                                              atom[0][j]->set_atmtype(ext->lists[k].chiral[0].
                                                                            change_list[n].new_name);
                                                              atom[0][j]->set_pdb_atmnam(ext->lists[k].chiral[0].
                                                                            change_list[n].new_name);
                                                         }							  
                                                    }
                                                    delete [] atom[0]; atom[0] = NULL;
                                               }
                                          }
                                     }
                                } else if (exist[1]) {
                                     if (atoms[1]) {
                                          change = 1;
                                          // atoms[1]->set_atmtype(ext->lists[k].chiral[1].change_to);
                                          // atoms[1]->set_pdb_atmnam(ext->lists[k].chiral[1].change_to);
                                          atom[1] = residue->find_atom(atoms[1]->atmtype(), num[1]);
                                          for (n = 0; n < num[1]; n++) {
                                               if (!strcmp(atom[1][n]->alt_loc(), "") ||
                                                   !strcmp(alt_loc, "") ||
                                                    atom[1][n]->alt_loc()[0] == alt_loc[m]) {
                                                    switch_link_atoms(atom[1][n]->atmtype(),
                                                      atom[1][n]->alt_loc(),
                                                      ext->lists[k].chiral[1].change_to,
                                                      chain->index(), residue->index(),
                                                      _n_link, _link);
                                                    atom[1][n]->set_atmtype(ext->lists[k].chiral[1].change_to);
                                                    atom[1][n]->set_pdb_atmnam(ext->lists[k].chiral[1].change_to);
                                               }
                                          }
                                          if (atom[1]) { delete [] atom[1]; atom[1] = NULL; }

                                          for (n = 0; n < ext->lists[k].chiral[1].n_change; n++) {
                                               atom[0] = residue->find_atom(ext->lists[k].chiral[1].
                                                            change_list[n].old_name, num[0]);
                                               if (atom[0]) {
                                                    for (j = 0; j < num[0]; j++) {
                                                         if (!strcmp(atom[0][j]->alt_loc(), "") ||
                                                             !strcmp(alt_loc, "") ||
                                                              atom[0][j]->alt_loc()[0] == alt_loc[m]) {
                                                              switch_link_atoms(atom[0][j]->atmtype(),
                                                                 atom[0][j]->alt_loc(), ext->lists[k].
                                                                 chiral[1].change_list[n].new_name,
                                                                 chain->index(), residue->index(),
                                                                 _n_link, _link);
                                                              atom[0][j]->set_atmtype(ext->lists[k].chiral[1].
                                                                            change_list[n].new_name);
                                                              atom[0][j]->set_pdb_atmnam(ext->lists[k].chiral[1].
                                                                            change_list[n].new_name);
                                                         }
                                                    }
                                                    delete [] atom[0]; atom[0] = NULL;
                                               }
                                          }
                                     }
                                }
                           }
                      }
                      if (change) {
                           residue->UpdateIndices();
                           residue->reorder_atoms();
                      }
                      prev = residue;
                      residue = chain->GetNextResidue();
                 }
/*
                 if (_experiment_type & EXPERIMENT_TYPE_NMR &&
                     chain->chain_type() == ATOMP_TOKEN) {
                      residue = chain->GetLastResidue();
                      if (residue)
                           ndb_check_E_Z_switch(residue, &_C_TERMINAL);
                 }
*/
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

static void check_first_flag(FILE *template_fp, FILE *fp, int& first_flag,
                             const int model_num) 
{
       if (first_flag) {
            first_flag = 0;
            ndb_writing_missing_words(template_fp, fp);
            if (model_num >= 0) {
                 fprintf(fp, "Model  Chain  Residue  Residue  Chiral       Chiral Neighbors      Original\n");
                 fprintf(fp, "               Name    Number   Center                             Atom Name\n");
                 fprintf(fp, "-----  -----  -------  ------- --------  ------------------------  ---------\n");
            } else {
                 fprintf(fp, "Chain  Residue  Residue  Chiral       Chiral Neighbors      Original\n");
                 fprintf(fp, "        Name    Number   Center                             Atom Name\n");
                 fprintf(fp, "-----  -------  ------- --------  ------------------------  ---------\n");
            }
       }
}

static void print_chiral_error(FILE *fp, const int model_num, const char* ResName,
                               const int pdb_res_no, const char *pdb_chnid, _ATOM *fst,
                               _ATOM *snd, _ATOM *trd, _ATOM *fth, _ATOM *f5h) 
{
       std::string name1, name2, name3, name4, name5, names;
       get_atom_name_with_alt_loc(name1, fst);
       get_atom_name_with_alt_loc(name2, snd);
       get_atom_name_with_alt_loc(name3, trd);
       get_atom_name_with_alt_loc(name4, fth);
       name5.clear();
       if (f5h) {
            get_atom_name_with_alt_loc(name5, f5h);
            names = name4;
       } else names = name2 + "," + name3 + "," + name4;
       if (model_num >= 0) fprintf(fp, "%3d    ", model_num);
       fprintf(fp, "%s      %4s    %5d    %-8s  %-24s  ", pdb_chnid, ResName,
                   pdb_res_no, name1.c_str(), names.c_str());
       if (f5h && strcmp(f5h->pdb_atmnam(), f5h->pub_atmnam()))
            fprintf(fp, "%s\n", name5.c_str());
       else fprintf(fp, "\n");
}

static void get_atom_name_with_alt_loc(std::string& name, _ATOM *atom)
{
       name = atom->pdb_atmnam();
       if (strcmp(atom->alt_loc(), "")) {
            name += "(";
            name += atom->alt_loc();
            name += ")";
       }
}

static void ndb_check_E_Z(_RESIDUE *residue, const int Chn_Index, const int n_link,
             _LINK **link)
{
       int i = 0;

       for (i = 0; i < NUM_EZ_CHECK; i++) {
            if (strcmp(residue->ResName(), ez_check[i].residue)) continue;
            ndb_check_E_Z_switch(residue, &ez_check[i], Chn_Index, n_link, link);
       }
}

static void ndb_check_E_Z_switch(_RESIDUE *residue, EZ_CHECK *ez_check,
             const int Chn_Index, const int n_link, _LINK **link)
{
       int j = 0, found_all_atoms = 1;
       _ATOM *atom1 = NULL, *atom2 = NULL, *atom3 = NULL, *atom4 = NULL;
       double torsion[2];

       for (j = 0; j < ez_check->n_check; j++) {
            atom1 = residue->find_atom(ez_check->check_list[j].atom);
            if (atom1 == NULL) {
                 found_all_atoms = 0; break;
            }
       }
       if (!found_all_atoms) return;

       if (strcmp(ez_check->no_atom, "") &&
           residue->find_atom(ez_check->no_atom)) return;

       atom1 = residue->find_atom(ez_check->atom1);
       atom2 = residue->find_atom(ez_check->atom2);
       atom3 = residue->find_atom(ez_check->atom3);
       atom4 = residue->find_atom(ez_check->atom4);
       if (!atom1 || !atom2 || !atom3 || !atom4) return;

       torsion[0] = cal_torsion(atom1, atom2, atom3, atom4);
       torsion[0] = mod(torsion[0], -180, 180);
       if (fabs(torsion[0]) < 90.0) return;

       for (j = 0; j < 2; j++) {
            atom4 = residue->find_atom(ez_check->check_list[j].atom);
            torsion[j] = cal_torsion(atom1, atom2, atom3, atom4);
            torsion[j] = mod(torsion[j], -180, 180);
       }
       if (fabs(torsion[0]) < fabs(torsion[1])) return;
       
       for (j = 0; j < ez_check->n_change; j++) {
            atom1 = residue->find_atom(ez_check->change_list[j].old_name);
            atom2 = residue->find_atom(ez_check->change_list[j].new_name);
            // if (atom1 && atom2) {
            if (atom1) {
                 switch_link_atoms(ez_check->change_list[j].old_name, "",
                       ez_check->change_list[j].new_name, Chn_Index,
                       residue->index(), n_link, link);
                 atom1->set_atmtype(ez_check->change_list[j].new_name);
                 atom1->set_pdb_atmnam(ez_check->change_list[j].new_name);
            }
            if (atom2) {
                 switch_link_atoms(ez_check->change_list[j].new_name, "",
                       ez_check->change_list[j].old_name, Chn_Index,
                       residue->index(), n_link, link);
                 atom2->set_atmtype(ez_check->change_list[j].old_name);
                 atom2->set_pdb_atmnam(ez_check->change_list[j].old_name);
            }
       }
       residue->UpdateIndices();
       residue->reorder_atoms();
}

static void switch_link_atoms(const char *old_name, const char *alt_loc,
          const char *new_name, const int ChnIndex, const int ResIndex,
          const int n_link, _LINK **link)
{
       if (!n_link || !link) return;

       for (int i = 0; i < n_link; i++) {
            if (!strcmp(link[i]->FstAtom, old_name) &&
                !strcmp(link[i]->altLoc1, alt_loc) &&
                link[i]->FstIndex == ChnIndex &&
                link[i]->FstRes == ResIndex) {
                 if (link[i]->fst_changed) break;
                 strcpy(link[i]->FstAtom, new_name);
                 link[i]->fst_changed = 1;
                 break;
            }
            if (!strcmp(link[i]->SndAtom, old_name) &&
                !strcmp(link[i]->altLoc2, alt_loc) &&
                link[i]->SndIndex == ChnIndex &&
                link[i]->SndRes == ResIndex) {
                 if (link[i]->snd_changed) break;
                 strcpy(link[i]->SndAtom, new_name);
                 link[i]->snd_changed = 1;
                 break;
            }
       }
}

static void get_defined_links(std::map<std::string, std::vector<std::string> >& links,
                              const int n_link, _LINK **link, _MOLECULE *mol)
{
       links.clear();
       if (!n_link || !link) return;

       std::vector<std::string> data;
       for (int i = 0; i < n_link; ++i) {
            if (strcmp(link[i]->SymOP_1, "") ||
                strcmp(link[i]->SymOP_2, "")) continue;
            _CHAIN *ch1 = mol->GetIndexChain(link[i]->FstIndex);
            _CHAIN *ch2 = mol->GetIndexChain(link[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            _FIELD *Init = ch1->SeqResIndex(link[i]->FstRes);
            _FIELD *End  = ch2->SeqResIndex(link[i]->SndRes);
            if (!Init || !End) continue;

            std::string cs = ch1->PDB_ChainID();
            cs += "_"; cs += link[i]->FstName;
            cs += "_"; cs += Init->Field[4];
            cs += "_"; cs += Init->InsCode;
            cs += "_"; cs += link[i]->FstAtom;
            cs += "_"; cs += link[i]->altLoc1;
            data.clear();
            data.push_back(ch2->PDB_ChainID());
            data.push_back(link[i]->SndName);
            data.push_back(End->Field[4]);
            data.push_back(End->InsCode);
            data.push_back(link[i]->SndAtom);
            data.push_back(link[i]->altLoc2);
            links.insert(std::make_pair(cs, data));

            cs = ch2->PDB_ChainID();
            cs += "_"; cs += link[i]->SndName;
            cs += "_"; cs += End->Field[4];
            cs += "_"; cs += End->InsCode;
            cs += "_"; cs += link[i]->SndAtom;
            cs += "_"; cs += link[i]->altLoc2;
            data.clear();
            data.push_back(ch1->PDB_ChainID());
            data.push_back(link[i]->FstName);
            data.push_back(Init->Field[4]);
            data.push_back(Init->InsCode);
            data.push_back(link[i]->FstAtom);
            data.push_back(link[i]->altLoc1);
            links.insert(std::make_pair(cs, data));
       }
}

static void additional_chiral_checking(FILE *template_fp, FILE *fp, int &first_flag,
                         const int _num_mol, _MOLECULE **_molecules,
                         std::map<std::string, std::vector<std::string> >& links,
                         CifString& error)
{
       std::map<std::string, std::vector<std::vector<std::string> > >
                        chiral_center_mapping;
       chiral_center_mapping.clear();
       std::map<std::string, _ATOM*> exist_links;
       ReVarCifArray<CifString> res_list;

       for (int i = 0; i < _num_mol; ++i) {
            int model = -1;
            if (_num_mol > 1) model = _molecules[i]->Mol_ID();
            get_exist_links(exist_links, _molecules[i], links);
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      residue->get_residue_list(res_list);
                      for (unsigned j = 0; j < res_list.Length(); ++j) {
                           if (chiral_center_mapping.find(res_list[j].Text()) ==
                               chiral_center_mapping.end()) {
                                get_chiral_center_mapping(chiral_center_mapping,
                                                          res_list[j].Text());
                           }
                           std::map<std::string,
                                    std::vector<std::vector<std::string> > >::iterator
                                pos = chiral_center_mapping.find(res_list[j].Text());
                           if (!pos->second.empty()) { 
                                checking_chiral_center(template_fp, fp, first_flag,
                                          model, pos->second, res_list[j].Text(),
                                          residue, exist_links, error);
                           }
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

static void switch_other_O1P_O2P_atoms(_CHAIN *chain, const int n_link, _LINK **link)
{
       return; // disable switch option
       std::vector<std::vector<std::string> > P_chiral_centers;
       _RESIDUE *residue = chain->GetFirstResidue();
       while (residue) {
            if (is_aa_residue(residue->ResName()) ||
                is_na_residue(residue->ResName()) ||
                !strcmp(residue->ResName(), "HOH") ||
                !strcmp(residue->ResName(), "DOD")) {
                 residue = chain->GetNextResidue();
                 continue;
            }
            get_P_chiral_centers(P_chiral_centers, residue->ResName());
            if (!P_chiral_centers.empty()) {
                 checking_P_chiral_center(P_chiral_centers, chain->index(), residue,
                                          n_link, link);
            }
            residue = chain->GetNextResidue();
       }
}

static void get_exist_links(std::map<std::string, _ATOM*>& exist_links, _MOLECULE* mol,
                            std::map<std::string, std::vector<std::string> >& links)
{
       exist_links.clear();
       if (links.empty()) return;

       int num_a;
       for (std::map<std::string, std::vector<std::string> >::iterator
            pos = links.begin(); pos != links.end(); ++pos) {
            _RESIDUE *res = mol->find_pdb_residue(pos->second[0].c_str(),
                         pos->second[1].c_str(), atoi(pos->second[2].c_str()),
                         pos->second[3].c_str());
            if (!res) continue;

            _ATOM **atom_a = res->find_atom(pos->second[4].c_str(), num_a);
            if (!atom_a) continue;

            for (int i = 0; i < num_a; ++i) {
                 if (pos->second[5] == atom_a[i]->alt_loc()) {
                      exist_links.insert(std::make_pair(pos->first, atom_a[i]));
                      break;
                 }
            }
            delete [] atom_a;
       }
}

static void get_chiral_center_mapping(std::map<std::string,
                  std::vector<std::vector<std::string> > >& mapping, const char *ResName)
{
       std::vector<std::vector<std::string> > list;
       list.clear();

       ConnectFormat *drug = find_drug(ResName);
       if (!drug) {
            mapping.insert(std::make_pair(ResName, list));
            return;
       }

       if (!drug->has_chiral_center()) {
            mapping.insert(std::make_pair(ResName, list));
            return;
       }

       std::vector<AtomFormat*> RefAtoms;
       std::vector<std::string> data;
       for (int i = 0; i < drug->natoms(); ++i) {
            if (drug->atoms(i)->stereo_config() == NULL) continue;
            if (!strcmp(drug->atoms(i)->stereo_config(), "") ||
                !strcmp(drug->atoms(i)->stereo_config(), "N") ||
                // carbon atom only
                strcmp(drug->atoms(i)->atomtype(), "C")) continue;
                // skip phosphate group
                // !strcmp(drug->atoms(i)->atomtype(), "P")) continue;

            RefAtoms.clear();
            RefAtoms.push_back(drug->atoms(i));
            for (int j = 0; j < drug->atoms(i)->nconect(); ++j) {
                 _SIndex *index = drug->find_atom(drug->atoms(i)->connectto(j));
                 if (!index) continue;
                 if (drug->atoms(index->pos())->ishydrogen()) continue;
                 RefAtoms.push_back(drug->atoms(index->pos()));
            }
            if (RefAtoms.size() < 4) continue;
            get_chiral_info(RefAtoms, data);
            if (!data.empty()) list.push_back(data);
       }
       mapping.insert(std::make_pair(ResName, list));
}

static void get_P_chiral_centers(std::vector<std::vector<std::string> >& chiral_centers,
                                 const char *ResName)
{
       chiral_centers.clear();

       ConnectFormat *drug = find_drug(ResName);
       if (!drug) return;

       if (!drug->has_chiral_center()) return;

       std::vector<AtomFormat*> RefAtoms;
       std::vector<std::string> data;
       for (int i = 0; i < drug->natoms(); ++i) {
            if (drug->atoms(i)->stereo_config() == NULL) continue;
            if (!strcmp(drug->atoms(i)->stereo_config(), "") ||
                !strcmp(drug->atoms(i)->stereo_config(), "N") ||
                strcmp(drug->atoms(i)->atomtype(), "P")) continue;

            RefAtoms.clear();
            RefAtoms.push_back(drug->atoms(i));
            for (int j = 0; j < drug->atoms(i)->nconect(); ++j) {
                 _SIndex *index = drug->find_atom(drug->atoms(i)->connectto(j));
                 if (!index) continue;
                 if (drug->atoms(index->pos())->ishydrogen()) continue;
                 RefAtoms.push_back(drug->atoms(index->pos()));
            }
            if (RefAtoms.size() < 4) continue;
            get_chiral_info(RefAtoms, data);
            if (!data.empty()) {
                 reorder_chiral_atoms(data, drug);
                 if (!data.empty()) chiral_centers.push_back(data);
            }
       }
}

static void checking_chiral_center(FILE *template_fp, FILE *fp, int& flag,
                 const int model_num, const std::vector<std::vector<std::string> >&
                 chiral_list, const char *ResName, _RESIDUE *residue,
                 std::map<std::string, _ATOM*>& links, CifString& error)
{
       for (unsigned int i = 0; i < chiral_list.size(); ++i) {
            int num_a = 0;
            int num_b = 0;
            int num_c = 0;
            int num_d = 0;
            _ATOM **atom1 = residue->find_atom(chiral_list[i][0].c_str(), num_a,
                                               ResName);
            if (!atom1) continue;
            _ATOM **atom2 = get_atom_list(atom1, num_a, chiral_list[i][1], num_b,
                                          ResName, residue, links);
            _ATOM **atom3 = get_atom_list(atom1, num_a, chiral_list[i][2], num_c,
                                          ResName, residue, links);
            _ATOM **atom4 = get_atom_list(atom1, num_a, chiral_list[i][3], num_d,
                                          ResName, residue, links);
            if (!atom1 || !atom2 || !atom3 || !atom4) {
                 if (atom1) delete [] atom1;
                 if (atom2) delete [] atom2;
                 if (atom3) delete [] atom3;
                 if (atom4) delete [] atom4;
                 continue;
            }
            int num = 0;
            _FOUR_ATOMS *four_atoms = find_four_atoms(num, atom1, num_a,
                             atom2, num_b, atom3, num_c, atom4, num_d);
            for (int j = 0; j < num; ++j) {
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].snd)) continue;
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].trd)) continue;
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].fth)) continue;
                 double volume = calc_chiral_volume(four_atoms[j].fst->orig(),
                             four_atoms[j].snd->orig(), four_atoms[j].trd->orig(),
                             four_atoms[j].fth->orig());
                 if (volume * atof(chiral_list[i][4].c_str()) < 0.0 ||
                     fabs(volume) <= 0.7) {
                      check_first_flag(template_fp, fp, flag, model_num);
                      print_chiral_error(fp, model_num, ResName,
                          residue->pdb_res_no(), four_atoms[j].fst->pdb_chnid(),
                          four_atoms[j].fst, four_atoms[j].snd, four_atoms[j].trd,
                          four_atoms[j].fth, NULL);
                      write_chiral_error(error, model_num, ResName,
                          residue->pdb_res_no(), four_atoms[j].fst->pdb_chnid(),
                          four_atoms[j].fst, four_atoms[j].snd,
                          four_atoms[j].trd, four_atoms[j].fth);
                 }
            }
            if (four_atoms) delete [] four_atoms;
            if (atom1) delete [] atom1;
            if (atom2) delete [] atom2;
            if (atom3) delete [] atom3;
            if (atom4) delete [] atom4;
       }
}

static void checking_P_chiral_center(const std::vector<std::vector<std::string> >&
                       chiral_list, const int chn_index, _RESIDUE *residue,
                       const int n_link, _LINK **link)
{
       bool changed = false;
       for (unsigned int i = 0; i < chiral_list.size(); ++i) {
            int num_a = 0;
            _ATOM **atom1 = residue->find_atom(chiral_list[i][5].c_str(), num_a);
            if (atom1) {
                 delete [] atom1;
                 continue;
            }
            int num_b = 0;
            int num_c = 0;
            int num_d = 0;
            atom1 = residue->find_atom(chiral_list[i][0].c_str(), num_a);
            _ATOM **atom2 = residue->find_atom(chiral_list[i][1].c_str(), num_b);
            _ATOM **atom3 = residue->find_atom(chiral_list[i][2].c_str(), num_c);
            _ATOM **atom4 = residue->find_atom(chiral_list[i][3].c_str(), num_d);
            if (!atom1 || !atom2 || !atom3 || !atom4) {
                 if (atom1) delete [] atom1;
                 if (atom2) delete [] atom2;
                 if (atom3) delete [] atom3;
                 if (atom4) delete [] atom4;
                 continue;
            }
            int num = 0;
            _FOUR_ATOMS *four_atoms = find_four_atoms(num, atom1, num_a,
                             atom2, num_b, atom3, num_c, atom4, num_d);
            for (int j = 0; j < num; ++j) {
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].snd)) continue;
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].trd)) continue;
                 if (!is_a_bond_loose(four_atoms[j].fst, four_atoms[j].fth)) continue;
                 double volume = calc_chiral_volume(four_atoms[j].fst->orig(),
                             four_atoms[j].snd->orig(), four_atoms[j].trd->orig(),
                             four_atoms[j].fth->orig());
                 if (volume * atof(chiral_list[i][4].c_str()) >= 0.0) continue;
                 changed = true;
                 switch_link_atoms(four_atoms[j].snd->atmtype(),
                                   four_atoms[j].snd->alt_loc(),
                                   chiral_list[i][2].c_str(), chn_index,
                                   residue->index(), n_link, link);
                 switch_link_atoms(four_atoms[j].trd->atmtype(),
                                   four_atoms[j].trd->alt_loc(),
                                   chiral_list[i][1].c_str(), chn_index,
                                   residue->index(), n_link, link);
                 four_atoms[j].snd->set_atmtype(chiral_list[i][2].c_str());
                 four_atoms[j].snd->set_pdb_atmnam(chiral_list[i][2].c_str());
                 four_atoms[j].trd->set_atmtype(chiral_list[i][1].c_str());
                 four_atoms[j].trd->set_pdb_atmnam(chiral_list[i][1].c_str());
            }
            if (four_atoms) delete [] four_atoms;
            if (atom1) delete [] atom1;
            if (atom2) delete [] atom2;
            if (atom3) delete [] atom3;
            if (atom4) delete [] atom4;
       }
       if (changed) {
            residue->UpdateIndices();
            residue->reorder_atoms();
       }
}
 
static void get_chiral_info(const std::vector<AtomFormat*>& RefAtoms,
                            std::vector<std::string>& data)
{
       data.clear();

       std::vector<COORD> coords;
       coords.clear();
       COORD coord;

       for (unsigned int i = 0; i < RefAtoms.size(); ++i) {
            if (RefAtoms[i]->coord_x() && strcmp(RefAtoms[i]->coord_x(), "") &&
                RefAtoms[i]->coord_y() && strcmp(RefAtoms[i]->coord_y(), "") &&
                RefAtoms[i]->coord_z() && strcmp(RefAtoms[i]->coord_z(), "")) {
                 coord.x = atof(RefAtoms[i]->coord_x());
                 coord.y = atof(RefAtoms[i]->coord_y());
                 coord.z = atof(RefAtoms[i]->coord_z());
                 coords.push_back(coord);
                 if (RefAtoms[i]->leaving_atom_flag() &&
                     !strcasecmp(RefAtoms[i]->leaving_atom_flag(), "Y")) {
                      std::string cs = "-";
                      cs += RefAtoms[i]->atomname();
                      data.push_back(cs);
                 } else data.push_back(RefAtoms[i]->atomname());
            }
       }
       if (coords.size() < 4) {
            data.clear();
            return;
       }

       double volume = calc_chiral_volume(coords[0], coords[1], coords[2], coords[3]);
       char buffer[20];
       memset(buffer, 0, 20);
       sprintf(buffer, "%.4f", volume);
       if (data.size() > 4)
            data[4] = buffer; 
       else data.push_back(buffer);
}

static _ATOM **get_atom_list(_ATOM **c_atom, const int c_num,  const std::string& name,
                             int& num, const char *ResName, _RESIDUE *residue,
                             std::map<std::string, _ATOM*>& links)
{
       num = 0;
       if (name[0] != '-') return residue->find_atom(name.c_str(), num, ResName);

       _ATOM **atom = residue->find_atom(name.substr(1).c_str(), num, ResName);
       if (atom != NULL || links.empty()) return atom;

       char buffer[10];
       std::vector<_ATOM*> atom_list;
       atom_list.clear();
 
       for (int i = 0; i < c_num; ++i) {
            std::string cs = c_atom[i]->pdb_chnid();
            cs += "_"; cs += c_atom[i]->pdb_resnam();
            sprintf(buffer, "%d", c_atom[i]->pdb_resnum());
            cs += "_"; cs += buffer;
            cs += "_"; cs += c_atom[i]->ins_code();
            cs += "_"; cs += c_atom[i]->pdb_atmnam();
            cs += "_"; cs += c_atom[i]->alt_loc();
            std::map<std::string, _ATOM*>::iterator pos = links.find(cs);
            if (pos == links.end()) continue;

            atom_list.push_back(pos->second);
       }

       if (!atom_list.empty()) {
            num = atom_list.size();
            atom = new _ATOM*[num];
            for (unsigned int i = 0; i < atom_list.size(); ++i) {
                 atom[i] = atom_list[i];
            }
       }

       return atom;
}

void static reorder_chiral_atoms(std::vector<std::string>& list, ConnectFormat *drug)
{
       int op1_idx = -1;
       int op2_idx = -1;
       std::string h1, h2;
       h1.clear();
       h2.clear();

       for (int i = 1; i < 4; ++i) {
            if (list[i].find("1") != std::string::npos &&
                is_right_oxygen(drug, list[0], list[i], h1)) op1_idx = i;
            if (list[i].find("2") != std::string::npos &&
                is_right_oxygen(drug, list[0], list[i], h2)) op2_idx = i;
       }
       if (op1_idx < 0 || op2_idx < 0 || (h1.empty() && h2.empty())) {
            list.clear();
            return;
       }

       std::vector<std::string> data;
       data.clear();
       data.push_back(list[0]);
       data.push_back(list[op1_idx]);
       data.push_back(list[op2_idx]);
       for (int i = 1; i < 4; ++i) {
            if (i == op1_idx || i == op2_idx) continue;
            data.push_back(list[i]);
            break;
       }
       if (data.size() != 4) {
            list.clear();
            return;
       }
       data.push_back(list[4]);     
       if (!h1.empty())
            data.push_back(h1);
       else if (!h2.empty())
            data.push_back(h2);
       list = data;
}

static bool is_right_oxygen(ConnectFormat *drug, const std::string& p_atom,
                            const std::string& o_atom, std::string& h_atom)
{
       h_atom.clear();

       _SIndex *index = drug->find_atom(o_atom.c_str());
       if (!index) return false;

       int i = index->pos();
       if (strcmp(drug->atoms(i)->atomtype(), "O")) return false;
       for (int j = 0; j < drug->atoms(i)->nconect(); ++j) { 
            if (p_atom == drug->atoms(i)->connectto(j)) continue;
            index = drug->find_atom(drug->atoms(i)->connectto(j));
            if (!index) return false;
            if (drug->atoms(index->pos())->ishydrogen())
                 h_atom = drug->atoms(i)->connectto(j);
            else return false;
       }
       return true;
}

static void write_chiral_error(CifString& error, const int model_num,
                               const char *ResName, const int pdb_res_no,
                               const char *pdb_chnid, _ATOM *fst, _ATOM *snd,
                               _ATOM *trd, _ATOM *fth)
{
       if (error == "") {
            if (model_num >= 0) {
                 error +=  "Model  Chain  Residue  Residue  Chiral       Chiral Neighbors      Original\n";
                 error +=  "               Name    Number   Center                             Atom Name\n";
                 error +=  "-----  -----  -------  ------- --------  ------------------------  ---------\n";
            } else {
                 error +=  "Chain  Residue  Residue  Chiral       Chiral Neighbors      Original\n";
                 error +=  "        Name    Number   Center                             Atom Name\n";
                 error +=  "-----  -------  ------- --------  ------------------------  ---------\n";
            }
       }
       char buffer[1000];
       std::string name1, name2, name3, name4, names;
       get_atom_name_with_alt_loc(name1, fst);
       get_atom_name_with_alt_loc(name2, snd);
       get_atom_name_with_alt_loc(name3, trd);
       get_atom_name_with_alt_loc(name4, fth);
       names = name2 + "," + name3 + "," + name4;
       if (model_num >= 0)
            sprintf(buffer, "%3d    %s      %4s    %5d    %-8s  %-24s\n",
                            model_num, pdb_chnid, ResName, pdb_res_no,
                            name1.c_str(), names.c_str());
       else sprintf(buffer, "%s      %4s    %5d    %-8s  %-24s\n", pdb_chnid,
                            ResName, pdb_res_no, name1.c_str(), names.c_str());
       error += buffer;
}

#if 0
static void ndb_check_methyl_protons(_RESIDUE *residue)
{
       int i, j;
       _ATOM *atom[NUM_ATOM];
       MP_TORSION torsion[3];

       for (i = 0; i < NUM_MP_CHECK; i++) {
            if (strcmp(residue->ResName(), mp_check[i].residue)) continue;
            for (j = 0; j < NUM_ATOM; j++) {
                 atom[j] = residue->find_atom(mp_check[i].check_list[j].atom);
                 if (!atom[j]) break;
            }
            if (j < NUM_ATOM) continue;

            torsion[0].torsion = cal_torsion(atom[FIRST], atom[SECOND], atom[THIRD], atom[FOURTH]);
            torsion[1].torsion = cal_torsion(atom[FIRST], atom[SECOND], atom[THIRD], atom[FIFTH]);
            torsion[2].torsion = cal_torsion(atom[FIRST], atom[SECOND], atom[THIRD], atom[SIXTH]);
            for (j = 0; j < 3; j++) {
                 torsion[j].torsion = mod(torsion[j].torsion, -180, 180);
                 torsion[j].old_index = j;
                 torsion[j].new_index = j;
            }
            find_new_atom_labeling(torsion);
            for (j = 0; j < 3; j++) {
                 atom[FOURTH+torsion[j].old_index]->
                        set_atmtype(mp_check[i].check_list[FOURTH+torsion[j].new_index].atom);
                 atom[FOURTH+torsion[j].old_index]->
                        set_pdb_atmnam(mp_check[i].check_list[FOURTH+torsion[j].new_index].atom);
            }
            residue->UpdateIndices();
            residue->reorder_atoms();
       }
}

static void find_new_atom_labeling(MP_TORSION *torsion)
{
       int i = 0, j, k, l;
       double minimum = 360.0;
       for (j = 0; j < 3; j++) {
            if (fabs(torsion[j].torsion) < minimum) {
                 minimum = fabs(torsion[j].torsion);
                 i = j;
            }
       }
       get_j_k(i, j, k);
       if (fabs(torsion[j].torsion) > fabs(torsion[k].torsion)) {
            l = k; k = j; j = l;
       }
       if (fabs(minimum - 60.0) < 0.001) {
            if (torsion[i].torsion < 0) {
                 if (fabs(torsion[j].torsion - 60.0) < 0.001) {
                      torsion[i].new_index = 2;
                      torsion[j].new_index = 0;
                      torsion[k].new_index = 1;
                 } else {
                      torsion[i].new_index = 0;
                      torsion[j].new_index = 1;
                      torsion[k].new_index = 2;
                 } 
            } else {
                 torsion[i].new_index = 0;
                 torsion[k].new_index = 1;
                 torsion[j].new_index = 2;
            }
       } else if (minimum < 5.0) {
            torsion[i].new_index = 0;
            if (torsion[j].torsion > 0 &&
                torsion[k].torsion < 0) {
                 torsion[j].new_index = 1;
                 torsion[k].new_index = 2;
            } else if (torsion[j].torsion < 0 &&
                       torsion[k].torsion > 0) {
                 torsion[j].new_index = 2;
                 torsion[k].new_index = 1;
            }
       } else if (torsion[i].torsion < 0) {
            torsion[i].new_index = 0;
            torsion[j].new_index = 1;
            torsion[k].new_index = 2;
       } else if (torsion[i].torsion > 0) {
            torsion[i].new_index = 0;
            torsion[k].new_index = 1;
            torsion[j].new_index = 2;
       }
}

static void get_j_k(int i, int &j, int &k)
{
       if (i == 0) {
            j = 1; k = 2;
       } else if (i == 1) {
            j = 0; k = 2;
       } else if (i == 2) {
            j = 0; k = 1;
       }
}
#endif
